
/*****************************************************************************************
 * Dipolar.java

 * Version:           0.1

 * Description: 

 * Represent and Read the experimental RDC data.

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 *  If you use publish any results derived from the use of this program
 *  please cite: Exact Solutions for Internuclear Vectors and Backbone
 *  Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 *  and Their Application in a Systematic Search Algorithm for
 *  Determining Protein Backbone Structure. L. Wang and
 *  B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 *  Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 *  <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;

public class Dipolar implements Cloneable{
    private int    residueNo;  //reside Number in the protein sequence
    private String residue;    //type of residue
    private double expRdc;    //experimental data
    private double expErr;    //error in RDCs, not needed by the program

    public Dipolar(){
	residueNo = 0;
	residue = null;
	expErr = 0.0;
	expRdc = 0.0;
    }		
    public Dipolar(int no){
	residueNo = no;
	residue = null;
	expErr = 0.0;
	expRdc = 0.0;
    }		
    public Dipolar(int no,String resid, double rdc, double err){
	residueNo = no;
	residue = resid;
	expRdc = rdc;
	expErr = err;	   
    }	

    //getting the values	
    public int getResidueNo(){
	return residueNo;
    }	
    public String getResidue(){
	return residue;
    }	
    public double getErr(){
	return expErr;
    }	
    public double getRdc(){
	return expRdc;
    }

    //set the values	
    public void setResidueNo(int R){
	residueNo = R;
    }	
    public void setErr(double err){
	expErr = err;
    }	
    public void setRdc(double rdc){
	expRdc = rdc;
    }
	
    public String toString(){
	String desc = String.valueOf(residueNo) +"    "+ String.valueOf(expRdc)
	    +"  "+String.valueOf(expErr);
	return desc;	
    }

    public static class rdcComparator implements Comparator{
	public int compare(Object o1, Object o2){
	    Dipolar n1 = (Dipolar)o1;
	    Dipolar n2 = (Dipolar)o2;
	    double d1 = n1.getResidueNo();
	    double d2 = n2.getResidueNo();
	    if (d1 < d2)
		return -1;
	    else if (d1 > d2)
		return 1;
	    else return 0;
	}
    }

    /**
     * A method for reading RDC files to extracing RDCs belong to a
     * 2ary type. In the current version, assume the RDC file has three columns each line
     @param filename the name of the RDC file
     @param idVec a vector of IDof2aryStructure objects (the secondary
                  structure identification for the protein)
     @param idType the 2ary structure type
     @return a vector of RDCs belong to a 2ary type like ("E" for beta-strand and "H" for helix)
     */
    public Vector rdcReader(String filename, final Vector idVec, String idType){
	Vector rdcVec = new Vector();
	int i, j, no;
	double rdc, err;
	try{
	    StreamTokenizer in =
		new StreamTokenizer(
				    new FileReader(filename));
	    while( in.nextToken() != StreamTokenizer.TT_EOF ){
		if (in.ttype == StreamTokenizer.TT_WORD)
		    ;
		else if (in.ttype == StreamTokenizer.TT_NUMBER){
		    no = (int)in.nval;
  		    in.nextToken();         
 		    rdc = in.nval;  
		    in.nextToken();
 		    err = in.nval;
//  		    in.nextToken();   //If the file has four columns
		    rdcVec.add(new Dipolar(no,"", rdc,err));
		}	
	    }			
	}catch (FileNotFoundException e) {
	    System.out.println("File not found: " + filename);
	}catch (IOException e) {
	    System.out.println("IOException: the stack trace is:");
	    e.printStackTrace();
	}
	Collections.sort(rdcVec,  new rdcComparator() ); 
	//Extract RDCs for individual 2ary structural elements
	IDof2aryStructure id2ary = new IDof2aryStructure();
	int index = -1;
	int beginNo = 0; 
	int endNo = 0; 
	String type2ary; 
	String type;
	Vector allRdcVec = new Vector();
	Vector oneElementRdc = new Vector();
	for (i=0; i<idVec.size(); i++){
	    id2ary = (IDof2aryStructure)idVec.elementAt(i);
	    beginNo = id2ary.getResidueNoA();
	    endNo   = id2ary.getResidueNoB();
	    type2ary= id2ary.getTypeAndID();
	    type = type2ary.substring(0, 1);
	    if (type.equals(idType)){
		index = Collections.binarySearch(rdcVec, new Dipolar(beginNo), new Dipolar.rdcComparator() );
		if (index > -1 ){
		    for (j = 0; j < (endNo - beginNo+1); j++)
			oneElementRdc.add(rdcVec.elementAt(index+j));
		}
		allRdcVec.add(oneElementRdc);
		oneElementRdc = new Vector();
	    }
	}	    
	return allRdcVec;
    }

    /**
     * A method for reading RDC restraint files. The required file format is: 
     *     residue No,  RDC value, experimental error. 
     * Must have only three items per line. Otherwise, errors.
     @param filename the name of the file
     @return a vector of RDC object
     */
    public Vector rdcRead(String filename){
	Vector inputs = new Vector();
	int no;
	double rdc, err;
	try{
	    StreamTokenizer in =
		new StreamTokenizer(
				    new FileReader(filename));
	    while( in.nextToken() != StreamTokenizer.TT_EOF ){
		if (in.ttype == StreamTokenizer.TT_WORD)
		    ;
		else if (in.ttype == StreamTokenizer.TT_NUMBER){
		    no = (int)in.nval;
  		    in.nextToken();         
 		    rdc = in.nval; 
		    in.nextToken();
 		    err = in.nval;
//  		    in.nextToken();   //If the file has four columns
		    inputs.add(new Dipolar(no,"", rdc,err));
		}	
	    }			
	}catch (FileNotFoundException e) {
	    System.out.println("File not found: " + filename);
	}catch (IOException e) {
	    System.out.println("IOException: the stack trace is:");
	    e.printStackTrace();
	}
	Collections.sort(inputs,  new rdcComparator() ); 
	return inputs;
    }
    protected Object clone(){
        try {
	    Object s = super.clone();     
	    return s;                     
        } catch (CloneNotSupportedException e) {
	    throw new InternalError();
        }
    }
}
