/*****************************************************************************************
 * Hbond.java

 * Version:           0.1

 * Description: 

 * Represent and read the hydrogen bond between two strands

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;
 
public class Hbond implements Cloneable{
    private int residueNoA;   //the Sequence No of one of the two partners of the Hbond
    private int residueNoB;   //the Sequence No of Another partner of the Hbond
    private String atomA;     //the name of Atom A
    private String atomB;     //the name of Atom B
    private String id;        //an id assigned to the H-bond
    private String type;      //the name of Hbond, NOT used in the current implementation

    /** Constructors */
    public Hbond() {
        residueNoA = 0;
        residueNoB = 0;
        atomA = null;
        atomB = null;
        id    = null;        
	type  = null;
    }

   public Hbond(int Ano, String A, int Bno,  String B,  String id, String t) {
        residueNoA = Ano;
        residueNoB = Bno;
        this.atomA = A;
        this.atomB = B;
        this.id   = id;
        this.type = t;
    }
 
    public int getResidueNoA() {
        return residueNoA;
    }
    public int getResidueNoB() {
        return residueNoB;
    }
    public String getAtomA()   {
        return atomA;
    }
    public String getAtomB()   {
        return atomB;
    }
    public String getID()   {
        return id;
    }
    public String getType()   {
        return type;
    }
  
    public String toString()    {
        String desc = residueNoA + atomA + "__"+residueNoB+atomB+"  "+id+"   "+type;
        return desc;
    }
    protected Object clone(){
        try{
            Object s = super.clone();     // clone the stack
            return s;                           // return the clone
        } catch (CloneNotSupportedException e)  {
            throw new InternalError();
        }
    }
 
    public static class HbondComparatorA implements Comparator{
        public int compare(Object o1, Object o2){
            Hbond n1 = (Hbond)o1;
            Hbond n2 = (Hbond)o2;
            int d1 = n1.getResidueNoA();
            int d2 = n2.getResidueNoA();
            if (d1 < d2)
                return -1;
            else if (d1 > d2)
                return 1;
            else return 0;
        }
    }
 
    public static class HbondComparatorB implements Comparator{
        public int compare(Object o1, Object o2){
            Hbond n1 = (Hbond)o1;
            Hbond n2 = (Hbond)o2;
            int d1 = n1.getResidueNoB();
            int d2 = n2.getResidueNoB();
            if (d1 < d2)
                return -1;
            else if (d1 > d2)
                return 1;
            else return 0;
        }
    }

    /*
     *A reader for H-bond data: see the attched file for the format
     @param hbFile the name of the file
     @return a vector of all Hbond objects
     */
    public Vector HbondReader(String hbFile){
        Vector inputs = new Vector();
        int noA, noB;
        String atomA, atomB, id, type;
        try{
            StreamTokenizer in =
                new StreamTokenizer(
                                    new FileReader(hbFile));
            while( in.nextToken() != StreamTokenizer.TT_EOF ){ //ignore the comments
                if (in.ttype == StreamTokenizer.TT_WORD)
                    ;
                else if (in.ttype == StreamTokenizer.TT_NUMBER){
                    noA = (int)in.nval;
                    in.nextToken();
                    atomA = in.sval;
                    in.nextToken();
                    noB = (int)in.nval;
                    in.nextToken();
                    atomB = in.sval;
                    in.nextToken();
                    id = in.sval;
                    in.nextToken();
                    type = in.sval;
                    inputs.add(new Hbond(noA, atomA, noB, atomB, id, type));
                }
            }
        }catch (FileNotFoundException e){
            System.out.println("File not found: " + hbFile);
        }catch (IOException e) {
            System.out.println("IOException: the stack trace is:");
            e.printStackTrace();
        }
        return inputs;
    }

    /**
     * Extract all the H-bonds belong to two paired strands
     @param hbVec a H-bond vector extracted by the above HbondReader method
     @param ids   the id for two paired strands
     */
    public Vector hbOfTwoStrands(Vector hbVec, String ids){
	String id;        //an id assigned to the H-bond
	Hbond hb = new  Hbond();
	Vector hb2Strands = new Vector();
	for (int i = 0; i < hbVec.size(); i++){
	    hb =  (Hbond)hbVec.elementAt(i);
	    if (hb.getID().equals(ids))
		hb2Strands.add(hb);
	}
	return hb2Strands;
    }
}
