/*****************************************************************************************
 * IDof2aryStructure.java

 * Version:           0.1

 * Description: 

 * Represent  secondary structure elements and read the input for their identification

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;
 
public class IDof2aryStructure implements Cloneable{
    private int residueNoA;     //The sequence number of the starting residue 
    private int residueNoB;     //The sequence number of the ending residue 
    private String typeAndID;   //The type of the secondary structure element (H or E)

    /** Constructors */
    public IDof2aryStructure(){
        residueNoA = 0;
        residueNoB = 0;
        typeAndID = null;
    }

   public IDof2aryStructure(int Ano, int Bno, String id){
        residueNoA = Ano;
        residueNoB = Bno;
        this.typeAndID = id;
    }
 
    public int getResidueNoA(){
        return residueNoA;
    }
    public int getResidueNoB(){
        return residueNoB;
    }
    public String getTypeAndID(){
        return typeAndID;
    }
  
    public String toString() {
        String desc = residueNoA + "  "+residueNoB+"  "+typeAndID;
        return desc;
    }
  
    /*
     * A very simple reader for secondary structure identification file
     * The the attached file for the format
     @param idFile the name of the file
     @return a vector of IDof2aryStructure objects
     *
     */
    public Vector IDof2aryStructureReader(String idFile){
        Vector inputs = new Vector();
        int noA, noB;
        String typeID = "";
	String ss = "";
        try{
            StreamTokenizer in =
                new StreamTokenizer(
                                    new FileReader(idFile));
            while( in.nextToken() != StreamTokenizer.TT_EOF ){ //ignore the comments
                if (in.ttype == StreamTokenizer.TT_WORD)
                    ;
                else if (in.ttype == StreamTokenizer.TT_NUMBER){
                    noA = (int)in.nval;
                    in.nextToken();
                    noB = (int)in.nval;
                    in.nextToken();
		    typeID = in.sval;
		    if  (typeID != null){
			ss = typeID.toUpperCase();
			if (ss.indexOf("H") > -1  ||  ss.indexOf("E") > -1 ){
			    inputs.add(new IDof2aryStructure(noA, noB, ss));
			}
		    }
                }
            }
        }catch (FileNotFoundException e) {
            System.out.println("File not found: " + idFile);
        }catch (IOException e) {
            System.out.println("IOException: the stack trace is:");
            e.printStackTrace();
        }
        return inputs;
    }

   /*
     * A reader for secondary structure identification file
     * Since the reader is very simple, you need to follow the file format strictly.
     @param idFile the name of the file
     @return a vector of IDof2aryStructure objects in the form "E1-E2" etc
             Which is consistent with those defined in H-bond file
     *
     */
    public Vector[] getOrder(String fileName){
	String ss ="", str ="";
        Vector inputs = new Vector();
	int index = -1;
        try{
	    BufferedReader in = new BufferedReader(new FileReader(fileName));	
	    ss = in.readLine();
	    while( ss != null ) {
		ss = ss.toUpperCase();
		if ( ss.indexOf("ORDER") > -1 ){
		    while ((ss = in.readLine()) != null &&  (index = ss.indexOf("-")) > -1 ){
			inputs.add(ss.trim());
		    }
		}
		if ( ss == null)
		    break;
		ss = in.readLine();
	    }	
	    in.close();
	}catch (FileNotFoundException e) {
            System.out.println("File not found: " + fileName);
        }catch (IOException e) {
            System.out.println("IOException: the stack trace is:");
            e.printStackTrace();
        }
	Vector[]  orderVec = new Vector[inputs.size()];
	for (int m=0; m<orderVec.length; m++)
	    orderVec[m] = new Vector();
	String idNum1 = "";
	String idNum2 = "";
	String id ="";
	int ind2 = -1;
	for (int i =0; i< inputs.size(); i++){
	    str = (String)inputs.elementAt(i);
	    StringTokenizer st = new StringTokenizer(str, "-");
	    int n = st.countTokens();
	    idNum1 = "E"+st.nextToken();  //"E" as the identification of strand is hard-coded
	    for(int j=0; j<n-1; j++){
		idNum2 = "E"+st.nextToken();
		id = idNum1+"-"+idNum2;
 		orderVec[i].add(id);
		idNum1 = idNum2;
	    }
	}
	return orderVec;
    }

    protected Object clone(){
        try{
            Object s = super.clone();     
            return s;                     
        } catch (CloneNotSupportedException e)  {
            throw new InternalError();
        }
    }
}
