/*****************************************************************************************
 * Model.java

 * Version:           0.1

 * Description: 

 * This is the main driver of our program for determining
 * a protein backbone substructure consisting of \alpha_helices and/or
 * \beta-strands. The program computes all the backbond dihedral
 * ($\phi$, $\psi$) angles of an $\alpha$-helix or a $\beta$-strand
 * based on exact solution and systematic search using two RDCs per
 * residue. To ovecome experimental errors in RDCs the target function
 * also include a term for \phi and \psi angles and a term for H-bond
 * energy for paring \beta-strands. In other words, The computed
 * dihedral angles will (a) satisfy the experimental RDCs in two media
 * as well as possible and (b) simultaneously have the ($\phi$,
 * $\psi$) angles as close as possible to their average values in the
 * PDB.  Please see the following J. Biomol NMR paper for details.

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;

public class Model{

    public Model(){
    }
    /**
     * A reader for parameter file
     @param fileName the name of the parameter 
     @return a vector of paramters for computing Saupe elements and secondary elements
     */
    public Vector parameterReader(String fileName){
	Vector parameterVec = new Vector();
	String ss ="", str ="";
	String ss2   = "";
	int    index = -1;
	int    ind2  = -1;
	String key   = "";
	String value = "";
	Map mapSaupe = new TreeMap();
	Map map1stStrand = new TreeMap();
	Map map2ndStrand = new TreeMap();
	Map mapOtherStrand = new TreeMap();
	try{
	    BufferedReader in = new BufferedReader(new FileReader(fileName));	
	    if ((ss = in.readLine()) != null)
		ss = ss.toUpperCase();
	    while( ss != null ) {
		if ( ss.indexOf("SAUPE") > -1 && ss.indexOf("ELEMENT") > -1 ){
 		    ss = (in.readLine()).toUpperCase();
		    index = ss.indexOf("=");
		    while (index == -1){
			ss = (in.readLine()).toUpperCase();
			index = ss.indexOf("=");
		    }
		    while ( ss != null && index  > -1 ){
			key = ss.substring(0, index).trim();
			ind2 = ss.indexOf(";", index);
			if (ind2 < 0){
			    System.out.println("Error in parameter file: lines need to be ended by \";\", see Example");
			    System.exit(1);
			}
			value = ss.substring(index+1, ind2).trim();
			mapSaupe.put(key, value);
			if ((ss = in.readLine()) != null){
			    ss = ss.toUpperCase();
			    index = ss.indexOf("=");
			}
		    }
		}else if( ss.indexOf("FIRST") > -1 && ss.indexOf("STRAND") > -1 ){
 		    ss = (in.readLine()).toUpperCase();
		    index = ss.indexOf("=");
		    while (index == -1){
			ss = (in.readLine()).toUpperCase();
			index = ss.indexOf("=");
		    }
		    while ( ss != null && index  > -1 ){
			key = ss.substring(0, index).trim();
			ind2 = ss.indexOf(";", index);
			if (ind2 < 0){
			    System.out.println("Error in parameter file: lines need to be ended by \";\", see Example");
			    System.exit(1);
			}
			value = ss.substring(index+1, ind2).trim();
			map1stStrand.put(key, value);
			if ((ss = in.readLine()) != null){
			    ss = ss.toUpperCase();
			    index = ss.indexOf("=");
			}
		    }
		}else if( ss.indexOf("SECOND") > -1 && ss.indexOf("STRAND") > -1 ){
 		    ss = (in.readLine()).toUpperCase();
		    index = ss.indexOf("=");
		    while (index == -1){
			ss = (in.readLine()).toUpperCase();
			index = ss.indexOf("=");
		    }
		    while ( ss != null && index  > -1 ){
			key = ss.substring(0, index).trim();
			ind2 = ss.indexOf(";", index);
			if (ind2 < 0){
			    System.out.println("Error in parameter file: lines need to be ended by \";\", see Example");
			    System.exit(1);
			}
			value = ss.substring(index+1, ind2).trim();
			map2ndStrand.put(key, value);
			if ((ss = in.readLine()) != null){
			    ss = ss.toUpperCase();
			    index = ss.indexOf("=");
			}
		    }
		}else if( ss.indexOf("OTHER") > -1 && ss.indexOf("STRAND") > -1 ){
 		    ss = (in.readLine()).toUpperCase();
		    index = ss.indexOf("=");
		    while (index == -1){
			ss = (in.readLine()).toUpperCase();
			index = ss.indexOf("=");
		    }
		    while ( ss != null && index  > -1 ){
			key = ss.substring(0, index).trim();
			ind2 = ss.indexOf(";", index);
			if (ind2 < 0){
			    System.out.println("Error in parameter file: lines need to be ended by \";\", see Example");
			    System.exit(1);
			}
			value = ss.substring(index+1, ind2).trim();
			mapOtherStrand.put(key, value);
			if ((ss = in.readLine()) != null){
			    ss = ss.toUpperCase();
			    index = ss.indexOf("=");
			}
		    }
		}
		if ((ss = in.readLine()) != null)
		    ss = ss.toUpperCase();
	    }
	    in.close();
	}catch (FileNotFoundException e) {
	    System.out.println("File not found: " + fileName);
	}catch (IOException e) {
	    System.out.println("IOException: the stack trace is:");
	    e.printStackTrace();
	}
	parameterVec.add(mapSaupe);
	parameterVec.add(map1stStrand);
	parameterVec.add(map2ndStrand);
	parameterVec.add(mapOtherStrand);
	return parameterVec;
    }
    /**
     * Compute s beta-sheet with the order specified by orderVec
     @param eRdc1Vec  RDC of medium 1 for all the strands of the sheet
     @param eRdc2Vec  RDC of medium 2 for all the strands of the sheet
     @param eigenValues the four Saupe elements
     @param rot         the relative rotation matrix between the two POFs
     @param HbondVec   the Hbonds between the strands
     @param paraVec    the set of parameters used to compute each strands including the 1st one
     @param orderVe    the order of the strands computed 
     @param printEulerSearch if true it will print out the the detail of rotation search
     @return the computed sheet Pdb vector
     */	
    public Vector sheetModel(Vector eRdc1Vec,  Vector eRdc2Vec, double[] eigenValues, Matrix rot, Vector HbondVec,
			    Vector paraVec, Vector[] orderVec,  boolean printEulerSearch) {
	int i=0, j=0;
	Vector strand1Rdc1  = (Vector)eRdc1Vec.elementAt(0); //2 - 7 representing individual ones (S1)
	Vector strand1Rdc2  = (Vector)eRdc2Vec.elementAt(0);
	Vector sheetPdb = new Vector();
	double[] ramaFilter = new double[]{Const.phiLowBeta,Const.phiHighBeta,Const.psiLowBeta,Const.psiHighBeta};
	double phiAve = Const.phiAveBeta;
	double psiAve = Const.psiAveBeta;

	Map map1stStrand = (Map)paraVec.elementAt(1); //The 2nd  parameter set are for computing the 1st strand
	double resolution = new Double((String)map1stStrand.get("RESOLUTION")).doubleValue();
	int refineCycle = Integer.parseInt((String)map1stStrand.get("REFINECYCLE"));
	int dfsCycle    = Integer.parseInt((String)map1stStrand.get("DFSCYCLE"));
	boolean debugDFS = new Boolean((String)map1stStrand.get("DEBUGDFS")).booleanValue();
	boolean printResults = new Boolean((String)map1stStrand.get("PRINTRESULTS")).booleanValue();
	double w4Angles   = new Double((String)map1stStrand.get("WEIGHT4ANGLES")).doubleValue();
  	PdbRmsd psd = new PdbRmsd();
	ModelRdc mdc = new ModelRdc();
	Hbond hb = new Hbond();

	//For Strand1, the first strand refined with no H-bond terms in score funtion
	Vector pdbStrand1 = mdc.refineStrand(strand1Rdc1, strand1Rdc2, rot, eigenValues, ramaFilter,
					     phiAve, psiAve, refineCycle, dfsCycle, w4Angles, resolution,
					     debugDFS, printResults);

	sheetPdb.addAll(pdbStrand1);
	// 	For Strands whose refinement needs H-bonds
	double w4HB = 4.0;
	Vector hBondVecOfEab = new Vector();
	int nn = 0;
	int index = -1;
	Vector strandRdc1  = new Vector();
	Vector strandRdc2  = new Vector();
	Vector pdbStrand1N = new Vector();
	Vector pdbStrand2 = new Vector();
 	String id2ary = "";

	double resolution2 = 0.0;
	int refineCycle2 = 1;
	int dfsCycle2    = 1;
	boolean debugDFS2  =  false;
	boolean printResults2 = false;
	double w4Angles2   = 1.0;
	double w4HB2       = 1.0;

	if (paraVec.size() > 2){
	    Map map2ndStrand = (Map)paraVec.elementAt(2); //The 3rd parameter set are for computing H-bonded strand
	    resolution2 = new Double((String) map2ndStrand.get("RESOLUTION")).doubleValue();
	    refineCycle2 = Integer.parseInt((String)  map2ndStrand.get("REFINECYCLE"));
	    dfsCycle2    = Integer.parseInt((String)  map2ndStrand.get("DFSCYCLE"));
	    debugDFS2    = new Boolean((String) map2ndStrand.get("DEBUGDFS")).booleanValue();
	    printResults2 = new Boolean((String) map2ndStrand.get("PRINTRESULTS")).booleanValue();
	    w4Angles2   = new Double((String) map2ndStrand.get("WEIGHT4ANGLES")).doubleValue();
	    w4HB2       = new Double((String) map2ndStrand.get("WEIGHT4HBONDS")).doubleValue();
	}
	if (paraVec.size() > 3){
	    Map mapOtherStrand = (Map)paraVec.elementAt(3); //The 4th parameter set are for computing H-bonded strand
	    resolution = new Double((String)mapOtherStrand.get("RESOLUTION")).doubleValue();
	    refineCycle = Integer.parseInt((String)mapOtherStrand.get("REFINECYCLE"));
	    dfsCycle    = Integer.parseInt((String)mapOtherStrand.get("DFSCYCLE"));
	    debugDFS    = new Boolean((String)mapOtherStrand.get("DEBUGDFS")).booleanValue();
	    printResults = new Boolean((String)mapOtherStrand.get("PRINTRESULTS")).booleanValue();
	    w4Angles   = new Double((String)mapOtherStrand.get("WEIGHT4ANGLES")).doubleValue();
	    w4HB       = new Double((String)mapOtherStrand.get("WEIGHT4HBONDS")).doubleValue();
	    for (i=0; i<orderVec.length; i++){
 		pdbStrand1N = pdbStrand1;
		for(j=0; j<orderVec[i].size(); j++){
		    id2ary = (String)orderVec[i].elementAt(j);
		    index = id2ary.lastIndexOf("E");
		    if (index > -1 ){
			nn = Integer.parseInt(id2ary.substring(index+1)) - 1;
			hBondVecOfEab = hb.hbOfTwoStrands(HbondVec, id2ary);
			strandRdc1 = (Vector)eRdc1Vec.elementAt(nn);
			strandRdc2 = (Vector)eRdc2Vec.elementAt(nn);
// 			System.out.println(strandRdc1.size());
// 			System.out.println(strandRdc2.size());
			if (i == 0 && j == 0)            //specifically for the second strand
			    pdbStrand2 = mdc.refineStrand(strandRdc1, strandRdc2, rot, eigenValues, ramaFilter, phiAve, psiAve, 
							   refineCycle2, dfsCycle2, resolution2, w4Angles2, w4HB2, pdbStrand1N,
							   hBondVecOfEab, printEulerSearch, debugDFS2, printResults2);
			else 
			    pdbStrand2 = mdc.refineStrand(strandRdc1, strandRdc2, rot, eigenValues, ramaFilter, phiAve, psiAve, 
							   refineCycle, dfsCycle, resolution, w4Angles, w4HB, pdbStrand1N,
							   hBondVecOfEab, printEulerSearch, debugDFS, printResults);

 			sheetPdb.addAll(pdbStrand2);
 			pdbStrand1N = pdbStrand2;
		    }
		}
	    }
	}
	Collections.sort(sheetPdb, new Pdb.PdbComparator());
	return sheetPdb;
    }

    public static void main(String[] argv){
	long startTime = System.currentTimeMillis();
	int i, j;
	String src   = "../inputFiles/"; //The directory has the input files
	ModelRdc mdc = new ModelRdc();
	Model mod = new Model();

	//Read Parameter file
	String paraFile = src+"parameter.txt";
	Vector paraVec = mod.parameterReader(paraFile);
	//read NOE table
	Noe ee = new Noe();
	String file   = src+"NOEs.txt";
	Vector noeVec = ee.noeReader(file);
	//read H-bonds
	Hbond hb = new Hbond();
	String HbondFile = src+"hydrogenBonds.txt";
	Vector HbondVec = hb.HbondReader(HbondFile);
	//read secondary structure identification
	IDof2aryStructure id2ary = new IDof2aryStructure();
	String id2aryFile = src+"secondaryElements.txt";
	Vector idVec = id2ary.IDof2aryStructureReader(id2aryFile);
	Vector[] orderVec = id2ary.getOrder(id2aryFile);

	Pdb pp = new Pdb();

	//RDC files for one helix used to compute Saupe elements
	String rdc1File = src+"1UBQrdc1.txt";
    	String rdc2File = src+"1UBQrdc2.txt";
  	Dipolar dd  =  new Dipolar();

	Vector eRdc1Vec = dd.rdcReader(rdc1File, idVec, "E"); //read RDCs for strands
	Vector eRdc2Vec = dd.rdcReader(rdc2File, idVec, "E");
	Vector hRdc1Vec = dd.rdcReader(rdc1File, idVec, "H"); //read RDCs for helices
	Vector hRdc2Vec = dd.rdcReader(rdc2File, idVec, "H");
	Vector helixRdc1Vec = (Vector)hRdc1Vec.elementAt(0); //1st helix used for computing Saupe elements
	Vector helixRdc2Vec = (Vector)hRdc2Vec.elementAt(0);

 	double[] saupe = new double[6]; //for returning Saupe elements
	Map saupeMap = (Map) paraVec.elementAt(0); //The first parameters are for computing Saupe elements
	int refineCycle = Integer.parseInt((String)saupeMap.get("REFINECYCLE"));
	int dfsCycle    = Integer.parseInt((String)saupeMap.get("DFSCYCLE"));
	boolean debugDFS = new Boolean((String)saupeMap.get("DEBUGDFS")).booleanValue();
        boolean printResults = new Boolean((String)saupeMap.get("PRINTRESULTS")).booleanValue();

	//For Helix: this helix is used to compute Saupe elements for two media (7 parameters)
	double [] ramaFilter ={Const.phiLowHelix,Const.phiHighHelix,Const.psiLowHelix,Const.psiHighHelix};
	double phiAve = Const.phiAveHelix;
	double psiAve = Const.psiAveHelix;

	double [][] rotArray = new double[3][3];
	Vector helixPdbVec = mdc.refineSaupe (helixRdc1Vec, helixRdc2Vec, saupe, rotArray, ramaFilter, 
					      phiAve, psiAve, refineCycle, dfsCycle, debugDFS, printResults);
	double Syy1 = saupe[0];
	double Szz1 = saupe[1];
	double rdc1Rmsd = saupe[2];
	double Syy2 = saupe[3];
	double Szz2 = saupe[4];
	double rdc2Rmsd = saupe[5];
	System.out.println(" Saupe elements and RMSDs in RDCs");
  	System.out.println(Syy1+",  "+Szz1+"; "+Syy2+",  "+Szz2+";  "+rdc1Rmsd+",  "+rdc2Rmsd);

 	Matrix rot = new Matrix(rotArray);
	double [] eigenValues = {Syy1, Szz1, Syy2, Szz2};
	//For betaSheet
	boolean printEulerSearch = false;
	PdbRmsd psd = new PdbRmsd();
	double[] noeRmsd = new double[1];
	boolean debugNOE = false;	
	Vector ubqByRdcVec = new Vector();
	Vector sheetVec = new Vector();
	System.out.println();
	System.out.println();
	if (paraVec.size() > 1){
	    sheetVec = mod.sheetModel(eRdc1Vec, eRdc2Vec, eigenValues, rot, HbondVec, 
				      paraVec, orderVec, printEulerSearch);
	    ubqByRdcVec = psd.positionByNOE(noeVec, helixPdbVec, sheetVec, noeRmsd, debugNOE);
	    System.out.println();
	    System.out.println();
	    System.out.println("PDB coordinates computed from RDCs");
	    System.out.println();
	    pp.print(ubqByRdcVec);
 	    System.out.println("TER");
 	    System.out.println("END");
 	}
	long endTime = System.currentTimeMillis() ;
	double totalTime = (double) ((endTime - startTime) / 60000.0); //in minutes
	System.out.println("Time= "+ totalTime +" minutes");
    }
}
