/*****************************************************************************************
 * ModelRdc.java

 * Version:           0.1

 * Description: 

 * A class implementing methods for building protein backbone models
 * from known backbone angles and lengths, and for computing Saupe
 * elements, and for computing individual secondary elements
 * (\alpha-helix and \beta-strands) by minimizing a target function
 * consisting of both the RDC term and the \phi and \psi angle terms,
 * and in the case of paring two \beta-strands, a h-bond term between
 * them.

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.FieldPosition;

public class ModelRdc implements Cloneable{
    private int residueNo;  //Though Phi and Psi are between two residues 
    private String residue; //We count it as belonging to the first of the two
    private double phi;
    private double psi;

    public ModelRdc(){
	residueNo = 0;
	residue = null;
        phi = 0.0;
        psi = 0.0;
    }	
    public ModelRdc(int No){
	residueNo = No;
	residue = ""; 
	phi = 0.0;
        psi = 0.0;
    }
    public ModelRdc(int no, String re, double x, double y){
	residueNo = no;
	residue = re;
        phi = x;
        psi = y;
    }
    //getting the values	
    public int getResidueNo(){
	return residueNo;
    }	
    public String getResidue(){
	return residue;
    }	
    public double getPhi(){
	return phi;
    }
    public double getPsi(){
	return psi;
    }
    public String toString(){
  	String desc = String.valueOf(residueNo) + "  " + residue
  	    +"  " + String.valueOf(phi*180/Math.PI) + "  " 
	    + String.valueOf(psi*180.0/Math.PI);
	return desc;	
    }

    public static class mdComparator implements Comparator{
	public int compare(Object o1, Object o2){
	    ModelRdc n1 = (ModelRdc)o1;
	    ModelRdc n2 = (ModelRdc)o2;
	    int d1 = n1.getResidueNo();
	    int d2 = n2.getResidueNo();
	    if (d1 < d2)
		return -1;
	    else if (d1 > d2)
		return 1;
	    else return 0;
	}
    }

    /**
     *  compute the length of a  vector v1 
     */
    public double length(double[] v1){
	double v1Len = Math.sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]);
	return v1Len;
    }
    /** calculate the angles between two vectors v1 and v2 
     * The returned angle is in the [0,Pi] range
     */
    public double interAngle(double[] v1, double[] v2){
	double v1Len = Math.sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]);
	double v2Len = Math.sqrt(v2[0]*v2[0] + v2[1]*v2[1] + v2[2]*v2[2]);
	double c = (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2]) / (v1Len * v2Len);
	return Math.acos(c);  //acos [0, PI] so should be OK
    }

    /** cal the directioal cosines of the vector vec */
    public double[] dirCos(double [] vec) {
	double len = 0.0;
	double [] dirs = new double[vec.length];
	for (int i=0; i<vec.length; i++)
	    len += vec[i] * vec[i];
	for (int j=0; j<vec.length; j++)
	    dirs[j] = vec[j] / Math.sqrt(len);
	return dirs;
    }

    /**
     * Compute the internuclear vector between two atoms
     @param n1 the coordinate for atom 1
     @param n2 the coordinate for atom 2
     @return a vector from n1->n2
    */
    public double[] internuclearVec(double[] n1, double[] n2){
        return new double[]{n2[0]-n1[0], n2[1]-n1[1], n2[2]-n1[2]};
    }

    /**
     * adding two vectors
     @param n1 the coordinate (direction cosine) for vector n1
     @param n2 the coordinate (direction cosine) for vector n2
     @return a vector from n1+n2
     */
    public double[] addCoords(double[] n1, double[] n2){
        return new double[]{n2[0]+n1[0], n2[1]+n1[1], n2[2]+n1[2]};
    }

    /**
     *A method for computing coordinates of backbone atoms
     @param theta1  an angle defined as in Const.class. 
     @param theta3  an angle defined as in Const.class. 
     @param theta5  an angle defined as in Const.class. 
     @param theta6  an angle defined as in Const.class. 
     @param theta7  an angle defined as in Const.class. 
     @param theta8  an angle defined as in Const.class. 
     @param theta9  an angle defined as in Const.class. 
     @param phi     The backbone \phi angle
     @param psi     The backbone \psi angle
     @param n1      the direction cosine for amide nitrogen of the original residue
     @param nh1     the direction cosine for amide proton
     @param ca1     the direction cosine for Ca
     @param residueNo  the residue number
     @param resid1     the name of the residue
     @param first    whether the residue is the first residue of the fragment
     @return Pdb object consisting of atoms: NH, CA, CO, HA and O of residue (i).
     * Please note that the coordinates of N, NH and CA of residue (i+1) are returned by
     * overwriting the original array for them: This may be BUG prone.
     **/    
    public Pdb coordByResidue (double theta1, double theta9, double theta3, double theta5, double theta6, 
			       double theta7, double theta8, double phi, double psi, double[] n1, 
			       double[] nh1, double[] ca1, int residueNo, String resid1, boolean first){
	Matrix rg = new Matrix(3,3);
	Matrix rgInv = new Matrix(3,3);
  	Matrix r2yInv = rg.rotationMat(phi, "-y"); //transpose == inverse == rotate in opposite direction
	Matrix r4zInv = rg.rotationMat(psi+Math.PI, "-z"); 
	double [] nToCO1Vec = new double[3];
	double [] coToCAVec = new double[3];
	double [] caToNVec  = new double[3];
	Vector atomVec    = new Vector();
	//intraResidue N(i)->NH(i) of the 1st residue
	double [] nToNHVec =  internuclearVec(n1,  nh1);
	//N(i)->CA(i) vector
	double [] nToCAVec = internuclearVec(n1, ca1);
	PhiPsi ff = new PhiPsi();
	if (first)
	    rg = Matrix.identity(3,3);
	else 
	    rg = ff.RgCal(nToNHVec, nToCAVec);
	rgInv = rg.transpose();

	//Compute the HA coordinate of residue (i)
 	Matrix matT = rgInv.times(Const.r1x9yInv.times(r2yInv));
	double [] coordHA = new double[]{0.0,  0.0, Const.dCA2HA};
	double[] caToHAVec =  matT.times(Const.rHA2HA1Inv.times(coordHA));
	coordHA =  addCoords(caToHAVec,  ca1);
	atomVec.add(new Cartesian("HA", coordHA[0], coordHA[1], coordHA[2]));

	//Compute the CO coordinate residue (i)
	matT = matT.times(Const.r3xInv);
	double [] coordCO   =  new double[]{0.0, 0.0, Const.dCA2CO};
	double [] caToCOVec = matT.times(coordCO);
	coordCO   = addCoords(caToCOVec,  ca1); 
	atomVec.add(new Cartesian("C", coordCO[0], coordCO[1], coordCO[2]));

	//Compute the O coordinate residue (i)
	double[] coordO = new double[]{0.0, 0.0, Const.dCO2O};
	double[] coToOVec = matT.times(r4zInv.times(Const.rOInv.times(coordO)));
	coordO  = addCoords(coToOVec,  coordCO);
	atomVec.add(new Cartesian("O", coordO[0], coordO[1], coordO[2]));

	//Compute the N coordinate residue (i+1)
	double [] coordN  = new double[]{0.0, Const.dCO2N, 0.0};
	matT = matT.times(r4zInv.times(Const.r5xInv));
	double [] co1ToNVec = matT.times(coordN);
	coordN = addCoords(coordCO, co1ToNVec);
	//updated N coordinate
	n1[0] = coordN[0];
	n1[1] = coordN[1];
	n1[2] = coordN[2];

	//Compute the CA coordinate residue (i+1)
	double [] coordCA = new double[]{0.0, Const.dN2CA, 0.0};
  	matT = matT.times(Const.r6yInv.times(Const.r7xInv));
	nToCAVec = matT.times(Const.r8zInv.times(Const.r1xInv.times(coordCA)));
	coordCA = addCoords(coordN, nToCAVec);
	//updated CA coordinate
	ca1[0] = coordCA[0];
	ca1[1] = coordCA[1];
	ca1[2] = coordCA[2];

	//Compute the NH coordinate residue (i+1)
	double [] coordNH = new double[]{0.0, 0.0,  -Const.dN2H};
	double [] nToNHVec2 = matT.times(coordNH);
	coordNH = addCoords(coordN, nToNHVec2);
	//updated NH coordinate
	nh1[0] = coordNH[0];
	nh1[1] = coordNH[1];
	nh1[2] = coordNH[2];
	return (new Pdb(residueNo, resid1, atomVec));
    }

    /*
     * Build model (compute the coordinates of backbone atoms (N, NH, CA, CO, O and HA) from 
     * the standard geometry and Phi/Psi angle.
     @param phiPsiVec  an array of Phi/Psi angles for the fragment.
     @param n1    the direction cosine for amide nitrogen of the first residue
     @param nh1   the direction cosine for amide proton
     @param ca1   the direction cosine for Ca
     @return a vector of Pdb object
     */
    public Vector modelBuild(Vector phiPsiVec, double [] n1, double[] nh1, double[] ca1){
	Vector tmpVec = new Vector();
	int i = 0, j = 0;
	double [] coordN  = new double[3]; 
	System.arraycopy(n1, 0, coordN, 0, 3);
	double [] coordNH = new double[3];
	System.arraycopy(nh1, 0, coordNH, 0, 3);
	double [] coordCA = new double[3]; 
	System.arraycopy(ca1, 0, coordCA, 0, 3);

	double [] nhVec1 = new double[3];
	double [] n2CA = new double[3];
	Vector atomVec = new Vector();
	Vector pdbVec = new Vector();
	Cartesian cc = new Cartesian();
 	atomVec.add(new Cartesian("N", n1));
	atomVec.add(new Cartesian("H", nh1));
 	atomVec.add(new Cartesian("CA", ca1));
	nhVec1 = internuclearVec(n1, nh1);
	Matrix rg = Matrix.identity(3,3);
	PhiPsi ff = (PhiPsi)phiPsiVec.elementAt(0);
	int no = ff.getResidueNo();
	String	resid1 = ff.getResidue();
 	double phi = ff.getPhi(); 
 	double psi = ff.getPsi(); 
	//The first residue
	Pdb pp = coordByResidue (Const.theta1, Const.theta9, Const.theta3, Const.theta5, Const.theta6, 
				 Const.theta7, Const.theta8, phi,  psi, coordN, coordNH, coordCA,
				 no, resid1, true);
	atomVec.addAll(pp.getAtomVec());
	pdbVec.add(new Pdb(no, resid1, atomVec));
	atomVec = new Vector();
	atomVec.add(new Cartesian("N", coordN));
	atomVec.add(new Cartesian("H", coordNH));
	atomVec.add(new Cartesian("CA", coordCA));
	//Compute New NH, N2CA and 
	nhVec1 = internuclearVec(coordN, coordNH); //update new NH1
	n2CA =   internuclearVec(coordN, coordCA);
	rg = ff.RgCal(nhVec1, n2CA);
	double[] co1ToN = new double[3];
	double[] co1    = new double[3];
	String atom = "";
	for (i=1; i<phiPsiVec.size(); i++){
	    ff = (PhiPsi)phiPsiVec.elementAt(i);
	    no = ff.getResidueNo();
	    resid1 = ff.getResidue();
   	    phi = ff.getPhi(); 
   	    psi = ff.getPsi(); 
	    pp = coordByResidue (Const.theta1, Const.theta9, Const.theta3, Const.theta5, Const.theta6, 
				 Const.theta7, Const.theta8, phi,  psi, coordN, coordNH, coordCA,
				 no, resid1, false);
	    atomVec.addAll(pp.getAtomVec());
	    for (j=0; j<atomVec.size(); j++){//We need to get CO coordinates BACK in case O and HA are there 
		cc = (Cartesian)atomVec.elementAt(j);
		atom = cc.getAtom();
		if (atom.equals("C"))
		    co1 = cc.getXYZ();
	    }
	    pdbVec.add(new Pdb(no, resid1, atomVec));
	    atomVec = new Vector();
	    atomVec.add(new Cartesian("N", coordN));
	    atomVec.add(new Cartesian("H", coordNH));
	    atomVec.add(new Cartesian("CA", coordCA));
	    //Update NEW NH, N2CA and CO1ToN vectors
 	    nhVec1 = internuclearVec(coordN, coordNH); //update new NH1
	    co1ToN = internuclearVec(coordN, co1);
	    n2CA   = internuclearVec(coordN, coordCA);
	    if (i == phiPsiVec.size() - 1)
		pdbVec.add(new Pdb(no+1, resid1, atomVec));
	    rg = ff.RgCal(nhVec1, n2CA);
	}
	return pdbVec;
    }

    /*
     * Build model (compute the coordinates of backbone atoms (N, NH, CA, CO, O and HA) from 
     * the standard geometry and Phi/Psi angle.
     @param phiArr an array of Phi angles for the fragment.
     @param psiArr an array of Psi angles for the fragment.
     @param n1  the direction cosine for amide nitrogen of the first residue
     @param nh1 the direction cosine for amide proton
     @param ca1 the direction cosine for Ca
     @param no1 the residue number of the first residue of a fragment
     @param first  whether the residue is the first residue of the fragment
     @return a vector of Pdb object
     */
    public Vector modelBuild(double[] phiArr, double[] psiArr, double[] n1, double[] nh1, double[] ca1, 
			     int no1, boolean first){
	Vector tmpVec = new Vector();
	int i = 0, j = 0;
	double [] coordN  = new double[3]; 
	System.arraycopy(n1, 0, coordN, 0, 3);
	double [] coordNH = new double[3];
	System.arraycopy(nh1, 0, coordNH, 0, 3);
	double [] coordCA = new double[3]; 
	System.arraycopy(ca1, 0, coordCA, 0, 3);
	int no = no1; 
	String resid ="ALA";
	PhiPsi ff = new PhiPsi();
	double [] n2CA = new double[3];
	Vector atomVec = new Vector();
	Vector pdbVec = new Vector();
	Cartesian cc = new Cartesian();
 	atomVec.add(new Cartesian("N", n1));
	atomVec.add(new Cartesian("H", nh1));
 	atomVec.add(new Cartesian("CA", ca1));
	double [] nhVec1 = internuclearVec(n1, nh1);
	Matrix rg = Matrix.identity(3,3);
	//The first residue
	Pdb pp = coordByResidue (Const.theta1, Const.theta9, Const.theta3, Const.theta5, Const.theta6, 
				 Const.theta7, Const.theta8, phiArr[0], psiArr[0], 
				 coordN, coordNH, coordCA, no, resid, first);
	atomVec.addAll(pp.getAtomVec());
	pdbVec.add(new Pdb(no, resid, atomVec));
	atomVec = new Vector();
	atomVec.add(new Cartesian("N",  coordN));
	atomVec.add(new Cartesian("H",  coordNH));
	atomVec.add(new Cartesian("CA", coordCA));
	//Compute New NH, N2CA and 
	nhVec1 = internuclearVec(coordN, coordNH); //update new NH1
	n2CA =   internuclearVec(coordN, coordCA);
	rg = ff.RgCal(nhVec1, n2CA);
	double[] co1ToN = new double[3];
	double[] co1    = new double[3];
	String atom = "";
	for (i=1; i<phiArr.length; i++){
	    no++;
	    pp = coordByResidue (Const.theta1, Const.theta9, Const.theta3, Const.theta5, Const.theta6, 
				 Const.theta7, Const.theta8, phiArr[i], psiArr[i], coordN, coordNH, coordCA,
				 no, resid, false);
	    atomVec.addAll(pp.getAtomVec());
	    for (j=0; j<atomVec.size(); j++){ //get CO coordinates BACK in case O and HA are there 
		cc = (Cartesian)atomVec.elementAt(j);
		atom = cc.getAtom();
		if (atom.equals("C"))
		    co1 = cc.getXYZ();
	    }
	    pdbVec.add(new Pdb(no, resid, atomVec));
	    atomVec = new Vector();
	    atomVec.add(new Cartesian("N",  coordN));
	    atomVec.add(new Cartesian("H",  coordNH));
	    atomVec.add(new Cartesian("CA", coordCA));
	    //Update NEW NH, N2CA and CO1ToN vectors
 	    nhVec1 = internuclearVec(coordN, coordNH);
	    co1ToN = internuclearVec(coordN, co1);
	    n2CA   = internuclearVec(coordN, coordCA);
	    if (i == phiArr.length - 1)
		pdbVec.add(new Pdb(no+1, resid, atomVec));
	    rg = ff.RgCal(nhVec1, n2CA);
	}
	return pdbVec;
    }

    /*       
     * Compute a helix model or the first strand (without H-bond term in the target function)
     @param  rdcVec1 the NH RDCs for medium 1
     @param  rdcVec2 the NH RDCs for medium 2
     @param  n2ca the N->CA vector of the first residue
     @param  Syy1 the diagnolized Saupe element for medium 1
     @param  Szz1 the diagnolized Saupe element for medium 1
     @param  rdc1Rmsd the RMSD between the measured RDCs and the back-computed RDCs 
                      before the refinement for medium 1
     @param  Syy2 the diagnolized Saupe element for medium 2
     @param  Szz2 the diagnolized Saupe element for medium 2
     @param  rdc2Rmsd the RMSD between the measured RDCs and the back-computed RDCs 
                      before the refinement for medium 2
     @param  rot the relative matrix between the two POFs
     @param  nCycle the number of DFS-based cycles
     @param  ramaFilter the Ramachandran filter for the favorable phi and psi region
     @param  phiAve  the average phi angle
     @param  psiAve  the average psi angle
     @param  weightAngles the relative weight for phi and psi angle terms in score function.
     @param  debugDFS  for monitoring the progress in systematic search-based minimization
     @param  printResults for printing the results for each iteration
     @return the refined Pdb vector
     */
    public Vector minHelix(final Vector rdcVec1, final Vector rdcVec2, final double[] n2ca, double Syy1,
			   double Szz1, double rdc1Rmsd, double Syy2, double Szz2, double rdc2Rmsd,
			   Matrix rot, int nCycle, double[] ramaFilter, double phiAve, double psiAve,
			   double weightAngles, boolean debugDFS, boolean printResults){
	Matrix rg = new Matrix(3,3);
	Matrix rgInv = new Matrix(3,3);
	int i = 0, j = 0, k = 0;  //i begin at no.18 for the helix : the real residueNo is 25
	int no = 0;
	int firstResidueNo = ((Dipolar)rdcVec1.elementAt(0)).getResidueNo();
	final int N = rdcVec1.size(); //the number of residues in the helix
	final int N1 = N - 1; //the number of residues in the helix
	double [] rdcV1 = new double[N]; //use array for efficience
	double [] rdcV2 = new double[N];
	double [] rdcV1Exp = new double[N];
	double [] rdcV2Exp = new double[N];
	double [] rdc1S = new double[N];
	double [] rdc2S = new double[N];
	int [] seqNos = new int[N];
	double [] coefs  = new double[14]; //the number 14 is hard-coded in the quarticCoef
	Root[] allRoots = new Root[N];
	int[] pathIndex = new int[N];
	for (j = 0; j < N; j++){
	    rdcV1Exp[j] = ((Dipolar)rdcVec1.elementAt(j)).getRdc();
	    seqNos[j] = ((Dipolar)rdcVec1.elementAt(j)).getResidueNo();
	    rdcV2Exp[j] = ((Dipolar)rdcVec2.elementAt(j)).getRdc();
	    pathIndex[j] = -1;
	}

    	PhiPsi ff = new PhiPsi();
	Vector rootVec = new Vector();
	Vector rootVec1 = new Vector();
	double[] nh1 = new  double[3];
	double[] nh1S = new  double[3];
	final double rRdcFinal1 = rdc1Rmsd * rdc1Rmsd * N; //Assuming independence
	final double rRdcFinal2 = rdc2Rmsd * rdc2Rmsd * N; // 
	Vector ppVec = new Vector();
	double rRdc1 = 1000.0, rRdc2 = 1000.0, rTotal = 10000.0, rT = 0.0;
	int depth = 0;
	double[] phiS = new double[N1];
	double[] psiS = new double[N1];
	double[] phiSave = new double[N1];
	double[] psiSave = new double[N1];
	double phiRmsd = 0.0, psiRmsd = 0.0;
	double rms1, rms2;
	boolean flag1 = false;
	boolean flagSln = false;
	long seed = -394715637; //using one seed in order to repeat the results
	Random rr = new Random(seed);

	if (debugDFS)
	    System.out.println("CycleNo,  rmsdRDC1,   rmsdRDC2,  rmsdPhi,    rmsdPsi,    totalScore");
	for (long mmm = 0; mmm < nCycle; mmm++){
    	    stop1:
  	    while (!flag1){
		flag1 = true;
		rRdc1 = 0.0;
		rRdc2 = 0.0;
		for (j = 0; j < N; j++){ //For the 1st beta strand of the pair
		    no = seqNos[j];
		    rdcV1[j] = rdcV1Exp[j] + rdc1Rmsd * rr.nextGaussian();
		    rdcV2[j] = rdcV2Exp[j] + rdc2Rmsd * rr.nextGaussian();
		    rms1 = Math.abs(rdcV1[j] - rdcV1Exp[j]);
		    rms2 = Math.abs(rdcV2[j] - rdcV2Exp[j]);
		    flagSln = ff.quarticCoef(Syy1, Szz1, Syy2, Szz2, rdcV1[j], rdcV2[j], rot, coefs);
		    if (!flagSln){
			flag1 = false; 
   			continue stop1;
		    }
		    rootVec = ff.quarticSolve(coefs);
		    if (rootVec.size() == 0 || rms1 > 2.0*rdc1Rmsd || rms2 > 2.0 * rdc2Rmsd){  
			flag1 = false; 
   			continue stop1;
		    }
		    allRoots[j] = new Root(no, rootVec);
		    rRdc1 += rms1 * rms1;
		    rRdc2 += rms2 * rms2;
		}
   		if ( (rRdc1 > rRdcFinal1) || (rRdc2 > rRdcFinal2) )
   	    	    flag1 = false;
  	    }
	    if(flag1){
		rootVec1 = allRoots[0].getRootVec();  
		for (i = 0; i < rootVec1.size(); i++){
		    pathIndex[depth] = i;
		    nh1 = (double[])(rootVec1.elementAt(i));
		    rg = ff.RgCal(nh1, n2ca);
		    ppVec = new Vector();
		    depth = 0;
		    ff.phiPsiChain(n2ca, nh1, allRoots, depth+1, ppVec, phiS, psiS, pathIndex, 0, ramaFilter);
		    //ppVec saves all the solutions
		    if (ppVec.size() > 0){//select one with smallest RMSD to the average \phipsi for the helix
			int noOfSln = (int)(ppVec.size() / (2*N1)); //the total # of solutions for the 2 sets
			for(k=0; k < noOfSln; k++){   //please note the index
			    phiRmsd = 0.0;
			    psiRmsd = 0.0;
			    for (int n = k* 2 * N1; n < (k+1)* 2 *N1; n += 2){  //original "N"
				phi = ((Double)ppVec.elementAt(n)).doubleValue();
				psi = ((Double)ppVec.elementAt(n+1)).doubleValue();
				phiS[(int)(n - k* 2 * N1) / 2] = phi;
				psiS[(int)(n - k* 2 * N1) / 2] = psi;
				phiRmsd += (phi - phiAve) * (phi - phiAve);
				psiRmsd += (psi - psiAve) * (psi - psiAve);
			    }
			    rT = Math.sqrt(rRdc1 / N)+ Math.sqrt(rRdc2 / N) 
				+ weightAngles * (Math.sqrt(phiRmsd / N1) + Math.sqrt(psiRmsd / N1));
			    if (rT < rTotal){
				if (debugDFS)
				    System.out.println( mmm +": "+Math.sqrt(rRdc1 / N )+"  "+ Math.sqrt(rRdc2 / N )
							+";  "+(Math.sqrt(phiRmsd /N1 ) / Const.cst)+"  "
							+(Math.sqrt(psiRmsd /N1) / Const.cst)+", RTotal = "+rTotal);
				rTotal = rT;
				System.arraycopy(phiS, 0, phiSave, 0, N1);
				System.arraycopy(psiS, 0, psiSave, 0, N1);
				System.arraycopy(rdcV1, 0, rdc1S, 0, N);
				System.arraycopy(rdcV2, 0, rdc2S, 0, N);
				System.arraycopy(nh1, 0, nh1S, 0, 3);
			    }
			}
		    }
		}
	    }
	    flag1 = false;
	}
	//The refined Model
	double[] n1 = new double[] {0.0, 0.0, 0.0};
	Vector pdbHelix = modelBuild(phiSave, psiSave, n1, nh1S, n2ca, firstResidueNo, false);
	if (printResults){
	    System.out.println();
	    System.out.println(" RDCs: ResidueNO  Sampled RDC   Experimental RDC  Difference ");
	    for (k=0; k < N; k++)
		System.out.println((firstResidueNo+k)+"  "+rdc1S[k]+"  "+rdcV1Exp[k]+"  "+Math.abs(rdc1S[k]-rdcV1Exp[k]));
	    System.out.println();
	    for (k=0; k < N; k++)
		System.out.println((firstResidueNo+k)+"  "+rdc2S[k]+"  "+rdcV2Exp[k]+"  "+Math.abs(rdc2S[k]-rdcV2Exp[k]));
	    System.out.println();
	    System.out.println(" Phi/Psi: ResidueNO    phi angle    psi angle");
	    for (k=0; k < N1; k++)
		System.out.println((firstResidueNo+k)+"  "+(phiSave[k]/Const.cst)+"  "+(psiSave[k]/Const.cst));
	}
	return pdbHelix;
    }

  /*       
     * Compute a strand model which is paired with a known strand (with the H-bond term in the target function)
     @param  rdcVec1 the NH RDCs for medium 1
     @param  rdcVec2 the NH RDCs for medium 2
     @param  n2ca the N->CA vector of the first residue
     @param  Syy1 the diagnolized Saupe element for medium 1
     @param  Szz1 the diagnolized Saupe element for medium 1
     @param  rdc1Rmsd the RMSD for medium 1
     @param  Syy2 the diagnolized Saupe element for medium 2
     @param  Szz2 the diagnolized Saupe element for medium 2
     @param  rdc2Rmsd the RMSD for medium 2
     @param  rot   the relative matrix between the two POFs
     @param  nCycle the number of DFS-based cycles
     @param  ramaFilter the Ramachandran filter for the favorable phi and psi region
     @param  phiAve  the average phi angle
     @param  psiAve  the average psi angle
     @param  weightAngles the relative weight for phi and psi angle terms in score function.
     @param  hbWeight the relative weight for H-bond terms in score function.
     @param  pdbS1 the strand which is the partner of the strand to be refined.
     @param  hbVecOfE12 Hbonds bewteen the two strands. 
     @param  debugDFS  for monitoring the progress in systematic search-based minimization
     @param  printResults for printing the results for each iteration
     @return the refined Pdb vector
     */
    public Vector minStrand(final Vector rdcVec1, final Vector rdcVec2, final double[] n2ca, double Syy1, double Szz1,
			    double rdc1Rmsd, double Syy2, double Szz2, double rdc2Rmsd, Matrix rot, int nCycle, 
			    double[] ramaFilter, double phiAve, double psiAve, final Vector pdbS1, Vector hbVecOfE12,
			    double angleWeight, double hbWeight, boolean debugDFS, boolean printResults){
	Matrix rg = new Matrix(3,3);
	Matrix rgInv = new Matrix(3,3);
	int i = 0, j = 0, k = 0;  //i begin at no.18 for the helix : the real residueNo is 25
	int no = 0;
	int firstResidueNo = ((Dipolar)rdcVec1.elementAt(0)).getResidueNo();
	final int N = rdcVec1.size(); //the number of residues in the helix
	final int N1 = N - 1; //the number of residues in the helix
	double [] rdcV1 = new double[N]; //use array for efficience
	double [] rdcV2 = new double[N];
	double [] rdcV1Exp = new double[N];
	double [] rdcV2Exp = new double[N];
	double [] rdc1S = new double[N];
	double [] rdc2S = new double[N];
	int [] seqNos = new int[N];
	double [] coefs  = new double[14]; //the number 14 is hard-coded in the quarticCoef
	Root[] allRoots = new Root[N];
	int[] pathIndex = new int[N];
	for (j = 0; j < N; j++){
	    rdcV1Exp[j] = ((Dipolar)rdcVec1.elementAt(j)).getRdc();
	    seqNos[j] = ((Dipolar)rdcVec1.elementAt(j)).getResidueNo();
	    rdcV2Exp[j] = ((Dipolar)rdcVec2.elementAt(j)).getRdc();
	    pathIndex[j] = -1;
	}

    	PhiPsi ff = new PhiPsi();
	Vector rootVec = new Vector();
	Vector rootVec1 = new Vector();
	Vector pdbVec2 = new Vector();
	double[] nh1 = new  double[3];
	double[] nh1S = new  double[3];
	final double rRdcFinal1 = rdc1Rmsd * rdc1Rmsd * N; // 
	final double rRdcFinal2 = rdc2Rmsd * rdc2Rmsd * N; // 
	Vector ppVec = new Vector();
	double rRdc1 = 1000.0, rRdc2 = 1000.0, rTotal = 10000.0, rT = 0.0;
	int depth = 0;
	double [] phiS = new double [N1];
	double [] psiS = new double [N1];
	double [] phiSave = new double [N1];
	double [] psiSave = new double [N1];
	Vector pdbVecS = new Vector();
	double phiRmsd = 0.0, psiRmsd = 0.0;
	double rms1, rms2;
	boolean flag1 = false;
	boolean flagSln = false;
	long seed = -394715637; //using one seed in order to repeat the results
	Random rr = new Random(seed);
	double matchScore = 0.0;
	PdbRmsd pR = new PdbRmsd();
	double[] amide = new double[] {0.0, 0.0, 0.0};
	for (long mmm = 0; mmm < nCycle; mmm++){
    	    stop1:
  	    while (!flag1){
		flag1 = true;
		rRdc1 = 0.0;
		rRdc2 = 0.0;
		for (j = 0; j < N; j++){   //For the 1st beta strand of the pair
		    no = seqNos[j];
		    rdcV1[j] = rdcV1Exp[j] + rdc1Rmsd * rr.nextGaussian();
		    rdcV2[j] = rdcV2Exp[j] + rdc2Rmsd * rr.nextGaussian();
		    rms1 = Math.abs(rdcV1[j] - rdcV1Exp[j]);
		    rms2 = Math.abs(rdcV2[j] - rdcV2Exp[j]);
		    flagSln = ff.quarticCoef(Syy1, Szz1, Syy2, Szz2, rdcV1[j], rdcV2[j], rot, coefs);
		    if (!flagSln){
			flag1 = false; 
   			continue stop1;
		    }
		    rootVec = ff.quarticSolve(coefs);
		    if (rootVec.size() == 0 || rms1 > 2.0*rdc1Rmsd || rms2 > 2.0 * rdc2Rmsd){  
			flag1 = false; 
   			continue stop1;
		    }
		    allRoots[j] = new Root(no, rootVec);
		    rRdc1 += rms1 * rms1;
		    rRdc2 += rms2 * rms2;
		}
   		if ( (rRdc1 > rRdcFinal1) || (rRdc2 > rRdcFinal2) )
   	    	    flag1 = false;
  	    }
	    if(flag1){
		rootVec1 = allRoots[0].getRootVec();  
		for (i = 0; i < rootVec1.size(); i++){
		    pathIndex[depth] = i;
		    nh1 = (double[])(rootVec1.elementAt(i));
		    rg = ff.RgCal(nh1, n2ca);
		    ppVec = new Vector();
		    depth = 0;
		    ff.phiPsiChain(n2ca, nh1, allRoots, depth+1, ppVec, phiS, psiS, pathIndex, 0, ramaFilter);
		    //ppVec save all the solutions
		    if (ppVec.size() > 0){ //select one with smallest RMSD to the average \phipsi for the helix
			int noOfSln = (int)(ppVec.size() / (2*N1)); //the total # of solutions for the 2 sets
			for(k=0; k < noOfSln; k++){   //please note the index
			    phiRmsd = 0.0;
			    psiRmsd = 0.0;
			    for (int n = k* 2 * N1; n < (k+1)* 2 *N1; n += 2){  //original "N"
				phi = ((Double)ppVec.elementAt(n)).doubleValue();
				psi = ((Double)ppVec.elementAt(n+1)).doubleValue();
				phiS[(int)(n - k* 2 * N1) / 2] = phi;
				psiS[(int)(n - k* 2 * N1) / 2] = psi;
				phiRmsd += (phi - phiAve) * (phi - phiAve);
				psiRmsd += (psi - psiAve) * (psi - psiAve);
			    }
 			    pdbVec2 = modelBuild(phiS, psiS, amide, nh1, n2ca, firstResidueNo, false);
 			    matchScore = pR.centerFit(pdbS1, pdbVec2, hbVecOfE12, false);
			    rT = Math.sqrt(rRdc1/N) + Math.sqrt(rRdc2/N) 
				+ angleWeight*(Math.sqrt(phiRmsd/N1) + Math.sqrt(psiRmsd/N1)) + hbWeight*matchScore;
			    if (rT < rTotal){
				if(debugDFS)
				    System.out.println( mmm +": "+Math.sqrt(rRdc1 / N )+"  "+ Math.sqrt(rRdc2 / N )
							+";  "+(Math.sqrt(phiRmsd /N1 ) / Const.cst)+"  "
							+(Math.sqrt(psiRmsd /N1) /Const.cst)+"  "+matchScore+", RTotal = "+rTotal);
				rTotal = rT;
				System.arraycopy(phiS, 0, phiSave, 0, N1);
				System.arraycopy(psiS, 0, psiSave, 0, N1);
				System.arraycopy(rdcV1, 0, rdc1S, 0, N);
				System.arraycopy(rdcV2, 0, rdc2S, 0, N);
				pdbVecS = pdbVec2;
			    }
			}
		    }
		}
	    }
	    flag1 = false;
	}
	if(printResults){
	    System.out.println(" RDC: ");
	    for (k=0; k < N; k++)
		System.out.println((firstResidueNo+k)+": "+rdc1S[k] +"  "+rdcV1Exp[k]+"  "+Math.abs(rdc1S[k] - rdcV1Exp[k]));
	    System.out.println();
	    for (k=0; k < N; k++)
		System.out.println((firstResidueNo+k)+": "+rdc2S[k] +"  "+rdcV2Exp[k]+"  "+ Math.abs(rdc2S[k] - rdcV2Exp[k]));

	    System.out.println(" Phi/Psi: ");
	    for (k=0; k < N1; k++)
		System.out.println((firstResidueNo+k)+"   "+(phiSave[k]/Const.cst)+"  "+(psiSave[k]/Const.cst));
	}
	return pdbVecS;
    }

    /**
     * Refine Saupe elements from an initial model
     @param rdc1Vec RDCs in medium 1
     @param rdc2Vec RDCs in medium 2
     @param eigenValues for returning the diagnolized Saupe elements and RMSDs in RDCs in
                  the order of {Syy1, Szz1, rmsd1, Syy2, Szz2, rmsd2}
     @param refineCycle the number of cycles for iterating through the refinement 
                         and computation of Saupe matrices. 
     @param nCycle  the number of cycles for the refinement by DFS itself.
     @param ramaFilter the Ramachandran filter for the favorable phi and psi region
     @param phiAve  the average phi angle
     @param psiAve  the average psi angle
     @param debugDFS  for monitoring the progress in systematic search-based minimization
     @param printResults for printing the results for each iteration
     @param matR12  for returning the relative matrix between the two POFs of RDCs
     @return  the refined Pdb vector
     */
    public Vector refineSaupe (final Vector rdc1Vec, final Vector rdc2Vec, double[] eigenValues, double[][] matR12,
			       double[] ramaFilter, double phiAve, double psiAve, int refineCycle,
			       int nCycle, boolean debugDFS, boolean printResults){
	int i = 0;
    	double [] n1 = {0.0, 0.0, 0.0};         //A coordinate frame in the 1st peptide plane
	double [] nh1 = {0.0, 0.0, -Const.dN2H}; 
    	double [] ca1 = {0.0, Const.dN2CA * Math.cos(Const.theta1), Const.dN2CA * Math.sin(Const.theta1)};

	//build an initial  helix model with average \phi and \psi angles.
	Pdb pp = new Pdb();
	Vector phiPsiVec = new Vector();
	Dipolar dd = (Dipolar)rdc1Vec.elementAt(0);
	int firstResidueNo =  dd.getResidueNo();
	int lastResidueNo  = firstResidueNo + rdc1Vec.size();
   	for (i=firstResidueNo; i<lastResidueNo; i++)  
    	    phiPsiVec.add(new PhiPsi(i, "ALA", phiAve, psiAve));
 	Vector pdbHelix = modelBuild(phiPsiVec, n1, nh1, ca1);
	//pp.print(pdbHelix);

	//Compute by SVD the 7 parameters used for computing other 2ary elements
 	PdbRdc pdr = new PdbRdc();
 	double[] saupe = new double[6];//Saving Saupe elements or rmsds in RDCs: Syy1,Szz1,rmsd1,Syy2,Szz2,rmsd2
	double[] n2ca  = new double[3];
	Matrix r = new Matrix(3, 3);
	double weight4Angles = 1.0;   //the relative weight for \phi and \psi angles vs. RDCs in target function
	for(i = 0; i< refineCycle; i++){
	    r = pdr.Tensors(pdbHelix, rdc1Vec, rdc2Vec, n2ca, saupe, printResults);
	    pdbHelix = minHelix(rdc1Vec, rdc2Vec, n2ca, saupe[0], saupe[1], saupe[2], saupe[3], saupe[4], saupe[5],
				r, nCycle, ramaFilter, phiAve, psiAve, weight4Angles, debugDFS, printResults);
	}
	r = pdr.Tensors(pdbHelix, rdc1Vec, rdc2Vec, n2ca, saupe, printResults);
	double[][] a1 = r.getArray();
	System.arraycopy(a1, 0, matR12, 0, 3);
	System.arraycopy(saupe, 0, eigenValues, 0, 6);
	Vector pdbVecN = pdr.rightHandedPdb(pdbHelix, rdc1Vec, rdc2Vec);
	return pdbVecN;
    }

    /**
     * For computing the first Strand which is NOT H-bonded to others.
     @param rdc1Vec RDCs in medium 1
     @param rdc2Vec RDCs in medium 2
     @param rot  the relative matrix between two POFs
     @param saupe four diagnolized Saupe elements and RMSDs in RDCs in
                  the order of {Syy1, Szz1, rmsd1, Syy2, Szz2, rmsd2}
     @param refineCycle the number of cycles for iterating through the refinement 
                        and computation of Saupe matrices. 
     @param initialCycle  the number of initial cycles for the DFS-based systematic search.
     @param ramaFilter the Ramachandran filter for the favorable phi and psi region
     @param phiAve  the average phi angle
     @param psiAve  the average psi angle
     @param w4Angles  weight for the \phi and \psi angles in the score funtion
     @param resolution the resolution for the grid-search for the first peptide plane
     @param debugDFS  for monitoring the progress in systematic search-based minimization
     @param printResults for printing the results for each iteration
     @return the refined Pdb vector
     */
    public Vector refineStrand (final Vector rdc1Vec, final Vector rdc2Vec, Matrix rot, double[] saupe, 
				double[] ramaFilter, double phiAve, double psiAve, int refineCycle, int initialCycle,
				double w4Angles, double resolution, boolean debugDFS, boolean printResults){
	int i = 0;
    	double [] n1  = {0.0, 0.0, 0.0};         //A coordinate frame in the 1st peptide plane
	double [] nh1 = {0.0, 0.0, -Const.dN2H}; 
    	double [] ca1 = {0.0, Const.dN2CA * Math.cos(Const.theta1), Const.dN2CA * Math.sin(Const.theta1)};

	//build an initial model
	Vector phiPsiVec = new Vector();
	Dipolar dd = (Dipolar)rdc1Vec.elementAt(0);
	Pdb pp = new Pdb();
	int firstResidueNo =  dd.getResidueNo();
	int lastResidueNo  = firstResidueNo + rdc1Vec.size();
   	for (i=firstResidueNo; i<lastResidueNo; i++)  
    	    phiPsiVec.add(new PhiPsi(i, "ALA", phiAve, psiAve));
 	Vector pdbStrand = modelBuild(phiPsiVec, n1, nh1, ca1);
//  	pp.print(pdbStrand);

	//Compute the orientation of the first peptide by grid-search
 	PdbRdc pdr = new PdbRdc();
	double [] rmsds = new double[2];
 	Vector pdbStrandN = new Vector();
	int N = rdc1Vec.size();
	Matrix iMat = Matrix.identity(3, 3);
  	double rdc1Rms = 0.0, rdc2Rms = 0.0; //S11-17 after search through only 180*180*180 for three angles.

	boolean debugEulerFit = false;
	Matrix rot4Strand = pdr.eulerFit(pdbStrand, rdc1Vec, rdc2Vec, saupe, rot, rmsds, resolution, debugEulerFit);
	rdc1Rms = Math.sqrt(rmsds[0] / N);
	rdc2Rms = Math.sqrt(rmsds[1] / N);
	if (printResults){
	    System.out.println("Results from grid-search: rmsdRdc1,  rmsdRdc2");
	    System.out.println(rdc1Rms+"  "+rdc2Rms);
	    rot4Strand.print(14,14);
	}
  	double [] n2ca = rot4Strand.times(ca1);
	int nCycle = 1; 
	for(i = 1; i< (refineCycle + 1); i++){
 	    nCycle = i*initialCycle;
	    pdbStrandN = minHelix(rdc1Vec,rdc2Vec,n2ca,saupe[0],saupe[1],rdc1Rms,saupe[2],saupe[3],
 				  rdc2Rms, rot, nCycle, ramaFilter, phiAve, psiAve, w4Angles, debugDFS, printResults);
	    rdc1Rms = pdr.BackCal(pdbStrandN,rdc1Vec,"N","H", iMat,-saupe[0]-saupe[1],saupe[0],saupe[1]);
	    rdc2Rms = pdr.BackCal(pdbStrandN,rdc2Vec,"N","H", rot.times(iMat),-saupe[2]-saupe[3],saupe[2],saupe[3]);
	}
 	return pdbStrandN;
    }

    /**
     *For computing the others Strands which IS H-bonded to others.
     @param rdc1Vec RDCs in medium 1
     @param rdc2Vec RDCs in medium 2
     @param nToCa   for returning the N->CA vector.
     @param saupe   four diagnolized Saupe elements in the order of {Syy1, Szz1, Syy2, Szz2}
     @param refineCycle  the number of cycles for iterating through the refinement 
                         and computation of Saupe matrices. 
     @param nCycle  the number of cycles for the refinement by DFS itself.
     @param  ramaFilter the Ramachandran filter for the favorable phi and psi region
     @param  phiAve  the average phi angle
     @param  psiAve  the average psi angle
     @param  angleWeights the relative weight for phi and psi angle terms in score function.
     @param  hbWeight the relative weight for H-bond terms in score function.
     @param  pdbS1 the strand which is the partner of the strand to be refined.
     @param  hbVecOfE12 Hbonds bewteen the two strands. 
     @param  debugDFS  for monitoring the progress in systematic search-based minimization
     @param  printResults for printing the results for each iteration
     @param  printEulerSearch if true it prints the progress in the grid-search for three euler angles.
     @return the refined Pdb vector
     */
    public Vector refineStrand(final Vector rdc1Vec, final Vector rdc2Vec, Matrix rot, double[] saupe,
			       double[] ramaFilter, double phiAve, double psiAve, int refineCycle, int initialCycle,
			       double resolution, double angleWeight, double hbWeight, Vector pdbVecS1, 
			       Vector hBVecOfE21, boolean printEulerSearch,  boolean debugDFS, boolean printResults){
	int i = 0;
    	double [] n1  = {0.0, 0.0, 0.0};         //A coordinate frame in the 1st peptide plane
	double [] nh1 = {0.0, 0.0, -Const.dN2H}; 
    	double [] ca1 = {0.0, Const.dN2CA * Math.cos(Const.theta1), Const.dN2CA * Math.sin(Const.theta1)};
	//build an initial ideal helix mode
	Vector phiPsiVec = new Vector();
	Dipolar dd = (Dipolar)rdc1Vec.elementAt(0);
	Pdb pp = new Pdb();
	int firstResidueNo =  dd.getResidueNo();
	int lastResidueNo  = firstResidueNo + rdc1Vec.size();
   	for (i=firstResidueNo; i<lastResidueNo; i++) {
	    phiPsiVec.add(new PhiPsi(i, "ALA", phiAve, psiAve));
	}
 	Vector pdbStrand = modelBuild(phiPsiVec, n1, nh1, ca1);
//   	pp.print(pdbStrand);

 	PdbRdc pdr = new PdbRdc();
 	PdbRmsd pR = new PdbRmsd();
	double [] rmsds = new double[3];
	Matrix r = new Matrix(3, 3);
	int N = rdc1Vec.size();
	Vector pdbStrandN = new Vector();
	Matrix iMat = Matrix.identity(3, 3);
	double[] hbScores = new double[4];
 	Matrix[] mm = new Matrix[4];
  	double[] rdc1Rmsd = new double[4];
  	double[] rdc2Rmsd = new double[4];
	int minInd4Rdc = 0;
	int minInd4HB = 0;
 	double w4HBEulerFit = 1.0;

 	mm[0] = pdr.eulerFitHB(pdbStrand, rdc1Vec, rdc2Vec, saupe, rot, rmsds, resolution, w4HBEulerFit, pdbVecS1, 
			       hBVecOfE21, printEulerSearch);
	double rdc1Rms = Math.sqrt(rmsds[0] / N);
	double rdc2Rms = Math.sqrt(rmsds[1] / N);
	if (printResults){
	    System.out.println("Results from grid-search: rmsdRdc1,  rmsdRdc2,  hbScore");
	    System.out.println(rdc1Rms+"   "+rdc2Rms+"   "+ rmsds[2]);
	}
	hbScores[0] = rmsds[2];
	for (i = 1; i < 4; i++){
	    mm[i] = Const.mat4ThreeDirs[i - 1].times(mm[0]);
	    pdbStrandN = pp.newPdb(pdbStrand, mm[i]);
	    hbScores[i] = pR.centerFit(pdbVecS1, pdbStrandN, hBVecOfE21, false);
	    rdc1Rmsd[i] = pdr.BackCal(pdbStrand,rdc1Vec,"N","H", mm[i], -saupe[0]-saupe[1], saupe[0], saupe[1]);
	    rdc2Rmsd[i] = pdr.BackCal(pdbStrand,rdc2Vec,"N","H",rot.times(mm[i]),-saupe[2]-saupe[3],saupe[2],saupe[3]);
	}
	for (i=1; i<rdc2Rmsd.length; i++){
	    if(rdc2Rmsd[i] < rdc2Rms){
		rdc2Rms = rdc2Rmsd[i];
		minInd4Rdc = i;
	    }
	}
	for (i=1; i<hbScores.length; i++){
	    if(hbScores[i] < hbScores[0])
		minInd4HB = i;
	}
	int nCycle = 1;
	double[] n2ca = new double[3];
 	if (minInd4HB != minInd4Rdc){
	    System.out.println("H-bond and RDCs not consistent for the element begin with residue No "  +  firstResidueNo);
// 	    System.out.println(minInd4HB+"  "+hbScores[minInd4HB]);
// 	    System.out.println(hbScores[0]+"   "+hbScores[1]+"   "+hbScores[2]+"   "+hbScores[3]);
// 	    System.out.println(minInd4Rdc+"  "+rdc1Rms+"   "+rdc2Rms);
// 	    System.out.println(rdc2Rmsd[0]+"   "+rdc2Rmsd[1]+"   "+rdc2Rmsd[2]+"   "+rdc2Rmsd[3]);
	}
	n2ca = mm[minInd4HB].times(ca1);
	for(i = 1; i< (refineCycle+1); i++){
	    nCycle = i* initialCycle;
	    pdbStrandN = minStrand(rdc1Vec, rdc2Vec, n2ca, saupe[0], saupe[1], rdc1Rms, saupe[2], saupe[3], rdc2Rms, rot, nCycle, 
				   ramaFilter, phiAve,  psiAve, pdbVecS1, hBVecOfE21, angleWeight, hbWeight, debugDFS, printResults);
	    rdc1Rms = pdr.BackCal(pdbStrandN, rdc1Vec,"N","H", iMat, -saupe[0]-saupe[1], saupe[0], saupe[1]);
	    rdc2Rms = pdr.BackCal(pdbStrandN, rdc2Vec,"N","H", rot,  -saupe[2]-saupe[3], saupe[2], saupe[3]);
	}
	Vector pdbVecN = pR.centerFitVec(pdbVecS1, pdbStrandN, hBVecOfE21, false);
  	return pdbVecN;
    }

    protected Object clone(){
        try {
	    Object s = super.clone();     
	    return s;                     
        } catch (CloneNotSupportedException e) {
	    throw new InternalError();
        }
    }
}
