/*****************************************************************************************
 * Noe.java

 * Version:           0.1

 * Description: 

 * Represent and read the NOE data

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;

public class Noe implements Cloneable{
    private int residueNoA;   //the residue  No of one of the two partners of the NOE
    private int residueNoB;   //the residue  No of Another partner of the NOE
    private String atomA;     //the name of Atom A
    private String atomB;     //the name of Atom B
    private double lower;     //the bound for NOEs
    private double upper;

    /** Constructors */
    public Noe()   {
    	residueNoA = 0;
	residueNoB = 0;
        atomA = null;
        atomB = null;
	lower = 0.0;
	upper = 0.0;
    }
    public Noe(int NoA, int noB) {
    	residueNoA = NoA;
	residueNoB = noB;
        atomA = null;
        atomB = null;
	lower = 0.0;
	upper = 0.0;
   }
   public Noe(int Ano, String A, int Bno,  String B,  double low, double up)   {
       	residueNoA = Ano;
	residueNoB = Bno;
        this.atomA = A;
        this.atomB = B;
	lower = low;
	upper = up;
    }

    public int getResidueNoA() {
	return residueNoA;
    }
    public int getResidueNoB() {
	return residueNoB;
    }
    public String getAtomA()   {
	return atomA;
    }
    public String getAtomB()   {
	return atomB;
    }

    public double[] getRange(){
	return new double[]{lower, upper};
    }

    public double getLower(){
	return lower;
    }

    public double getUpper(){
	return upper;
    }

    public String toString()	{
	String desc = residueNoA + atomA + "__"+residueNoB+atomB+"  "+lower+"---"+upper;
	return desc;
    }
    protected Object clone(){
      	try{
	    Object s = super.clone();     // clone the stack
	    return s;                           // return the clone
      	} catch (CloneNotSupportedException e)	{
	    throw new InternalError();
	}
    }

    public static class noeComparatorA implements Comparator{
	public int compare(Object o1, Object o2){
	    Noe n1 = (Noe)o1;
	    Noe n2 = (Noe)o2;
	    int d1 = n1.getResidueNoA();
	    int d2 = n2.getResidueNoA();
	    if (d1 < d2)
		return -1;
	    else if (d1 > d2)
		return 1;
	    else return 0;
	}
    }

    public static class noeComparatorB implements Comparator{
	public int compare(Object o1, Object o2){
	    Noe n1 = (Noe)o1;
	    Noe n2 = (Noe)o2;
	    int d1 = n1.getResidueNoB();
	    int d2 = n2.getResidueNoB();
	    if (d1 < d2)
		return -1;
	    else if (d1 > d2)
		return 1;
	    else return 0;
	}
    }

    /*
     * A very simple reader for NOE data, see the attached file for format.
     @param noeFile the name of the file
     @return a vector of NOE object
     */
    public Vector noeReader(String noeFile){
	Vector inputs = new Vector();
	int noA, noB;
	String atomA, atomB;
	double noeValue;
	try{
	    StreamTokenizer in =
		new StreamTokenizer(
				    new FileReader(noeFile));
	    while( in.nextToken() != StreamTokenizer.TT_EOF ){ //ignore the comments
		if (in.ttype == StreamTokenizer.TT_WORD)
		    ;
		else if (in.ttype == StreamTokenizer.TT_NUMBER){
		    noA = (int)in.nval;
  		    in.nextToken();         
		    atomA = in.sval;
		    in.nextToken();
		    noB = (int)in.nval;
		    in.nextToken();
		    atomB = in.sval;
		    in.nextToken();
 		    lower = in.nval;
		    in.nextToken();
 		    upper = in.nval;
		    inputs.add(new Noe(noA, atomA, noB, atomB, lower, upper));
		}	
	    }			
	}catch (FileNotFoundException e) {
	    System.out.println("File not found: " + noeFile);
	}catch (IOException e) {
	    System.out.println("IOException: the stack trace is:");
	    e.printStackTrace();
	}
	return inputs;
    }
}






















