/*****************************************************************************************
 * Pdb.java

 * Version:           0.1

 * Description: 

 * Represent, read and generate PDB files and extract various
 * backbone angles etc.

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.FieldPosition;

public class Pdb{
    private int residueNo;
    private String residue;
    private Vector atomVec = new Vector(); //all the atoms belong to the residue

    public Pdb(){
	residueNo = 0;
	residue = null;
	atomVec = null;
    }
    public Pdb(int No){
	residueNo = No;
	atomVec = new Vector();
	residue = "";
    }
    public Pdb(int No,  String res, Vector aa){
	residueNo = No;
	atomVec = aa;
	residue = res;
    }

    //getting the values	
    public int getResidueNo(){
	return residueNo;
    }	
    public Vector getAtomVec(){
	return atomVec;
    }	
    public String getResidue(){
	return residue;
    }	
    public String toString(){
	String desc = String.valueOf(residueNo) + "  "+ residue+"\n";
	String str="";
	for (int i=0; i<atomVec.size(); i++)
	    str += ((Cartesian) atomVec.elementAt(i)).toString();
	return (desc+str);	
    }

    public static class PdbComparator implements Comparator{
	public int compare(Object o1, Object o2){
	    Pdb n1 = (Pdb)o1;
	    Pdb n2 = (Pdb)o2;
	    int d1 = n1.getResidueNo();
	    int d2 = n2.getResidueNo();
	    if (d1 < d2)
		return -1;
	    else if (d1 > d2)
		return 1;
	    else return 0;
	}
    } 

    public double length(double[] v1){
	double v1Len = Math.sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]);
	return v1Len;
    }
    /** cal the directioal cosines of the vector vec */
    public double[] dirCos(double [] vec) {
	double len = 0.0;
	double [] dirs = new double[vec.length];
	for (int i=0; i<vec.length; i++)
	    len += vec[i] * vec[i];
	for (int j=0; j<vec.length; j++)
	    dirs[j] = vec[j] / Math.sqrt(len);
	return dirs;
    }

    /** calculate the angles between two vectors v1 and v2 
     * The returned angle is the [0,Pi] region
     */
    public double interAngle(double[] v1, double[] v2){
	double v1Len = Math.sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]);
	double v2Len = Math.sqrt(v2[0]*v2[0] + v2[1]*v2[1] + v2[2]*v2[2]);
	double c = (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2]) / (v1Len * v2Len);
	return Math.acos(c);  //acos [0, PI] so should be OK
    }

 /**
     * Compute the internuclear vector between two atoms
     @param n1 the coordinate for atom 1
     @param n2 the coordinate for atom 2
     @return a vector from n1->n2
    */
    public double[] internuclearVec(double[] n1, double[] n2){
        return new double[]{n2[0]-n1[0], n2[1]-n1[1], n2[2]-n1[2]};
    }

    /**
     * Reading PDB file in PDB Format!!!
     * the magic number for substring is based on format for pdb file
     @param  pdbFile the name of the file
     @return a vector of Pdb object 
    * Comment: There is always some irregularity in PDB files so when reading 
    * a real PDB file you must be cautious and ready to check the output from the reader
    * However, in our case we simply need coordinates which is much easier.
    */
    public Vector readPdb(String pdbFile){	
	String ss="", str="", a1_type, a2_type, r_type="", keywords, pdbID="";
	int Atoms_No, residueNo = 0, noTmp = 0;
	double x_in,y_in,z_in;
	Vector pdbVec = new Vector();
	Vector atoms = new Vector();
	String form="";
	try{
	    BufferedReader in = new BufferedReader(new FileReader(pdbFile));			
	    ss = in.readLine();
	    //	    System.out.println("ss :"+ss);
	    if (ss.length() > 62)
	        pdbID = ss.substring(62, 66).trim();
	    stop:
	    while(true) {
		if (ss.length() > 5)
		    keywords = ss.substring(0, 6).trim();
		else keywords = ss.trim();
		while(keywords.equals("ATOM") )	{
		    if ( ss.length() > 54){
			form = ss.substring(16,17).trim(); //We choose only one of it
			if (form.equals("B")){  //Some Pdb file like 1CBN has A and B 
			    if ((ss = in.readLine()) == null)
				break stop;							
			    if (ss.length() > 54)
				str = (ss.substring(0,6)).trim();
			    else continue stop;
//  			    System.out.println(" Sick Line "+ss);
			    while (str.equals("ANISOU")){
				if ((ss = in.readLine()) == null)
				    break stop;		
//  				System.out.println(" Sick Line "+ss);
				if (ss.length() > 54)
				    str = (ss.substring(0,6)).trim();
				else continue stop;
			    }
			    if (  !(str.equals("ATOM")) ){
				if(r_type.length() == 3){
				    pdbVec.add(new Pdb(residueNo, r_type, atoms));
				    atoms = new  Vector();	
				}
				continue stop;
			    }
			}
			a1_type = ss.substring(12, 16).trim();
			r_type = ss.substring(17, 20).trim();
			residueNo = Integer.parseInt(ss.substring(23, 26).trim());
			x_in = new Double(ss.substring(30, 38).trim()).doubleValue();
			y_in = new Double(ss.substring(38, 46).trim()).doubleValue();
			z_in = new Double(ss.substring(46, 54).trim()).doubleValue();
			atoms.add(new Cartesian(a1_type, x_in, y_in, z_in));
			if ((ss = in.readLine()) == null)
			    break stop;		
			if (ss.length() > 54)
			    str = (ss.substring(0,6)).trim();
			else continue stop;
			while (str.equals("ANISOU")){
			    if ((ss = in.readLine()) == null)
				break stop;	
//  			    System.out.println(" Sick Line "+ss);
			    if (ss.length() > 54)
				str = (ss.substring(0,6)).trim();
			    else continue stop;
			}
			if (  !(str.equals("ATOM")) ){
			    if(r_type.length() == 3){
				pdbVec.add(new Pdb(residueNo, r_type, atoms));
				atoms = new  Vector();	
			    }
			    continue stop;
			}
			form = ss.substring(16,17).trim();
			if (form.equals("B")){
			    if ((ss = in.readLine()) == null)
				break stop;							
			    if (ss.length() > 54)
				str = (ss.substring(0,6)).trim();
			    else continue stop;
//  			    System.out.println(" Sick Line "+ss);
			    while (str.equals("ANISOU")){
				if ((ss = in.readLine()) == null)
				    break stop;		
//  				System.out.println(" Sick Line "+ss);
				if (ss.length() > 54)
				    str = (ss.substring(0,6)).trim();
				else continue stop;
			    }
			    if (  !(str.equals("ATOM")) ){
				if(r_type.length() == 3){
				    pdbVec.add(new Pdb(residueNo, r_type, atoms));
				    atoms = new  Vector();		
				}
				continue stop;
			    }
			}
			noTmp = Integer.parseInt(ss.substring(23, 26).trim());
			while (noTmp == residueNo){
			    a1_type = ss.substring(12, 16).trim(); 
			    r_type = ss.substring(17, 20).trim();
			    residueNo = Integer.parseInt(ss.substring(23, 26).trim());
			    x_in = new Double(ss.substring(30, 38).trim()).doubleValue();
			    y_in = new Double(ss.substring(38, 46).trim()).doubleValue();
			    z_in = new Double(ss.substring(46, 54).trim()).doubleValue();
			    atoms.add(new Cartesian(a1_type, x_in, y_in, z_in));
			    if ((ss = in.readLine()) == null)
				break stop;							
			    if (ss.length() > 54)
				str = (ss.substring(0,6)).trim();
			    else continue stop;
//  			    System.out.println(" Sick Line "+ss);
			    while (str.equals("ANISOU")){
				if ((ss = in.readLine()) == null)
				    break stop;		
//  				System.out.println(" Sick Line "+ss);
				if (ss.length() > 54)
				    str = (ss.substring(0,6)).trim();
				else continue stop;
			    }

			    if (!(str.equals("ATOM")) ){
				if(r_type.length() == 3){
				    pdbVec.add(new Pdb(residueNo, r_type, atoms));
				    atoms = new  Vector();		
				}
				continue stop;
			    }
			    form = ss.substring(16,17).trim();
			    if (form.equals("B")){
				if ((ss = in.readLine()) == null)
				    break stop;							
				if (ss.length() > 54)
				    str = (ss.substring(0,6)).trim();
				else continue stop;
//  				System.out.println(" Sick Line "+ss);
				while (str.equals("ANISOU")){
				    if ((ss = in.readLine()) == null)
					break stop;		
//  				    System.out.println(" Sick Line "+ss);
				    if (ss.length() > 54)
					str = (ss.substring(0,6)).trim();
				    else continue stop;
				}
				if (  !(str.equals("ATOM")) ){
				    if(r_type.length() == 3){
					pdbVec.add(new Pdb(residueNo, r_type, atoms));
					atoms = new  Vector();		
				    }
				    continue stop;
				}
			    }
			    noTmp = Integer.parseInt(ss.substring(23, 26).trim());
			}
			if(r_type.length() == 3){
			    pdbVec.add(new Pdb(residueNo, r_type, atoms));
			    atoms = new  Vector();
			}
			keywords = str;		
		    }
		}
		if ((ss = in.readLine()) == null){
		    pdbVec.add(new Pdb(residueNo, r_type, atoms));
		    break stop;
		}
	    }
	}catch (FileNotFoundException e) {
	    System.out.println("File not found: " + pdbFile);
     	}catch (IOException e) {
	    System.out.println("IOException: the stack trace is:");
	    e.printStackTrace();
	}
	Collections.sort(pdbVec,new PdbComparator() );
	return pdbVec;
    }

    /** 
     * compute new PDB coordinates after a rotation by matrix R
     @param pdbVec original PDB coordinates
     @param R      the rotation matrix
     @return       a new vector of PDB object
     */
    public Vector newPdb (final Vector pdbVec, final Matrix R){
	Cartesian atom = null;
	String atomName="", resid;
	Vector atoms = new Vector();
	Vector atomVec = new Vector();
	Vector pdbVecN = new Vector();
	int no;
	double [] coord = new double[3];
	double [] coordN = new double[3];
	Pdb pp = new Pdb();
	for (int i=0; i<pdbVec.size(); i++){
	    pp = (Pdb)pdbVec.elementAt(i);
	    no = pp.getResidueNo();
	    resid = (pp.getResidue()).trim();
	    atoms = pp.getAtomVec();
	    for (int j=0; j<atoms.size(); j++){
		atom = (Cartesian) atoms.elementAt(j);
		atomName = atom.getAtom();
		coord[0] = atom.getX();
		coord[1] = atom.getY();
		coord[2] = atom.getZ();
		coordN = R.times(coord);
		atomVec.add(new Cartesian(atomName, coordN[0], coordN[1], coordN[2]));
	    }
	    pdbVecN.add(new Pdb(no, resid, atomVec));
	    atomVec = new Vector();
	}
	return pdbVecN;
    }

    /** 
     * compute new PDB coordinates after a translation
     @param pdbVec  original PDB coordinates
     @param vT      the center in which the PDB is being tranlated to
     @return        a new vector of PDB object
    */
    public Vector newPdbByTranslation (final Vector pdbVec, double[]  vT){
	Cartesian atom = null;
	String atomName ="", resid = "";
	Vector atoms = new Vector();
	Vector atomVec = new Vector();
	Vector pdbVecN = new Vector();
	int no;
	double [] coord = new double[3];
	double [] coordN = new double[3];
	Pdb pp = new Pdb();
	for (int i=0; i<pdbVec.size(); i++){
	    pp = (Pdb)pdbVec.elementAt(i);
	    no = pp.getResidueNo();
	    resid = (pp.getResidue()).trim();
	    atoms = pp.getAtomVec();
	    for (int j=0; j<atoms.size(); j++){
		atom = (Cartesian) atoms.elementAt(j);
		atomName = atom.getAtom();
		coord[0] = atom.getX();
		coord[1] = atom.getY();
		coord[2] = atom.getZ();
		coordN =  new double[] {coord[0] + vT[0], coord[1] + vT[1], coord[2] + vT[2]}; //new coordinates
		atomVec.add(new Cartesian(atomName, coordN[0], coordN[1], coordN[2]));
	    }
	    pdbVecN.add(new Pdb(no, resid, atomVec));
	    atomVec = new Vector();
	}
	return pdbVecN;
    }

    /** 
     * Compute Phi and Psi directly from three backbone vectors (not
     * NH vector), a helper
     @param n2CO1  N->CO vector
     @param n2CA   N->CA vector
     @param ca2CO  CA->CO vector
     @return phipsi angles
     */  
    public double PhiPsiCalPDB(double[] n2CO1, double[] n2CA, double[] ca2CO){
	double [] nToCO1 = dirCos(n2CO1);
	double [] nToCA = dirCos(n2CA);
	double [] caToCO = dirCos(ca2CO);
	// Cal the rotate matrix: First cal the polar angle phi and theta of the 
	// intermediate vector(N->CA) from it's directional cosines
	double theta_nToCa = Math.acos(nToCA[2]); //acos [0,PI) so should be OK
	double phi_nToCa = 0.0;
	double x = nToCA[0];
	double y = nToCA[1];
	if (theta_nToCa != 0.0){
	    if (x > 0) 
		phi_nToCa = Math.atan(y/x);
	    else if (x < 0)
		phi_nToCa = Math.atan(y/x) + Math.PI;
	    else {
		if (y > 0)
		    phi_nToCa = Math.PI / 2.0;
		else phi_nToCa = 3*Math.PI / 2.0;
	    }
	} else  phi_nToCa = 0.0;
	final int m = 3; //The dimensions of the Rotation Matrix
	final int n = 3;
	final double numMin = 1.0E-10;
	//construct the rotation matrix A needed to bring N-CA vector onto the +y-axis
	double [][] A = new double[3][3];
	A[0][0] = Math.sin(phi_nToCa);
	A[0][1] = -Math.cos(phi_nToCa);
	A[0][2] = 0.0;
	A[1][0] = Math.sin(theta_nToCa) * Math.cos(phi_nToCa);
	A[1][1] = Math.sin(theta_nToCa) * Math.sin(phi_nToCa);
	A[1][2] = Math.cos(theta_nToCa);
	A[2][0] = -Math.cos(theta_nToCa) * Math.cos(phi_nToCa);
	A[2][1] = -Math.cos(theta_nToCa) * Math.sin(phi_nToCa);
	A[2][2] = Math.sin(theta_nToCa);
	
	Matrix Rn2ca = new Matrix(A);
	//Rotate the vectors N(i)->CO(i-1) and CA(i)->CO(i) 
	double [] n2Co1 = Rn2ca.times(nToCO1);
	double [] ca2Co = Rn2ca.times(caToCO);
	double [] n2Ca  = Rn2ca.times(nToCA);
	double x1 = ca2Co[0];
	double y1 = ca2Co[1];
	double z1 = ca2Co[2];  
	double x2 = n2Co1[0];
	double y2 = n2Co1[1];
	double z2 = n2Co1[2];
	//For standard geometry gamma1 and gamma2 can NOT be zero
	double sin_gamma1 = Math.sqrt(1 - y1*y1);
	double sin_gamma2 = Math.sqrt(1 - y2*y2);
	double cosPhi = (z1*z2+x1*x2) / (sin_gamma1 * sin_gamma2);
	double sinPhi = (x1*z2-z1*x2) / (sin_gamma1 * sin_gamma2);
	if (sinPhi > 0){
	    if ( (sinPhi > 1.0) && ( Math.abs(sinPhi - 1.0) < numMin) )
		return 0.0;
	    else if (  (sinPhi > 1.0) && (Math.abs(sinPhi - 1.0) >= numMin)){
		System.out.println("Math Error in PhiPsiCalPDB method");
		return 0.0;
	    }
	    else return (Math.acos(cosPhi));
	}else if (sinPhi < 0){
	    if ( (sinPhi < -1.0) && (Math.abs(sinPhi + 1.0) < numMin) )
		return Math.PI;
	    else if ( (sinPhi < -1.0) &&  (Math.abs(sinPhi + 1.0) >= numMin) ){
		System.out.println("Math Error in PhiPsiCalPDB method");
		return 0.0;
	    }else return (-Math.acos(cosPhi));
	}else {
	    if (cosPhi > 0)
		return 0.0;
	    else return Math.PI;
	}
    }

    /**
     * Compute all the backbone Phi and Psi angles from PDB from the
     * backbone backbone bond vectors between heavy atoms.
     @param  pdbVec a vector of Pdb object derived from a Pdb file
     @return a vector of PhiPsi object
     */  
    public Vector PhiPsiTotalPdb(final Vector pdbVec){
	double [] amide1 = new double[3];
	double [] amide2 = new double[3];
	double [] co1 = new double[3];
	double [] co2 = new double[3];
	double [] ca1 = new double[3];
	double [] ca2 = new double[3];

	int no1 = 0, no2 = 0, no3 = 0, i=0, j=0, k=0;
	String resid1="", resid2="", resid3="";
	Cartesian cc = new Cartesian();
	String atom = "";
	//First Residue
	Pdb pp = (Pdb)pdbVec.elementAt(i); 
	no1 = pp.getResidueNo();
	resid1 = pp.getResidue();
	Vector atomVec = pp.getAtomVec();
	//the Atoms N, CA and CO of residue i
	for (j=0; j<atomVec.size(); j++){
	    cc = (Cartesian)atomVec.elementAt(j);
	    atom = cc.getAtom();
	    if (atom.equals("N"))
		amide1 = cc.getXYZ();
	    else if (atom.equals("CA"))
		ca1 = cc.getXYZ();
	    else if (atom.equals("C"))
		co1 = cc.getXYZ();
	}
	// Beginning With Second Residue
	double [] nToCO1Vec = new double[3];
	double [] co1ToNVec = new double[3];
	Vector phiPsiVec = new Vector(pdbVec.size());
	double[] phiArr = new double[pdbVec.size()];
	double[] psiArr = new double[pdbVec.size()];
	for (int m=0; m<phiPsiVec.size(); m++){
	    phiArr[m] = 999.0;         //For labelling
	    psiArr[m] = 999.0;
	}
	phiArr[i] = 0.0; //Assign the phi of the first Residue as 0.0
	//Cal the intraresidue CA->N and CA->CO vector
	double [] caToNVec  =  internuclearVec(ca1, amide1);
	double [] caToCOVec = internuclearVec(ca1, co1);
	i++;       //The second residue
	pp  = (Pdb)pdbVec.elementAt(i);
	no2 = pp.getResidueNo();
	resid2  = pp.getResidue();
	atomVec = pp.getAtomVec();
	for (j=0; j<atomVec.size(); j++){ //The atoms N, CA and C of residue i+1
	    cc = (Cartesian)atomVec.elementAt(j);
	    atom = cc.getAtom();
	    if (atom.equals("N"))
		amide2 = cc.getXYZ();
	    else if (atom.equals("CA"))
		ca2 = cc.getXYZ();
	    else if (atom.equals("C"))
		co2 = cc.getXYZ();
	}
	// Cal the CO(i-1)->N(i) Vector
	if ( no2 == (no1+1) ){
	    co1ToNVec =  internuclearVec(co1, amide2);
	    psiArr[i-1] = PhiPsiCalPDB(caToNVec, caToCOVec, co1ToNVec); //Cal the 1st Psi
	}
	phiPsiVec.addElement(new PhiPsi(no1, resid1, phiArr[i-1], psiArr[i-1]));
	//The intraResidue N(i)->CA(i) vector
	double [] nToCAVec = internuclearVec(amide2, ca2);
	//The intraResidue CA(i)->CO(i) vector
	caToCOVec = internuclearVec(ca2, co2);
	if (no2 == (no1 + 1)){  // if consecutive, compute the interResidue N(i)->CO(i-1) vector
	    nToCO1Vec =  internuclearVec(amide2, co1);
	    phiArr[i] = PhiPsiCalPDB(nToCO1Vec, nToCAVec, caToCOVec); //calculate phi
	}
        i++; //Begin with the third residue we go into the While loop
	while (i < pdbVec.size()){
	    pp = (Pdb)pdbVec.elementAt(i);
	    no3 = pp.getResidueNo();
	    resid3 = pp.getResidue();
	    atomVec = pp.getAtomVec();
	    for (j=0; j<atomVec.size(); j++){
		cc = (Cartesian)atomVec.elementAt(j);
		atom = cc.getAtom();
		if (atom.equals("N"))
		    amide1 = cc.getXYZ();
		else if (atom.equals("CA"))
		    ca1 = cc.getXYZ();
		else if (atom.equals("C"))
		    co1 = cc.getXYZ();
	    }
	    //Cal the Vector CO(i)->N(i+1), co2 from previous residue
	    if ( no3 == (no2+1) ){
		co1ToNVec =  internuclearVec(co2, amide1);
		for (k=0; k<nToCAVec.length; k++) //reverse the direction of the N(i)->CA(i)
		    caToNVec[k] = -nToCAVec[k];
		psiArr[i-1] = PhiPsiCalPDB(caToNVec, caToCOVec, co1ToNVec);
	    }
	    phiPsiVec.addElement(new PhiPsi(no2,resid2,phiArr[i-1], psiArr[i-1]));
	    //Cal the Vector N(i)->CA(i)    		
	    nToCAVec  = internuclearVec(amide1, ca1);
	    //Cal the Vector CA(i)->CO(i)    		
	    caToCOVec = internuclearVec(ca1, co1);
	    if ( no3 == (no2+1) ){
		for (k=0; k<co1ToNVec.length; k++) //reverse the direction of CO(i)->N(i+1)
		    nToCO1Vec[k] = -co1ToNVec[k];
		phiArr[i] = PhiPsiCalPDB(nToCO1Vec, nToCAVec, caToCOVec);
	    }
	    for (k=0; k<co1.length; k++)  //update co2
		co2[k] = co1[k];
	    i++;      
	    no2 = no3;   //Updating the residueNO and ResidueType
	    resid2 = resid3;
	}
	psiArr[i-1] = 0.0;  //The last Psi is assigned to 0.0
	phiPsiVec.addElement(new PhiPsi(no2,resid2,phiArr[i-1],psiArr[i-1]));
// 	System.out.println((no2-1)+"  "+(phiArr[i-2]/cst)+"  "+(psiArr[i-2]/cst));
	return phiPsiVec;
    }
 
    /**
     *  Extract backbone atoms as defined in the code from a PDB file
     @param pdbVec the original Pdb vector
     @return a Pdb file with only protein backbone atoms
     *
     */
    public Vector extractBackbone(final Vector pdbVec){
	Cartesian cc = null;
	String atom="", resid;
	Vector atomVec = new Vector();
	Vector atomVecN = new Vector();
	Vector pdbVecN = new Vector();
	int no;
	Pdb pp = new Pdb();
	for (int i=0; i<pdbVec.size(); i++){
	    pp = (Pdb)pdbVec.elementAt(i);
	    no = pp.getResidueNo();
	    resid = (pp.getResidue()).trim();
	    atomVec = pp.getAtomVec();
	    for (int j=0; j<atomVec.size(); j++){
		cc = (Cartesian) atomVec.elementAt(j);
		atom = cc.getAtom();
		if (atom.equals("N") )
		    atomVecN.add(new Cartesian(atom, cc.getXYZ()));
		else if (atom.equals("H"))
		    atomVecN.add(new Cartesian(atom, cc.getXYZ()));
		else if (atom.equals("CA"))
		    atomVecN.add(new Cartesian(atom, cc.getXYZ()));
		else if (atom.equals("HA"))
		    atomVecN.add(new Cartesian(atom, cc.getXYZ()));
		else if (atom.equals("C"))
		    atomVecN.add(new Cartesian(atom, cc.getXYZ()));
		else if (atom.equals("O"))
		    atomVecN.add(new Cartesian(atom, cc.getXYZ()));
	    }
	    pdbVecN.add(new Pdb(no, resid, atomVecN));
	    atomVecN = new Vector();
	}
	return pdbVecN;
    }

    /** 
     * A method for printing the coordinates in PDB format
     * the magic Number comes from PDB file format
     @param pdbVec a vector of Pdb object
     */
    public void print (Vector pdbVec) {
	final int d = 3;              //number of digits after the decimal
	int w = 0;
	DecimalFormat format = new DecimalFormat();
	format.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.US));
	format.setMinimumIntegerDigits(1);
	format.setMaximumFractionDigits(d);
	format.setMinimumFractionDigits(d);
	format.setGroupingUsed(false);
      	Vector atomVec = new Vector();
	Cartesian cc = new Cartesian();
	Pdb pp = new Pdb();
	String noS;
	String resid, atom;
	double x,y,z;
	String sX, sY, sZ;
	int padding = 0;
	String atomStr;
	int i, j, k;
	int atomNo=0;
	String str="";
 	for (i=0; i<pdbVec.size(); i++){
	    pp = (Pdb)pdbVec.elementAt(i);
	    noS = String.valueOf(pp.getResidueNo());
	    resid = pp.getResidue();
	    atomVec = pp.getAtomVec();
	    for (j=0; j<atomVec.size(); j++){
		cc = (Cartesian) atomVec.elementAt(j);
		atom = cc.getAtom();
		x = cc.getX();
		y = cc.getY();
		z = cc.getZ();
		sX = format.format(x);
		sY = format.format(y);
		sZ = format.format(z);
		w = 39-31;
		padding = w - sX.length(); 
		for (k = 0; k < padding; k++)
		    sX = " "+sX;
		padding = w - sY.length(); 
		for (k = 0; k < padding; k++)
		    sY = " "+sY;
		padding = w - sZ.length(); 
		for (k = 0; k < padding; k++)
		    sZ = " "+sZ;
      
		str += "ATOM  ";
		w = 12 - 7 ;
 		atomNo++;
		atomStr = String.valueOf(atomNo);
		padding = w - atomStr.length(); 
		for (k = 0; k < padding; k++)
		    atomStr = " "+atomStr;
		str += atomStr+" "; //add 2 empty space
		
		w = 17 - 13; 
		if (Character.isDigit(atom.charAt(0))){
		    padding = w - atom.length();
		    for (k = 0; k < padding; k++)
			atom += " "; 
		}
		else {
		    atom = " "+ atom;
		    padding =  w - atom.length();
		    for (k = 0; k < padding; k++)
			atom += " "; 
		}
		str += atom+" "; //add 1 empty space

		str += resid +"  "; //add 2 empty space, if has chainID go to here

		w = 27 - 23 ;
		padding = w - noS.length(); // At _least_ 1 space
		for (k = 0; k < padding; k++)
		    noS = " "+noS;
		str += noS+"    "; //add 4 empty space, if has iCode go to here

		str += sX+sY+sZ+" ";
		str += " 1.00"+" ";  //Occupancy number
		str += " 0.00"+" ";  //tempaerature Factor
		padding = 80 - str.length(); //finally make the length 80. Some junk stuff go to here
		for (k = 0; k < padding; k++)
			str += " "; 
		System.out.println(str);
		noS = String.valueOf(pp.getResidueNo());
		str="";
	    }
	}
    }

    //A simple test driver as a example
    public static void main(String[] args) {      	
	Pdb pp = new Pdb();
	String src = "../inputFiles/";
 	String pdbFile = src+"1UBQH.pdb";
  	Vector pdbVec = pp.readPdb(pdbFile);
 	Vector pdbBB = pp.extractBackbone(pdbVec);
	pp.print(pdbBB);
    }
}
