/*****************************************************************************************
 * PdbRdc.java

 * Version:           0.1

 * Description: 

 * This class implement SVD methods for best fitting RDCs and a
 * structural fragment and the grid search method for computing the
 * three Euler angles needed for computing the first peptide plane

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;

public class PdbRdc {
    private Pdb pdbCoord;  //an Pdb object associated with the RDC value
    private double rdcVal;
  
    public PdbRdc (){
	pdbCoord = null;
	rdcVal = 0.0;
    }
    public PdbRdc (Pdb pp, double rdc){
	pdbCoord = pp;
	rdcVal = rdc;
    }
    public Pdb getPdb(){
	return pdbCoord;
    }
    public double getRdc(){
	return rdcVal;
    }

    /** cal the directioal cosines of the vector vec */
    public double[] dirCos(double [] vec) {
	double len = 0.0;
	double [] dirs = new double[vec.length];
	for (int i=0; i<vec.length; i++)
	    len += vec[i] * vec[i];
	for (int j=0; j<vec.length; j++)
	    dirs[j] = vec[j] / Math.sqrt(len);
	return dirs;
    }

    /**
     * Compute the internuclear vector between two atoms
     @param n1 the coordinate for atom 1
     @param n2 the coordinate for atom 2
     @return a vector from n1->n2
    */
    public double[] internuclearVec(double[] n1, double[] n2){
        return new double[]{n2[0]-n1[0], n2[1]-n1[1], n2[2]-n1[2]};
    }

    /** 
     *Computing the Saupe matrix by SVD method and then
     *diagonalizing it for computing the diagonalized elements and
     *rotation matrix to one of POF. 
     @param pdbVec  a vector of PDB objects.
     @param rdcVec  a corresponding RDC vector.
     @param rmsdArr for returning the five Saupe elements and the best-fitted RMSD 
     @param atom1   the nucleus type whose RDC for the internuclear vector are recorded
     @param atom2   the nucleus type whose RDC for the internuclear vector are recorded
     @param printResults if true, it prints the back-computed RDCs.
     @return True if the SVD matrix is NOT singular, otherwise False
     *The order of atom1 and atom2 is important (different order == sign inversion)
     */
    public boolean bestFit(final Vector pdbVec, final Vector rdcVec, String atom1, String atom2,
			   double[] rmsdArr, boolean printResults){
	Pdb pp = new Pdb();
	Dipolar dd = new Dipolar();
	int i, j, k;
	int no;
	String resid;
	Vector atomVec = new Vector();
	double [] nToNHVec = new double[3];
	double [] amide = new double[3];
	double [] nh = new double[3];
	double cosX = 0.0, cosY = 0.0, cosZ = 0.0;
	double rdcExp = 0.0;
	Vector pdbRdcVec = new Vector();
	Cartesian cc = new Cartesian();
	String atom = "";
	int index;
	for (i=0; i<rdcVec.size(); i++){
	    dd = (Dipolar)rdcVec.elementAt(i); 
	    rdcExp = dd.getRdc();
	    no = dd.getResidueNo();
	    index = Collections.binarySearch(pdbVec, new Pdb(no), new Pdb.PdbComparator());
	    if (index > -1)
		pdbRdcVec.add(new PdbRdc((Pdb)pdbVec.elementAt(index), rdcExp)); 
	}
	int m = pdbRdcVec.size();   //number of rows
	int n = 5;                    //number of columns
	double [][] equ = new double[m][n];
	double [] b     = new double[m];
	for (i=0; i<pdbRdcVec.size(); i++){
	    pp = ((PdbRdc)pdbRdcVec.elementAt(i)).getPdb();
	    atomVec = pp.getAtomVec();
	    for (j=0; j<atomVec.size(); j++){
		cc = (Cartesian)atomVec.elementAt(j);
		atom = cc.getAtom();
		if (atom.equals(atom1))
		    amide = cc.getXYZ();
		else if (atom.equals(atom2))
		    nh = cc.getXYZ();
	    }
	    nToNHVec = internuclearVec(amide, nh);
	    nToNHVec = dirCos(nToNHVec);
	    cosX = nToNHVec[0];
	    cosY = nToNHVec[1];
	    cosZ = nToNHVec[2];
	    equ[i][0] = cosY * cosY - cosX * cosX;
	    equ[i][1] = cosZ * cosZ - cosX * cosX;
	    equ[i][2] = 2.0 * cosX * cosY;
	    equ[i][3] = 2.0 * cosX * cosZ;
	    equ[i][4] = 2.0 * cosY * cosZ;
	    b[i] = ((PdbRdc)pdbRdcVec.elementAt(i)).getRdc();
	}
	Matrix A = new Matrix(equ, m, n);
	SingularValueDecomposition SVD = A.svd();
	double [] singularvalues = SVD.getSingularValues();
	double[] singulars = new double[6];
// 	System.out.println("Singular Values: ");
	for (k=0; k<singularvalues.length; k++){
	    if (singularvalues[k] != 0.0)
		singulars[k] = 1 / singularvalues[k];
	    else return false;
	}
	double [][] ArrS = { {singulars[0], 0.0, 0.0, 0.0, 0.0},
			     {0.0, singulars[1], 0.0, 0.0, 0.0},
			     {0.0, 0.0, singulars[2], 0.0, 0.0},
			     {0.0, 0.0, 0.0, singulars[3], 0.0},
			     {0.0, 0.0, 0.0, 0.0, singulars[4]}};
	Matrix U = SVD.getU();
	Matrix V = SVD.getV();
	Matrix S  = SVD.getS();
	Matrix invS = new Matrix(ArrS, n, n);
	int errorCount = 0;
	/** comment out at your own risk */
	try {
	    check(A,SVD.getU().times(SVD.getS().times(SVD.getV().transpose())));
	    try_success("SingularValueDecomposition...","");
	} catch ( java.lang.RuntimeException e ) {
	    errorCount = try_failure(errorCount,"SingularValueDecomposition...",
				     "incorrect singular value decomposition calculation");
	}
	/** comment out at your own risk */

	double [] saupe = V.times(invS.times(U.transpose().times(b)));
	double Syy = saupe[0], Szz = saupe[1], Sxy = saupe[2], Sxz = saupe[3], Syz = saupe[4];
	double r = 0.0, rdcCal = 0.0;
	//back-computing RDCs	
	if (printResults)
	    System.out.println("ResidueNO,   back-computed RDC,   measured RDC,  difference");
	for (i = 0; i < pdbRdcVec.size(); i++){
	    pp = ((PdbRdc)pdbRdcVec.elementAt(i)).getPdb(); 
	    atomVec = pp.getAtomVec();
	    for (j=0; j<atomVec.size(); j++){
		cc = (Cartesian)atomVec.elementAt(j);
		atom = cc.getAtom();
		if (atom.equals(atom1))
		    amide = cc.getXYZ();
		else if (atom.equals(atom2))
		    nh = cc.getXYZ();
	    }
	    nToNHVec = internuclearVec(amide, nh);
	    nToNHVec = dirCos(nToNHVec);
	    cosX = nToNHVec[0];
	    cosY = nToNHVec[1];
	    cosZ = nToNHVec[2];
	    rdcCal =  (cosY * cosY - cosX * cosX) * Syy
		       + (cosZ * cosZ - cosX * cosX) * Szz
		       + 2.0 * cosX * cosY * Sxy
		       + 2.0 * cosX * cosZ * Sxz
		       + 2.0 * cosY * cosZ * Syz;
	    rdcExp = ((PdbRdc)pdbRdcVec.elementAt(i)).getRdc();
	    if (printResults)
		System.out.println(pp.getResidueNo()+"   "+rdcCal+"   "+rdcExp+"   "+(rdcCal-rdcExp));
	    r += (rdcCal-rdcExp) * (rdcCal-rdcExp);
	}
	if (pdbRdcVec.size() < 2){
	    System.out.println("Error in bestFit method, possible reason: either Pdb or RDC file is not read properly");
	    System.exit(1);
	}
	if (printResults)
	    System.out.println("rmsd :"+ Math.sqrt( r / pdbRdcVec.size() ) );
	for (k=0; k<saupe.length; k++)
	    rmsdArr[k] = saupe[k];
	rmsdArr[5] = Math.sqrt(r / pdbRdcVec.size() );  //The last one
	return true;
    }
 
    /**
     * Computing the seven parameters from the RDCs in two media
     @param pdbVec   the structure model
     @param rdc1Vec  the first set of RDCs
     @param rdc2Vec  the second set of RDCs
     @param n2ca     returning the rotated N->CA vector.
     @param eignValues returning the four Saupe elements in the order of Syy1, Szz1, Syy2, Szz2.
     @param printResults if true it prints the back-computed RDCs in bestFit method
     @return relative matrix between the two POFs
     */
    public Matrix Tensors(Vector pdbVec, Vector rdc1Vec, Vector rdc2Vec, double[] n2ca, 
			  double[] eignValues, boolean printResults){
 	double[] saupe = new double[6];
	boolean flagSVD = bestFit(pdbVec, rdc1Vec, "N", "H", saupe, printResults); //N and H are relics 
	double Syy = saupe[0], Szz = saupe[1], Sxy = saupe[2], Sxz = saupe[3], Syz = saupe[4];
	double[][] mArr = {{-Syy-Szz, Sxy, Sxz},
			   {Sxy,    Syy,   Syz},
			   {Sxz,    Syz,   Szz}};
	Matrix mm = new Matrix(mArr);
	EigenvalueDecomposition eigs = new EigenvalueDecomposition(mm, true);
	Matrix m1 = eigs.getV();
	Matrix DD = eigs.getD();
	eignValues[0] = DD.get(1, 1);
	eignValues[1] = DD.get(2, 2);
	eignValues[2] = saupe[5];
	Matrix r1 = m1.transpose();

	//Make sure the coordinate system has the right rightedness after the diagnolization
	Pdb pp = new Pdb();
	Vector phiPsiVec = pp.PhiPsiTotalPdb(pdbVec);  //We are lazy, only the first \psi angle is compared
	Vector pdbVecN = pp.newPdb(pdbVec, r1);
	Vector phiPsiVec1 = pp.PhiPsiTotalPdb(pdbVecN);
	if (Math.abs(((PhiPsi)phiPsiVec.elementAt(0)).getPsi()-((PhiPsi)phiPsiVec1.elementAt(0)).getPsi()) 
	    > 10*Const.eps)
	    r1 = Const.mLeftHand.times(r1);

	flagSVD = bestFit(pdbVec, rdc2Vec, "N", "H", saupe, printResults); 
	if (!flagSVD){
	    System.out.println(" SVD failed in Tensors method ");
	    System.exit(0);
	}
	Syy = saupe[0]; 
	Szz = saupe[1]; 
	Sxy = saupe[2]; 
	Sxz = saupe[3];
	Syz = saupe[4];
	mArr = new double[][]{{-Syy-Szz, Sxy, Sxz},
			      {Sxy,   Syy,    Syz},
			      {Sxz,   Syz,    Szz}};
	mm = new Matrix(mArr);
	eigs = new EigenvalueDecomposition(mm, true);
	Matrix m2 = eigs.getV();
	DD = eigs.getD();
	eignValues[3] = DD.get(1, 1);
	eignValues[4] = DD.get(2, 2);
	eignValues[5] = saupe[5];
	Matrix r2 = m2.transpose();

	//Make sure the coordinate system has the right rightedness after the diagnolization
	pdbVecN = pp.newPdb(pdbVec, r2);
	Vector phiPsiVec2 = pp.PhiPsiTotalPdb(pdbVecN);
	if (Math.abs(((PhiPsi)phiPsiVec.elementAt(0)).getPsi() - ((PhiPsi)phiPsiVec2.elementAt(0)).getPsi()) 
	    > 10*Const.eps )
	    r2 = Const.mLeftHand.times(r2);
	//Compute the rotated first N->CA vector
	double [] amide = new double[3];
	double [] ca = new double[3];
	Cartesian cc = new Cartesian();
	String atom = "";
	pp = (Pdb)pdbVec.elementAt(0);  
	Vector atomVec = pp.getAtomVec();
	for (int j=0; j<atomVec.size(); j++){ 
	    cc = (Cartesian)atomVec.elementAt(j);
	    atom = cc.getAtom();
	    if (atom.equals("N"))
		amide = cc.getXYZ();
	    else if (atom.equals("CA"))
		ca = cc.getXYZ();
	}
	double [] ca1 = pp.internuclearVec(amide, ca);
	double [] nca = r1.times(ca1);
	System.arraycopy(nca, 0, n2ca, 0, 3);
 	return (r2.times(r1.transpose()));
    }

    /**
     * Turn the Pdb coordinate to righthanded if the original one is left-handed after SVD
     * SVD and the EigenValueDecomposition method can return either handedness
     @param pdbVec   the structural model
     @param rdc1Vec  the first set of RDCs
     @param rdc2Vec  the second set of RDCs
     @return The righthanded PDB vector
     */
    public Vector rightHandedPdb(Vector pdbVec, Vector rdc1Vec, Vector rdc2Vec){
 	double[] saupe = new double[6];
	boolean printResults = false;
	boolean flagSVD = bestFit(pdbVec, rdc1Vec, "N", "H", saupe, printResults); //N and H are relics 
	double Syy = saupe[0], Szz = saupe[1], Sxy = saupe[2], Sxz = saupe[3], Syz = saupe[4];
	double[][] mArr = {{-Syy-Szz, Sxy, Sxz},
			   {Sxy,    Syy,   Syz},
			   {Sxz,    Syz,   Szz}};
	Matrix mm = new Matrix(mArr);
	EigenvalueDecomposition eigs = new EigenvalueDecomposition(mm, true);
	Matrix m1 = eigs.getV();
	Matrix r1 = m1.transpose();
	//Make sure the coordinate system has the right rightedness after the diagnolization
	Pdb pp = new Pdb();
	Vector phiPsiVec = pp.PhiPsiTotalPdb(pdbVec);  //We are lazy, only the first \psi angle is compared
	Vector pdbVecN = pp.newPdb(pdbVec, r1);
	Vector phiPsiVec1 = pp.PhiPsiTotalPdb(pdbVecN);
	if (Math.abs(((PhiPsi)phiPsiVec.elementAt(0)).getPsi()-((PhiPsi)phiPsiVec1.elementAt(0)).getPsi()) 
	    > 10*Const.eps)
	    r1 = Const.mLeftHand.times(r1);
 	return (pp.newPdb(pdbVec, r1));
    }

    /**
     * back-compute RDCs
     @param pdbVec the pdb coordinates
     @param rdcVec the meausred RDC values
     @param atom1 the atoms of which the internuclear vector has RDC values 
     @param atom2
     @param mm the matrix used to rotate the original PDB frame to one of POF 
     @param Sxx the three diagonalized Saupe elements
     @param Syy 
     @param Szz
     @return RMSDs between the computed RDCs and the experimental RDCs.
     **/
    public double BackCal(final Vector pdbVec, final Vector rdcVec, String atom1, String atom2, 
			    Matrix mm, double Sxx, double Syy, double Szz){
	int i, j, no;
	int index = 0;
	Cartesian cc = new Cartesian();
	Vector atomVec = new Vector();
	double [] nToNHVec = new double[3];
	double [] amide = new double[3];
	double [] nh = new double[3];

	double r = 0.0, rdcCal = 0.0;
	Dipolar dd = new Dipolar();
	Pdb pp = new Pdb();
	double cosX = 0.0, cosY = 0.0, cosZ = 0.0;
	double rdcExp = 0.0;
	Vector pdbRdcVec = new Vector();
	String atom = "";
	for (i=0; i<rdcVec.size(); i++){
	    dd = (Dipolar)rdcVec.elementAt(i); 
	    rdcExp = dd.getRdc();
	    no = dd.getResidueNo();
	    index = Collections.binarySearch(pdbVec, new Pdb(no), new Pdb.PdbComparator());
	    if (index > -1)
		pdbRdcVec.add(new PdbRdc((Pdb)pdbVec.elementAt(index), rdcExp)); 
	}
	for (i = 0; i < pdbRdcVec.size(); i++){
	    pp = ((PdbRdc)pdbRdcVec.elementAt(i)).getPdb(); 
	    atomVec = pp.getAtomVec();
	    for (j=0; j<atomVec.size(); j++){
		cc = (Cartesian)atomVec.elementAt(j);
		atom = cc.getAtom();
		if (atom.equals(atom1))
		    amide = cc.getXYZ();
		else if (atom.equals(atom2))
		    nh = cc.getXYZ();
	    }
	    nToNHVec = internuclearVec(amide, nh);
	    nToNHVec = mm.times(dirCos(nToNHVec));
	    cosX = nToNHVec[0];
	    cosY = nToNHVec[1];
	    cosZ = nToNHVec[2];
	    rdcCal = cosX * cosX * Sxx+ cosY * cosY * Syy + cosZ * cosZ * Szz; 
	    rdcExp = ((PdbRdc)pdbRdcVec.elementAt(i)).getRdc();
	    //System.out.println(pp.getResidueNo()+"   "+rdcCal+"   "+rdcExp+"   "+(rdcCal-rdcExp));
	    r += (rdcCal-rdcExp) * (rdcCal-rdcExp);
	}
	if (pdbRdcVec.size() < 2){
	    System.out.println("Error in file Size 1401: "+pdbRdcVec.size());
	    System.exit(1);
	}
	//System.out.println("rmsd :"+ Math.sqrt(r / pdbRdcVec.size() ));
	//For back-computing RDCs of those residue with missing experimental RDCs
	for (i = 0; i < pdbVec.size(); i++){
	    pp = (Pdb)pdbVec.elementAt(i); 
	    atomVec = pp.getAtomVec();
	    for (j=0; j<atomVec.size(); j++){
		cc = (Cartesian)atomVec.elementAt(j);
		atom = cc.getAtom();
		if (atom.equals(atom1))
		    amide = cc.getXYZ();
		else if (atom.equals(atom2))
		    nh = cc.getXYZ();
	    }
	    nToNHVec = internuclearVec(amide, nh);
	    nToNHVec = mm.times(dirCos(nToNHVec));
	    cosX = nToNHVec[0];
	    cosY = nToNHVec[1];
	    cosZ = nToNHVec[2];
	    rdcCal = cosX * cosX * Sxx + cosY * cosY * Syy + cosZ * cosZ * Szz; 
	    //System.out.println(pp.getResidueNo()+"   "+rdcCal+"    "+1.0+"   "+0.0);
	}
	return 	(Math.sqrt(r / pdbRdcVec.size() ));
    }

    /** 
     * Compute three Euler angles using two sets of RDC data and the corresponding Saupe elements
     * Basic steps:
     *   extract the NH vectors with the corresponding RDC data as an array (for efficiency)
     *   For all the values of three Euler angles first compute the new NH directions after the rotation
     *   then compute the RDCs and RMSD from the new NH directions.
     @param pdbVec a vector of Pdb object corresponding to the structural fragment.
     @param rdcVec1 a vector of RDC values from medium 1
     @param rdcVec2 a vector of RDC values from medium 2
     @param saupe   four diagnolized Saupe elements in the order of {Syy1, Szz1, Syy2, Szz2}
     @param rot     The relative rotation matrix between the 2 POFs.
     @param rmsds   for returning the RDC RMSDs in the order of RDC for medium 1 and medium 2
     @param resolution the resolution in the grid-search
     @param debug   if true, it will print the progress of the fitting
     @return the rotation matrix which minimizing the score function (RDCs)
     * We only search through 1/4 of the region to obtain one possible orientation, 
     * other three can be computed much faster by multiplying by one of the three constant 
     * matrices.
     */
    public Matrix eulerFit(final Vector pdbVec, final Vector rdcVec1, final Vector rdcVec2, 
			   double[] saupe, Matrix rot, double[] rmsds, double resolution, boolean debug){
	double alpha = 0.0, beta = 0.0, gamma = 0.0;
	double alphaS = 0.0, betaS = 0.0, gammaS = 0.0;
	int i, j, k, no;
	int index = 0;
	Matrix A = new Matrix(3,3);
	Cartesian cc = new Cartesian();
	Vector atomVec = new Vector();
	Pdb pp = new Pdb();
	String atom = "", resid = "";

	Dipolar dd = new Dipolar();
	double rdc1Exp = 0.0;
	double rdc2Exp = 0.0;
	Vector pdbRdcVec = new Vector();
	// an array with three NH directional cosines and the last element as RDC value
	int N = rdcVec1.size();  //the number of RDC data
	double [][] nhRdcArr = new double[N][5]; 
	double [] rdcArr1 = new double[N]; 
	double [] rdcArr2 = new double[N]; 
	double [] rdcArr1S =new double[N]; 
	double [] rdcArr2S =new double[N]; 
	double [] amide = new double[3];
	double [] nh = new double[3];
	double [] nh2 = new double[3];
	double [] nhB = new double[3];
	double [][] mat = new double[3][3];
	double r = 0.0, r1 = 0.0, r2 = 0.0; 
	double rT = 10000.0;
	double r1s = 0.0,  r2s = 0.0;
	double rdc1Cal = 0.0;
	double rdc2Cal = 0.0;
	//extract the NH vectors based on the available RDC values.
	for (i=0; i<N; i++){ 
	    dd = (Dipolar)rdcVec1.elementAt(i); 
	    rdc1Exp = dd.getRdc();
	    dd = (Dipolar)rdcVec2.elementAt(i); 
	    rdc2Exp = dd.getRdc();
	    no = dd.getResidueNo();
	    index = Collections.binarySearch(pdbVec, new Pdb(no), new Pdb.PdbComparator());

	    if (index > -1){
		pp = (Pdb)pdbVec.elementAt(index); 
		no = pp.getResidueNo();
		atomVec = pp.getAtomVec();
		for (j=0; j<atomVec.size(); j++){
		    cc = (Cartesian)atomVec.elementAt(j);
		    atom = cc.getAtom();
		    if (atom.equals("N"))
			amide = cc.getXYZ();
		    else if (atom.equals("H"))
			nh = cc.getXYZ();
		}
		nhRdcArr[i][0] =(nh[0] - amide[0]) / Const.dN2H; //the first three are direction cosines.
		nhRdcArr[i][1] =(nh[1] - amide[1]) / Const.dN2H;
		nhRdcArr[i][2] =(nh[2] - amide[2]) / Const.dN2H;
		nhRdcArr[i][3] = rdc1Exp;   //the last two elements are RDC values
		nhRdcArr[i][4] = rdc2Exp;
	    }
	}
	//Do the grid search ( 0 =< alpha < 180, 0 =< beta <180, 0 =< gamma < 180) 
	for (i = 0; i < (180/resolution); i++){
	    alpha = i * resolution * Math.PI / 180.0;
	    for (j=0; j< (180 / resolution); j++){
		beta = j * resolution * Math.PI / 180.0;
		for (k=0; k< (180 / resolution ); k++){
		    gamma = k * resolution * Math.PI / 180.0;
		    //Generate the Euler matrix
		    mat[0][0] = Math.cos(alpha)*Math.cos(beta)*Math.cos(gamma) - Math.sin(alpha)*Math.sin(gamma);
		    mat[0][1] = Math.sin(alpha)*Math.cos(beta)*Math.cos(gamma) + Math.cos(alpha)*Math.sin(gamma);
		    mat[0][2] = -Math.sin(beta)*Math.cos(gamma);
		    mat[1][0] = -Math.cos(alpha)*Math.cos(beta)*Math.sin(gamma) - Math.sin(alpha)*Math.cos(gamma);
		    mat[1][1] = -Math.sin(alpha)*Math.cos(beta)*Math.sin(gamma) + Math.cos(alpha)*Math.cos(gamma);
		    mat[1][2] = Math.sin(gamma)*Math.sin(beta);
		    mat[2][0] = Math.sin(beta)*Math.cos(alpha);
		    mat[2][1] = Math.sin(alpha)*Math.sin(beta);
		    mat[2][2] = Math.cos(beta);
		    A = new Matrix(mat);
		    r = 0.0;
		    r1 = 0.0; 
		    r2 = 0.0;
		    for (int m = 0; m < N; m++){
			nhB[0] = nhRdcArr[m][0];
			nhB[1] = nhRdcArr[m][1];
			nhB[2] = nhRdcArr[m][2];
   			nh = A.times(nhB); 
//  			nh2 = rot.times(nh);//rotate the NH vector for the second RDC data set
       			nh2 = nh;     // NOT rotate the NH vector for the second RDC data set
			rdc1Cal =  (nh[1] * nh[1] - nh[0] * nh[0]) * saupe[0] //(cosy*cosy-cosx*cosx)*Syy1 ...
			    + (nh[2] * nh[2] - nh[0] * nh[0]) * saupe[1]; //Szz1;
			rdc2Cal =  (nh2[1] * nh2[1] - nh2[0] * nh2[0]) * saupe[2] // Syy2
			    + (nh2[2] * nh2[2] - nh2[0] * nh2[0]) * saupe[3]; //Szz2;
			r1 += (rdc1Cal - nhRdcArr[m][3]) * (rdc1Cal - nhRdcArr[m][3]);
			r2 += (rdc2Cal - nhRdcArr[m][4]) * (rdc2Cal - nhRdcArr[m][4]);
			rdcArr1[m] = rdc1Cal;  
			rdcArr2[m] = rdc2Cal;
		    }
		    r = r1 + r2;
		    if( r < rT || Math.abs(r - rT) < Const.eps) {//finding all four possible sets of Euler angles.
			alphaS = alpha;
			betaS = beta;
			gammaS = gamma;
			r1s = r1;
			r2s = r2;
			rT = r;
			System.arraycopy(rdcArr1, 0, rdcArr1S, 0, N); //save the back computed RDC values
			System.arraycopy(rdcArr2, 0, rdcArr2S, 0, N);
			if (debug)
			    System.out.println(" rT = "+rT+"  "+r1+"  "+r2);
		    }
		}
	    }
	}
	if (debug){
	    System.out.println();
	    System.out.println("Rdc1: ");
	    System.out.println("back-computed RDCs,    measured RDCs");
	    for (i=0; i<N; i++){
		System.out.println(rdcArr1S[i]+"   "+nhRdcArr[i][3]);
	    }
	    System.out.println("RMSD = "+ Math.sqrt(r1s/N));
	    System.out.println();
	    System.out.println("Rdc2: ");
	    System.out.println("back-computed RDCs,    measured RDCs");
	    for (i=0; i<N; i++){
		System.out.println(rdcArr2S[i]+"   "+nhRdcArr[i][4]);
	    }
	    System.out.println("RMSD = "+ Math.sqrt(r2s/N));
	}
	rmsds[0] = r1s;
	rmsds[1] = r2s;
	Matrix mm = A.eulerMat(alphaS, betaS, gammaS);
	return 	mm;
    }

  /** 
     * Compute three Euler angles using two sets of RDC data and the
     * corresponding Saupe elements and H-bonds between the two strands.
     * Basic steps: extract the NH
     * vectors with the corresponding RDC data as an array (for
     * efficiency) For all the values of three Euler angles first
     * compute the new NH directions after the rotation then compute
     * the RDCs and RMSD from the new NH directions.
     @param pdbVec   a vector of Pdb object corresponding to the structural fragment.
     @param rdcVec1  a vector of RDC values from medium 1
     @param rdcVec2  a vector of RDC values from medium 2
     @param saupe    four diagnolized Saupe elements in the order of {Syy1, Szz1, Syy2, Szz2}
     @param rot      The relative rotation matrix between the 2 POFs.
     @param rmsds    saving the RDC RMSDs and hbScore in the order: rmsdRDC1, rmsdRDC2, hbScore.
     @param resolution the resolution in grid search on angle space
     @param weight4HB relative weight for the H-bond in the score function
     @param pdbVecS1  the H-bond parter of current fragment being searched
     @param hbVec  the identity of H-bonds
     @param debug if true will print the progress of the search
     @return the rotation matrix which minimizing the score function (RDCs)
     */
    public Matrix eulerFitHB(final Vector pdbVec, final Vector rdcVec1,final Vector rdcVec2, double[] saupe, Matrix rot, 
			     double[] rmsds, double resolution, double weight4HB, Vector pdbVecS1, Vector hbVec,
			     boolean debug){
	double alpha = 0.0, beta = 0.0, gamma = 0.0;
	double alphaS = 0.0, betaS = 0.0, gammaS = 0.0;
	int i, j, k, no;
	int index = 0;
	Matrix A = new Matrix(3,3);
	Cartesian cc = new Cartesian();
	Vector atomVec = new Vector();
	Pdb pp = new Pdb();
	String atom = "", resid = "";

	Dipolar dd = new Dipolar();
	double rdc1Exp = 0.0;
	double rdc2Exp = 0.0;
	Vector pdbRdcVec = new Vector();
	// an array with three NH directional cosines and the last element as RDC value
	int N = rdcVec1.size();  //the number of RDC data
	double [][] nhRdcArr = new double[N][5]; 
	double [] rdcArr1 = new double[N]; 
	double [] rdcArr2 = new double[N]; 
	double [] rdcArr1S =new double[N]; 
	double [] rdcArr2S =new double[N]; 
	double [] amide = new double[3];
	double [] nh = new double[3];
	double [] nh2 = new double[3];
	double [] nhB = new double[3];
	double [][] mat = new double[3][3];
	double r = 0.0, r1 = 0.0, r2 = 0.0; 
	double rT = 10000.0;
	double r1s = 0.0,  r2s = 0.0;
	double rdc1Cal = 0.0;
	double rdc2Cal = 0.0;
	Vector pdbVecN = new Vector();
	double hScore = 0.0, hScoreS = 0.0;
	PdbRmsd pR = new PdbRmsd();
	//extract the NH vectors based on the available RDC values.
	for (i=0; i<N; i++){ 
	    dd = (Dipolar)rdcVec1.elementAt(i); 
	    rdc1Exp = dd.getRdc();
	    dd = (Dipolar)rdcVec2.elementAt(i); 
	    rdc2Exp = dd.getRdc();
	    no = dd.getResidueNo();
	    index = Collections.binarySearch(pdbVec, new Pdb(no), new Pdb.PdbComparator());

	    if (index > -1){
		pp = (Pdb)pdbVec.elementAt(index); 
		no = pp.getResidueNo();
		atomVec = pp.getAtomVec();
		for (j=0; j<atomVec.size(); j++){
		    cc = (Cartesian)atomVec.elementAt(j);
		    atom = cc.getAtom();
		    if (atom.equals("N"))
			amide = cc.getXYZ();
		    else if (atom.equals("H"))
			nh = cc.getXYZ();
		}
		nhRdcArr[i][0] =(nh[0] - amide[0]) / Const.dN2H; //the first three are direction cosines.
		nhRdcArr[i][1] =(nh[1] - amide[1]) / Const.dN2H;
		nhRdcArr[i][2] =(nh[2] - amide[2]) / Const.dN2H;
		nhRdcArr[i][3] = rdc1Exp;   //the last two elements are RDC values
		nhRdcArr[i][4] = rdc2Exp;
	    }
	}
	//Do the grid search ( 0 =< alpha <360, 0 =< beta <180, 0 =< gamma <360) 
	if(debug){
	    System.out.println(" totalScore, rmsdRDC1,  rmsdRDC2,  hbScore");
	}

	for (i = 0; i < (180/resolution); i++){
	    alpha = i * resolution * Math.PI / 180.0;
	    for (j=0; j< (180 / resolution); j++){
		beta = j * resolution * Math.PI / 180.0;
		for (k=0; k< (180 / resolution ); k++){
		    gamma = k * resolution * Math.PI / 180.0;
		    //Generate the Euler matrix
		    mat[0][0] = Math.cos(alpha)*Math.cos(beta)*Math.cos(gamma) - Math.sin(alpha)*Math.sin(gamma);
		    mat[0][1] = Math.sin(alpha)*Math.cos(beta)*Math.cos(gamma) + Math.cos(alpha)*Math.sin(gamma);
		    mat[0][2] = -Math.sin(beta)*Math.cos(gamma);
		    mat[1][0] = -Math.cos(alpha)*Math.cos(beta)*Math.sin(gamma) - Math.sin(alpha)*Math.cos(gamma);
		    mat[1][1] = -Math.sin(alpha)*Math.cos(beta)*Math.sin(gamma) + Math.cos(alpha)*Math.cos(gamma);
		    mat[1][2] = Math.sin(gamma)*Math.sin(beta);
		    mat[2][0] = Math.sin(beta)*Math.cos(alpha);
		    mat[2][1] = Math.sin(alpha)*Math.sin(beta);
		    mat[2][2] = Math.cos(beta);
		    A = new Matrix(mat);
		    r = 0.0;
		    r1 = 0.0; 
		    r2 = 0.0;
		    for (int m = 0; m < N; m++){
			nhB[0] = nhRdcArr[m][0];
			nhB[1] = nhRdcArr[m][1];
			nhB[2] = nhRdcArr[m][2];
   			nh = A.times(nhB); 
// 			nh2 = rot.times(nh);  //rotate the NH vector for the second RDC data set
     			nh2 = nh;             // NOT rotate the NH vector for the second RDC data set
			rdc1Cal =  (nh[1] * nh[1] - nh[0] * nh[0]) * saupe[0] //Syy1//(cosy*cosy-cosx*cosx)*Syy1 ...
			    + (nh[2] * nh[2] - nh[0] * nh[0]) * saupe[1]; //Szz1;
			rdc2Cal =  (nh2[1] * nh2[1] - nh2[0] * nh2[0]) * saupe[2] // Syy2
			    + (nh2[2] * nh2[2] - nh2[0] * nh2[0]) * saupe[3]; //Szz2;
			r1 += (rdc1Cal - nhRdcArr[m][3]) * (rdc1Cal - nhRdcArr[m][3]);
			r2 += (rdc2Cal - nhRdcArr[m][4]) * (rdc2Cal - nhRdcArr[m][4]);
			rdcArr1[m] = rdc1Cal;  
			rdcArr2[m] = rdc2Cal;
		    }
		    pdbVecN = pp.newPdb(pdbVec, A);
		    hScore = pR.centerFit(pdbVecS1, pdbVecN, hbVec, false);
		    r = weight4HB * hScore + Math.sqrt(r1 / N ) +  Math.sqrt(r2 / N );
		    if( r < rT || Math.abs(r - rT) < Const.eps) {//finding all four possible sets of Euler angles.
			alphaS = alpha;
			betaS = beta;
			gammaS = gamma;
			r1s = r1;
			r2s = r2;
			hScoreS = hScore;
			rT = r;
			if(debug){
			    System.arraycopy(rdcArr1, 0, rdcArr1S, 0, N); //save the back computed RDC values
			    System.arraycopy(rdcArr2, 0, rdcArr2S, 0, N);
			    System.out.println(" rT = "+rT+"  "+r1+"  "+r2+"  "+hScore);
			}
		    }
		}
	    }
	}
	if (debug){
	    System.out.println();
	    System.out.println("Rdc1: ");
	    System.out.println("back-computed RDCs,    measured RDCs");
	    for (i=0; i<N; i++){
		System.out.println(rdcArr1S[i]+"   "+nhRdcArr[i][3]);
	    }
	    System.out.println("RDC1 RMSD = "+ Math.sqrt(r1s/N));
	    System.out.println();
	    System.out.println("Rdc2: ");
	    System.out.println("back-computed RDCs,    measured RDCs");
	    for (i=0; i<N; i++){
		System.out.println(rdcArr2S[i]+"   "+nhRdcArr[i][4]);
	    }
	    System.out.println("RDC2 RMSD = "+ Math.sqrt(r2s/N));
	}
	rmsds[0] = r1s;
	rmsds[1] = r2s;
	rmsds[2] = hScoreS;
	Matrix mm = A.eulerMat(alphaS, betaS, gammaS);
	return 	mm;
    }

    /** private utility routines for checking the SVD method **/
    /** Check magnitude of difference of scalars. **/
    private static void check(double x, double y) {
	double eps = Math.pow(2.0,-52.0);
	if (x == 0 & Math.abs(y) < 10*eps) return;
	if (y == 0 & Math.abs(x) < 10*eps) return;
	if (Math.abs(x-y) > 10*eps*Math.max(Math.abs(x), Math.abs(y))) {
	    throw new RuntimeException("The difference x-y is too large: x = " 
				       + Double.toString(x) + "  y = " + Double.toString(y));
	}
    }
    /** Check norm of difference of "vectors". **/
    private static void check(double[] x, double[] y) {
	if (x.length == y.length ) {
	    for (int i=0;i<x.length;i++) {
		check(x[i],y[i]);
	    } 
	} else {
	    throw new RuntimeException("Attempt to compare vectors of different lengths");
	}
    }

    /** Check norm of difference of arrays. **/
   private static void check(double[][] x, double[][] y) {
      Matrix A = new Matrix(x);
      Matrix B = new Matrix(y);
      check(A,B);
   }

   /** Check norm of difference of Matrices. **/
   private static void check(Matrix X, Matrix Y) {
      double eps = Math.pow(2.0,-52.0);
      if (X.norm1() == 0. & Y.norm1() < 10*eps) return;
      if (Y.norm1() == 0. & X.norm1() < 10*eps) return;
      if (X.minus(Y).norm1() > 1000*eps*Math.max(X.norm1(),Y.norm1())) {
         throw new RuntimeException("The norm of (X-Y) is too large: " +  Double.toString(X.minus(Y).norm1()));
      }
   }

    /** Shorten spelling of print. **/
    private static void print (String s) {
	System.out.print(s);
    }

    /** Print appropriate messages for successful outcome try **/
    private static void try_success (String s,String e) {
	print(">    " + s + "success\n");
	if ( e != "" ) {
	    print(">      Message: " + e + "\n");
	}
    }
    /** Print appropriate messages for unsuccessful outcome try **/
    private static int try_failure (int count, String s,String e) {
	print(">    " + s + "*** failure ***\n>      Message: " + e + "\n");
	return ++count;
    }
}
