/*****************************************************************************************
 * PdbRmsd.java

 * Version:           0.1

 * Description: 

 * A class implement the method for Computing the Hbonds between two
 * strands and determining the relative positions of a helix (25-34)
 * and a \beta sheet based on the NOE distances between them. called 
 * PdbRmsd since the class is coded, originally, for computing RMSDs
 * between two structures.

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;

public class PdbRmsd {
    private Vector pdbVecA = new Vector();
    private Vector pdbVecB = new Vector();
    static final String[] atomNameArrA = {"N", "C"}; //atoms compared
    static final String[] atomNameArrB = {"N", "C"};  //A and B can, in principle, be different
    public PdbRmsd(){
	pdbVecA = new Vector();
	pdbVecB = new Vector();
    }
    public Vector getPdbVecA(){
	return pdbVecA;
    }
    public Vector getPdbVecB(){
	return pdbVecB;
    } 
    /** calculate the angles between two vectors v1 and v2 
     * The returned angle is in the [0,Pi] range or the cosine itself.
     * or just return the interAngle cosine
     */
    public double interAngle(double[] v1, double[] v2){
	double v1Len = Math.sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]);
	double v2Len = Math.sqrt(v2[0]*v2[0] + v2[1]*v2[1] + v2[2]*v2[2]);
	return ( (v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]) / (v1Len * v2Len) );
//  	return Math.acos(c);  //acos [0, PI] so should be OK
    }

    /** cal the directioal cosines of the vector vec */
    public double[] dirCos(double [] vec) {
	double len = 0.0;
	double [] dirs = new double[vec.length];
	for (int i=0; i<vec.length; i++)
	    len += vec[i] * vec[i];
	for (int j=0; j<vec.length; j++)
	    dirs[j] = vec[j] / Math.sqrt(len);
	return dirs;
    }

    /**
     * compute the center of a PDB object (the 3D coordinates)
     */
    public double[] centerCal(Vector pdbVec){
	int i, j, k;
	Pdb pp = new Pdb();
	Cartesian cc = new Cartesian();
	Vector atomVec = new Vector();
	double xC1=0.0, yC1=0.0, zC1=0.0;
	int mm = pdbVec.size();
	int m = 0;
	String atom = "";
	for (i=0; i<pdbVec.size(); i++){
	    pp = (Pdb)pdbVec.elementAt(i); 
	    atomVec = pp.getAtomVec();
	    for (j=0; j<atomVec.size(); j++){ //save all those of B with atom Name as in the array
		cc = (Cartesian)atomVec.elementAt(j);
		atom = cc.getAtom();
		for (k =0; k< atomNameArrA.length; k++){ //since this is supposely very short so linear search OK
		    if (atom.equals(atomNameArrA[k])){
			cc = (Cartesian)atomVec.elementAt(j);
			xC1 += cc.getX(); 
			yC1 += cc.getY();
			zC1 += cc.getZ();
			m++;
		    }
		}
	    }
	}
	// Compute the center
	if (m > 0){
	    xC1 /= m;
	    yC1 /= m;
	    zC1 /= m;
	}
	return new double[]{xC1, yC1, zC1};
    }

    /**
     * compute the center of a PDB coordinates when the coordinates are represented as an array of nx3
     */
    public double[] centerCal(double[][]  pdbVec){
	double xC1=0.0, yC1=0.0, zC1=0.0;
	int mm = pdbVec.length;
	for (int i=0; i<mm; i++){
	    xC1 += pdbVec[i][0];
	    yC1 += pdbVec[i][1];
	    zC1 += pdbVec[i][2];
	}
	// Compute the center
	if (mm > 0){
	    xC1 /= mm;
	    yC1 /= mm;
	    zC1 /= mm;
	}
	return new double[]{xC1, yC1, zC1};
    }

    /**
     *compute the H-bond energy between the two paired strands.
     @param pdbVecA     strand A
     @param pdbVecB     strand B, A's partner
     @param hbVec       the vector of Hbond information
     @param debug if true it print the information for debuging
    */
    public double centerFit(Vector pdbVecA, Vector pdbVecB, Vector hbVec, boolean debug){
	boolean debug2 = false;
	final int  noOfAtoms = hbVec.size(); 
	int i, j, k;
	//One pair (N->H...O or O...H<-N), their partner's coordinates are computed based on perfect H-bond.
	double[][] hPartnerOfA = new double[noOfAtoms][3]; //Coodinates of its H-bond partner,
	double[][] coordOfA    = new double[noOfAtoms][3]; //Coodinates of atoms N and O, in the order of N, O
	double[][] dirsOfA     = new double[noOfAtoms][3]; //and directional cosines

	double[][] hPartnerOfB = new double[noOfAtoms][3]; //the corresponding ones for strand B
	double[][] coordOfB    = new double[noOfAtoms][3]; 
	double[][] dirsOfB     = new double[noOfAtoms][3]; 

	Pdb pp = new Pdb();
	Cartesian cc = new Cartesian();
	Vector atomVec = new Vector();
	String atom = "";
	double [] amide = new double[3];
	double [] nh = new double[3];
	double [] n2NHVec  = new double[3];
	double [] co = new double[3];
	double [] o  = new double[3];
	double [] co2OVec = new double[3];
	int residueNo = 0;
	int residueNoA;   //the Sequence No of one of the two partners of the NOE
	int residueNoB;   //the Sequence No of Another partner of the NOE
	String atomA;     //the name of Atom A
	String atomB;     //the name of Atom B
	String type;      //the name of A
	Hbond hb = new  Hbond();
	//Extract atoms NH2O and O2HN of strandA, and compute the directions of N2H and CO2O for both strand A and B
	boolean isN = false;
	int index  =  -1;
	for (i = 0; i < noOfAtoms; i++){
	    hb = (Hbond)hbVec.elementAt(i);
	    residueNoA = hb.getResidueNoA();
	    atomA = hb.getAtomA();
	    index = Collections.binarySearch(pdbVecA, new Pdb(residueNoA), new Pdb.PdbComparator());
	    //Extract the coordinates and direction of the H-bond partners, one partner:
	    if ( index > -1 ){
		pp = (Pdb)pdbVecA.elementAt(index);
		residueNo = pp.getResidueNo();
		atomVec = pp.getAtomVec();
		if  (atomA.equals("O")){
		    for (j=0; j<atomVec.size(); j++){
			cc = (Cartesian)atomVec.elementAt(j);
			atom = cc.getAtom();
			if (atom.equals("C"))
			    co = cc.getXYZ();
			else if (atom.equals("O")){
			    o = cc.getXYZ();
			    coordOfA[i] = new double[]{o[0], o[1], o[2]};
			}
		    }
		    //The coordinate of the H-bond partner of the atom O 
		    hPartnerOfA[i] = new double[]{o[0] + (o[0] - co[0]) * Const.dN2COHBond / Const.dCO2O, 
						  o[1] + (o[1] - co[1]) * Const.dN2COHBond / Const.dCO2O,
						  o[2] + (o[2] - co[2]) * Const.dN2COHBond / Const.dCO2O};
		    dirsOfA[i] = new double[]{o[0] - co[0],  o[1]-co[1], o[2]-co[2]};
		    if (debug2){
			System.out.println(residueNoA);
			System.out.println("O: "+i+"  "+coordOfA[i][0]+"  "+coordOfA[i][1]+"   "+coordOfA[i][2]);
			System.out.println("A: "+i+"  "+dirsOfA[i][0]+"  "+dirsOfA[i][1]+"   "+dirsOfA[i][2]);
		    }
		} else if (atomA.equals("N")) {
		    for (j=0; j<atomVec.size(); j++){
			cc = (Cartesian)atomVec.elementAt(j);
			atom = cc.getAtom();
			if (atom.equals("N")){
			    amide = cc.getXYZ();
			    coordOfA[i] = new double[]{amide[0], amide[1], amide[2]};
			}else if (atom.equals("H"))
			    nh = cc.getXYZ();
		    }
		    //The coordinate of the H-bond partner of the atom N
		    hPartnerOfA[i] = new double[]{amide[0] + (nh[0] - amide[0]) * Const.dN2COHBond / Const.dN2H, 
						  amide[1] + (nh[1] - amide[1]) * Const.dN2COHBond / Const.dN2H, 
						  amide[2] + (nh[2] - amide[2]) * Const.dN2COHBond / Const.dN2H};
		    dirsOfA[i] = new double[]{nh[0]-amide[0], nh[1]-amide[1], nh[2]-amide[2]};
		    if (debug2){
			System.out.println(residueNoA);
			System.out.println("N: "+i+"  "+coordOfA[i][0]+"  "+coordOfA[i][1]+"   "+coordOfA[i][2]);
			System.out.println("A: "+i+"  "+dirsOfA[i][0]+"  "+dirsOfA[i][1]+"   "+dirsOfA[i][2]);
		    }
		}
	    }
	    //Extract the coordinates and direction of the H-bond partners, the corresponding partner:
	    residueNoB = hb.getResidueNoB();
	    atomB = hb.getAtomB();
	    index = Collections.binarySearch(pdbVecB, new Pdb(residueNoB), new Pdb.PdbComparator());
	    if ( index > -1 ){
		pp = (Pdb)pdbVecB.elementAt(index);
		residueNo = pp.getResidueNo();
		atomVec = pp.getAtomVec();
		if  (atomB.equals("O")){
		    for (j=0; j<atomVec.size(); j++){
			cc = (Cartesian)atomVec.elementAt(j);
			atom = cc.getAtom();
			if (atom.equals("C"))
			    co = cc.getXYZ();
			else if (atom.equals("O")){
			    o = cc.getXYZ();
			    coordOfB[i] = new double[]{o[0], o[1], o[2]};
			}
		    }
		    dirsOfB[i] = new double[]{o[0] - co[0], o[1]-co[1], o[2]-co[2]};
		    if (debug2){
			System.out.println(residueNoB);
			System.out.println("O: "+i+"  "+coordOfB[i][0]+" : "+coordOfB[i][1]+"   "+coordOfB[i][2]);
		    }
		}else if (atomB.equals("N")) {
		    for (j=0; j<atomVec.size(); j++){
			cc = (Cartesian)atomVec.elementAt(j);
			atom = cc.getAtom();
			if (atom.equals("N")){
			    amide = cc.getXYZ();
			    coordOfB[i] = new double[]{amide[0], amide[1], amide[2]};
			}else if (atom.equals("H"))
			    nh = cc.getXYZ();
		    }
		    //The coordinate of the atom O of the H-bond partner of the NH vector
		    hPartnerOfB[i] = new double[]{amide[0] + (nh[0] - amide[0]) * Const.dN2COHBond / Const.dN2H, 
						  amide[1] + (nh[1] - amide[1]) * Const.dN2COHBond / Const.dN2H, 
						  amide[2] + (nh[2] - amide[2]) * Const.dN2COHBond / Const.dN2H};
		    dirsOfB[i] = new double[]{nh[0]-amide[0], nh[1]-amide[1], nh[2]-amide[2]};
		    if (debug2){
			System.out.println(residueNoB);
			System.out.println("N: "+i+"  "+coordOfB[i][0]+" : "+coordOfB[i][1]+"   "+coordOfB[i][2]);
		    }
		}
	    }
	}
	double[] centerOfPartnersOfA = centerCal(hPartnerOfA); //the center of the coords of the H-bond partners of A.
	double x1, y1, z1, x2, y2, z2;  //for compute the H-bond distances.
	double disHBond;

	double[] centerOfB = centerCal(coordOfB); //the center of the coords of B.
	double[] cT={centerOfPartnersOfA[0]-centerOfB[0], centerOfPartnersOfA[1]-centerOfB[1], centerOfPartnersOfA[2]-centerOfB[2]};
	Vector pdbVec2N = pp.newPdbByTranslation(pdbVecB, cT); //the new coordinate for fragment B after the fitting
	double cosAngle, disRms = 0.0, angleRms = 0.0;
	double[][] coordOfBN  = new double[noOfAtoms][3]; 

	for (i=0; i<coordOfB.length; i++){
	    coordOfBN[i] = new double[]{coordOfB[i][0]+cT[0], coordOfB[i][1]+cT[1], coordOfB[i][2]+cT[2]};
 	    x2 = coordOfBN[i][0];   
 	    y2 = coordOfBN[i][1];
 	    z2 = coordOfBN[i][2];
   	    x1 = coordOfA[i][0];   
   	    y1 = coordOfA[i][1];
   	    z1 = coordOfA[i][2];
  	    disHBond =  Math.sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
	    disRms += (disHBond - Const.dN2COHBond) * (disHBond - Const.dN2COHBond);  //deviation from the ideal value
	    cosAngle = interAngle(dirsOfA[i], dirsOfB[i]);
	    angleRms += (cosAngle + 1.0) * (cosAngle + 1.0); //- (-1.0) since if they are colinear
	    if (debug)
		System.out.println("Distance = "+ disHBond +"  Angle = "+cosAngle);  //distance between N and O of an H-bond
	}
	if (noOfAtoms > 0){
	    angleRms = Math.sqrt(angleRms / coordOfB.length ); 
	    disRms   = Math.sqrt(disRms / coordOfB.length ); 
	}
	if (debug)
	    System.out.println(disRms+"  "+angleRms);
	return (angleRms+disRms);
    }

  

    /**
     *compute the H-bond energy term between the two paired strands.
     *Identical to the centerFit method except that the newly paired
     *Pdb vector are returned
     @param pdbVecA     strand A
     @param pdbVecB     strand B, A's partner
     @param hbVec       the vector of Hbond information
     @return the translated and paired Pdb vector
    */
    public Vector centerFitVec(Vector pdbVecA, Vector pdbVecB, Vector hbVec, boolean debug){
	boolean debug2 = false;
	final int  noOfAtoms = hbVec.size(); 
	int i, j, k;
	//One pair (N->H...O or O...H<-N), their partner's coordinates are computed based on perfect H-bond.
	double[][] hPartnerOfA = new double[noOfAtoms][3]; //Coodinates of its H-bond partner,
	double[][] coordOfA    = new double[noOfAtoms][3]; //Coodinates of atoms N and O, in the order of N, O
	double[][] dirsOfA     = new double[noOfAtoms][3]; //and directional cosines

	double[][] hPartnerOfB = new double[noOfAtoms][3]; //the corresponding ones for strand B
	double[][] coordOfB    = new double[noOfAtoms][3]; 
	double[][] dirsOfB     = new double[noOfAtoms][3]; 

	Pdb pp = new Pdb();
	Cartesian cc = new Cartesian();
	Vector atomVec = new Vector();
	String atom = "";
	double [] amide = new double[3];
	double [] nh = new double[3];
	double [] n2NHVec  = new double[3];
	double [] co = new double[3];
	double [] o  = new double[3];
	double [] co2OVec = new double[3];
	int residueNo = 0;
	int residueNoA;   //the Sequence No of one of the two partners of the NOE
	int residueNoB;   //the Sequence No of Another partner of the NOE
	String atomA;     //the name of Atom A
	String atomB;     //the name of Atom B
	String type;      //the name of A
	Hbond hb = new  Hbond();
	//Extract atoms NH2O and O2HN of strandA, and compute the directions of N2H and CO2O for both strand A and B
	boolean isN = false;
	int index  =  -1;
	for (i = 0; i < noOfAtoms; i++){
	    hb = (Hbond)hbVec.elementAt(i);
	    residueNoA = hb.getResidueNoA();
	    atomA = hb.getAtomA();
	    index = Collections.binarySearch(pdbVecA, new Pdb(residueNoA), new Pdb.PdbComparator());
	    //Extract the coordinates and direction of the H-bond partners, one partner:
	    if ( index > -1 ){
		pp = (Pdb)pdbVecA.elementAt(index);
		residueNo = pp.getResidueNo();
		atomVec = pp.getAtomVec();
		if  (atomA.equals("O")){
		    for (j=0; j<atomVec.size(); j++){
			cc = (Cartesian)atomVec.elementAt(j);
			atom = cc.getAtom();
			if (atom.equals("C"))
			    co = cc.getXYZ();
			else if (atom.equals("O")){
			    o = cc.getXYZ();
			    coordOfA[i] = new double[]{o[0], o[1], o[2]};
			}
		    }
		    //The coordinate of the H-bond partner of the atom O 
		    hPartnerOfA[i] = new double[]{o[0] + (o[0] - co[0]) * Const.dN2COHBond / Const.dCO2O, 
						  o[1] + (o[1] - co[1]) * Const.dN2COHBond / Const.dCO2O,
						  o[2] + (o[2] - co[2]) * Const.dN2COHBond / Const.dCO2O};
		    dirsOfA[i] = new double[]{o[0] - co[0],  o[1]-co[1], o[2]-co[2]};
		    if (debug2){
			System.out.println(residueNoA);
			System.out.println("O: "+i+"  "+coordOfA[i][0]+"  "+coordOfA[i][1]+"   "+coordOfA[i][2]);
			System.out.println("A: "+i+"  "+dirsOfA[i][0]+"  "+dirsOfA[i][1]+"   "+dirsOfA[i][2]);
		    }
		} else if (atomA.equals("N")) {
		    for (j=0; j<atomVec.size(); j++){
			cc = (Cartesian)atomVec.elementAt(j);
			atom = cc.getAtom();
			if (atom.equals("N")){
			    amide = cc.getXYZ();
			    coordOfA[i] = new double[]{amide[0], amide[1], amide[2]};
			}else if (atom.equals("H"))
			    nh = cc.getXYZ();
		    }
		    //The coordinate of the H-bond partner of the atom N
		    hPartnerOfA[i] = new double[]{amide[0] + (nh[0] - amide[0]) * Const.dN2COHBond / Const.dN2H, 
						  amide[1] + (nh[1] - amide[1]) * Const.dN2COHBond / Const.dN2H, 
						  amide[2] + (nh[2] - amide[2]) * Const.dN2COHBond / Const.dN2H};
		    dirsOfA[i] = new double[]{nh[0]-amide[0], nh[1]-amide[1], nh[2]-amide[2]};
		    if (debug2){
			System.out.println(residueNoA);
			System.out.println("N: "+i+"  "+coordOfA[i][0]+"  "+coordOfA[i][1]+"   "+coordOfA[i][2]);
			System.out.println("A: "+i+"  "+dirsOfA[i][0]+"  "+dirsOfA[i][1]+"   "+dirsOfA[i][2]);
		    }
		}
	    }
	    //Extract the coordinates and direction of the H-bond partners, the corresponding partner:
	    residueNoB = hb.getResidueNoB();
	    atomB = hb.getAtomB();
	    index = Collections.binarySearch(pdbVecB, new Pdb(residueNoB), new Pdb.PdbComparator());
	    if ( index > -1 ){
		pp = (Pdb)pdbVecB.elementAt(index);
		residueNo = pp.getResidueNo();
		atomVec = pp.getAtomVec();
		if  (atomB.equals("O")){
		    for (j=0; j<atomVec.size(); j++){
			cc = (Cartesian)atomVec.elementAt(j);
			atom = cc.getAtom();
			if (atom.equals("C"))
			    co = cc.getXYZ();
			else if (atom.equals("O")){
			    o = cc.getXYZ();
			    coordOfB[i] = new double[]{o[0], o[1], o[2]};
			}
		    }
		    dirsOfB[i] = new double[]{o[0] - co[0], o[1]-co[1], o[2]-co[2]};
		    if (debug2){
			System.out.println(residueNoB);
			System.out.println("O: "+i+"  "+coordOfB[i][0]+" : "+coordOfB[i][1]+"   "+coordOfB[i][2]);
		    }
		}else if (atomB.equals("N")) {
		    for (j=0; j<atomVec.size(); j++){
			cc = (Cartesian)atomVec.elementAt(j);
			atom = cc.getAtom();
			if (atom.equals("N")){
			    amide = cc.getXYZ();
			    coordOfB[i] = new double[]{amide[0], amide[1], amide[2]};
			}else if (atom.equals("H"))
			    nh = cc.getXYZ();
		    }
		    //The coordinate of the atom O of the H-bond partner of the NH vector
		    hPartnerOfB[i] = new double[]{amide[0] + (nh[0] - amide[0]) * Const.dN2COHBond / Const.dN2H, 
						  amide[1] + (nh[1] - amide[1]) * Const.dN2COHBond / Const.dN2H, 
						  amide[2] + (nh[2] - amide[2]) * Const.dN2COHBond / Const.dN2H};
		    dirsOfB[i] = new double[]{nh[0]-amide[0], nh[1]-amide[1], nh[2]-amide[2]};
		    if (debug2){
			System.out.println(residueNoB);
			System.out.println("N: "+i+"  "+coordOfB[i][0]+" : "+coordOfB[i][1]+"   "+coordOfB[i][2]);
		    }
		}
	    }
	}
	double[] centerOfPartnersOfA = centerCal(hPartnerOfA); //the center of the coords of the H-bond partners of A.
	double x1, y1, z1, x2, y2, z2;  //for compute the H-bond distances.
	double disHBond;

	double[] centerOfB = centerCal(coordOfB); //the center of the coords of B.
	double[] cT1={centerOfPartnersOfA[0]-centerOfB[0], centerOfPartnersOfA[1]-centerOfB[1], centerOfPartnersOfA[2]-centerOfB[2]};
	double cosAngle = 0.0, disRms = 0.0, angleRms = 0.0;
	double[][] coordOfBN  = new double[noOfAtoms][3]; 
	double [] cT = new double[3];
	double [] cTSave = new double[3];
	long seed = 9752864; //419057923;
	Random rr = new Random(seed);
	int nCycle = 1;//4096;
	double rT = 0, rTotal = 100.0;
	double 	disRmsS = 0.0;
	double 	angleRmsS = 0.0;
	double range = 0.0;
	for (int mm = 0; mm< nCycle; mm++){
	    disRms = 0.0;
	    angleRms = 0.0;
	    cT[0] = cT1[0]+  range * (0.50-rr.nextDouble());
	    cT[1] = cT1[1]+  range * (0.50-rr.nextDouble());
	    cT[2] = cT1[2]+  range * (0.50-rr.nextDouble());
	    for (i=0; i<coordOfB.length; i++){
		coordOfBN[i] = new double[]{coordOfB[i][0]+cT[0], coordOfB[i][1]+cT[1], coordOfB[i][2]+cT[2]};
		x2 = coordOfBN[i][0];   
		y2 = coordOfBN[i][1];
		z2 = coordOfBN[i][2];
		x1 = coordOfA[i][0];   
		y1 = coordOfA[i][1];
		z1 = coordOfA[i][2];
		disHBond =  Math.sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
		disRms += (disHBond - Const.dN2COHBond) * (disHBond - Const.dN2COHBond);  //deviation from the ideal value
		cosAngle = interAngle(dirsOfA[i], dirsOfB[i]);
		angleRms += (cosAngle + 1.0) * (cosAngle + 1.0); //- (-1.0) since if they are colinear
	    }
	    rT = disRms;
	    if (rT < rTotal){
		if (debug)
		    System.out.println("DisRms= "+Math.sqrt(disRms/coordOfB.length )+"; AngleRms="+Math.sqrt(angleRms/coordOfB.length));
		rTotal = rT;
		System.arraycopy(cT, 0, cTSave, 0, 3);
		disRmsS = disRms;
		angleRmsS = angleRms;
	    }
	}

	for (i=0; i<coordOfB.length; i++){
	    coordOfBN[i] = new double[]{coordOfB[i][0]+cTSave[0], coordOfB[i][1]+cTSave[1], coordOfB[i][2]+cTSave[2]};
 	    x2 = coordOfBN[i][0];   
 	    y2 = coordOfBN[i][1];
 	    z2 = coordOfBN[i][2];
   	    x1 = coordOfA[i][0];   
   	    y1 = coordOfA[i][1];
   	    z1 = coordOfA[i][2];
  	    disHBond =  Math.sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
	    disRms += (disHBond - Const.dN2COHBond) * (disHBond - Const.dN2COHBond);  //deviation from the ideal value
	    cosAngle = interAngle(dirsOfA[i], dirsOfB[i]);
	    angleRms += (cosAngle + 1.0) * (cosAngle + 1.0); //- (-1.0) since if they are colinear
	    if (debug)
		System.out.println("Dis= "+ disHBond +"  Angle = "+cosAngle);  //distance between N and O of an H-bond
	}
	if (noOfAtoms > 0){
	    angleRms = Math.sqrt(angleRmsS / coordOfB.length ); 
	    disRms   = Math.sqrt(disRmsS / coordOfB.length ); 
	}
	Vector pdbVec2N = pp.newPdbByTranslation(pdbVecB, cTSave); //the new coordinate for fragment B after the fitting
	if (debug)
	    System.out.println(disRms+"  "+angleRms);
	return pdbVec2N;
    }

  
    /** 
     * Compute the relative positions between two 2ary oriented elements based on NOEs
     @param noeVec the NOE distance restraints between the fragment A and B
     @param pdbVecA one fragment
     @param pdbVecB another fragment
     @param noeRms saving the NOE rmsd
     @param debugNOE if true it will print out the NOE fitting Rmsds
     @return the Pdb vector for translated fragment which is B while
     *     Pdb vector for A is NOT changed
     */
    public Vector positionByNOE(final Vector noeVec, Vector pdbVecA, Vector pdbVecB, double[] noeRms, boolean debugNOE){
	//Extract neccessary information from the NOE table.
	int i, j;
	int noA, noB;
	int index = -1;
	String atom  = "";
	String atomA = "";
	String atomB = "";
	Vector atomVec = new Vector();
	Cartesian cc = new Cartesian();
	Pdb pp = new Pdb();
	int sizeA = 0;
	int sizeB = 0;
	double[][] pdbA2 = new double[noeVec.size()][3];
	double[][] pdbB2 = new double[noeVec.size()][3];
	Vector id4A = new Vector();
	Vector id4B = new Vector();
	String id = "";  //Combining the residueNO and Atom name for easy comparion.
	double[][] pdbA = new double[noeVec.size()][3];
	double[][] pdbB = new double[noeVec.size()][3];
	double [][] noeDistance = new double[noeVec.size()][2];
	for (i=0; i<noeVec.size(); i++){
	    noA = ((Noe)noeVec.elementAt(i)).getResidueNoA();
	    atomA = ((Noe)noeVec.elementAt(i)).getAtomA();
	    id = String.valueOf(noA)+atomA;
	    index = Collections.binarySearch(pdbVecA, new Pdb(noA), new Pdb.PdbComparator() );
	    if ( index > -1 ){
		pp = (Pdb)pdbVecA.elementAt(index);
		atomVec = pp.getAtomVec();
		for (j=0; j<atomVec.size(); j++){
		    cc = (Cartesian)atomVec.elementAt(j);
		    atom = cc.getAtom();
		    if ( atom.equals(atomA) )
			pdbA[i] = cc.getXYZ();
		    //for deleting repeatation in NOEs when computing the centers 
		    if ( atom.equals(atomA) && !id4A.contains(id) ) {  
			pdbA2[sizeA] = cc.getXYZ();	
			id4A.add(id);
			sizeA++;
		    }
		}
	    }
	    noB = ((Noe)noeVec.elementAt(i)).getResidueNoB();
	    atomB = ((Noe)noeVec.elementAt(i)).getAtomB();
	    id = String.valueOf(noB)+atomB;
	    index = Collections.binarySearch(pdbVecB, new Pdb(noB), new Pdb.PdbComparator() );
	    if ( index > -1 ){
		pp = (Pdb)pdbVecB.elementAt(index);
		atomVec = pp.getAtomVec();
		for (j=0; j<atomVec.size(); j++){
		    cc = (Cartesian)atomVec.elementAt(j);
		    atom = cc.getAtom();
		    if (atom.equals(atomB))
			pdbB[i] = cc.getXYZ();
		    if ( atom.equals(atomB) && !id4B.contains(id) ) {
			pdbB2[sizeB] = cc.getXYZ();	
			id4B.add(id);
			sizeB++;
		    }
		}
	    }
	    noeDistance[i] = ((Noe)noeVec.elementAt(i)).getRange();
	}
	// Compute the center
	final int n = 3;
	int mm = pdbA.length;   
	double xC1 = 0.0, yC1 = 0.0, zC1 = 0.0, xC2 = 0.0, yC2 = 0.0, zC2 = 0.0; 
	double noeAve    = 0.0;
	for (i=0; i<sizeA; i++){
	    xC1 += pdbA2[i][0];
	    yC1 += pdbA2[i][1];
	    zC1 += pdbA2[i][2];
	}
	if (sizeA > 0){ // Compute the center
 	    xC1 /= sizeA;
 	    yC1 /= sizeA;
 	    zC1 /= sizeA;
	}
	for (i=0; i<sizeB; i++){
	    xC2 += pdbB2[i][0]; 
	    yC2 += pdbB2[i][1];
	    zC2 += pdbB2[i][2];
	}
	if (sizeB > 0){ // Compute the center
 	    xC2 /= sizeB;
 	    yC2 /= sizeB;
 	    zC2 /= sizeB;
	}
	//compute average NOEs
	for (i=0; i<mm; i++)
	    noeAve += (noeDistance[i][0] + noeDistance[i][1]) / 2.0;
	noeAve /= mm;

	double lengthR = noeAve;
	double [][] arrA = new double[mm][n];
	double [][] arrB = new double[mm][n];
	double [][] newArrB = new double[mm][n];
	double [] centerA = {xC1, yC1, zC1};
 	double diffCenter = Math.sqrt((xC1-xC2)* (xC1-xC2)+(yC1-yC2) * (yC1-yC2)+(zC1-zC2) * (zC1-zC2)); 	
	double disRms = 0.0;
	double[] noeDis = new double[mm];
	for (i=0; i<mm; i++){  //move both centers to the origin
	    arrA[i][0] = pdbA[i][0] - xC1; 
	    arrA[i][1] = pdbA[i][1] - yC1;
	    arrA[i][2] = pdbA[i][2] - zC1;
	    arrB[i][0] = pdbB[i][0] - xC2; 
	    arrB[i][1] = pdbB[i][1] - yC2;
	    arrB[i][2] = pdbB[i][2] - zC2; 
	    noeDis[i]  = Math.sqrt(  (arrA[i][0] - arrB[i][0]) * (arrA[i][0] - arrB[i][0]) 
			       + (arrA[i][1] - arrB[i][1]) * (arrA[i][1] - arrB[i][1]) 
			       + (arrA[i][2] - arrB[i][2]) * (arrA[i][2] - arrB[i][2]));
	}
 	double x1, y1, z1, x2, y2, z2;
	double ranTheta, ranPhi,  ranThetaS = 0.0, ranPhiS = 0.0;
	double disRmsd = 1000.0;
	long seed = 9752864; //419057923;
	Random rr = new Random(seed);
	double low = 0.0, upper = 0.0;
	int resolution = 1;
	int noeRangeRes = 100;
 	double lenS = 0.0;
	for (int m = 0; m < (180/resolution); m++){
	    ranTheta = m * resolution * Math.PI / 180.0;
	    for (j=0; j< (360 / resolution); j++){
		ranPhi = j * resolution * Math.PI / 360.0;
		for (int k =0 ; k< noeRangeRes; k++){
		    lengthR  = noeAve + 3.0*(0.50-rr.nextDouble());
		    //move the center of B to a new position on the upper sphere, recompute B'coordinates
		    disRms = 0.0;
		    for (i = 0; i < mm; i++){
			newArrB[i][0] = arrB[i][0] + lengthR * Math.sin(ranTheta) * Math.cos(ranPhi); 
			newArrB[i][1] = arrB[i][1] + lengthR * Math.sin(ranTheta) * Math.sin(ranPhi);
			newArrB[i][2] = arrB[i][2] + lengthR * Math.cos(ranTheta);
			noeDis[i]  = Math.sqrt((arrA[i][0] - newArrB[i][0]) * (arrA[i][0] - newArrB[i][0]) 
					       +(arrA[i][1] - newArrB[i][1]) * (arrA[i][1] -newArrB[i][1]) 
					       +(arrA[i][2] - newArrB[i][2]) * (arrA[i][2] -newArrB[i][2]));
			if (noeDis[i] > noeDistance[i][0] && noeDis[i] < noeDistance[i][1])
			    disRms += 0.0;
			else if(noeDis[i] <= noeDistance[i][0] )
			    disRms += noeDistance[i][0] - noeDis[i];
			else if(noeDis[i] >= noeDistance[i][1])
			    disRms += noeDis[i] - noeDistance[i][1];
		    }
		    disRms = Math.sqrt(disRms / mm);
		    if ( disRms < disRmsd || Math.abs(disRms - disRmsd) < Const.eps ){
			ranThetaS = ranTheta;
			ranPhiS =  ranPhi;
			disRmsd = disRms;
			lenS = lengthR;
			//System.out.println(disRmsd+" ... "+ranThetaS+"  "+ranPhiS+"  "+lenS); //Check the progress
		    }
		}
	    }
	}
	noeRms[0] = disRmsd;
	if (debugNOE){
	    System.out.println("NOE rmsd = "+disRmsd);
	    System.out.println("NOE distances difference for individual ones:");
	}
	double[] centerBNew = {-xC2 + lenS * Math.sin(ranThetaS) * Math.cos(ranPhiS),
			       -yC2 + lenS * Math.sin(ranThetaS) * Math.sin(ranPhiS),
			       -zC2 + lenS * Math.cos(ranThetaS)};
	for (i=0; i<mm; i++){  //move both centers to the origin
	    arrB[i][0] = pdbB[i][0] + centerBNew[0]; 
	    arrB[i][1] = pdbB[i][1] + centerBNew[1];
	    arrB[i][2] = pdbB[i][2] + centerBNew[2];
	    noeDis[i]  = Math.sqrt( (arrA[i][0] - arrB[i][0]) * (arrA[i][0] - arrB[i][0]) 
				   +(arrA[i][1] - arrB[i][1]) * (arrA[i][1] - arrB[i][1]) 
				   +(arrA[i][2] - arrB[i][2]) * (arrA[i][2] - arrB[i][2]));
	    if (debugNOE)  	    //Print the NOE distance after the grid-search.
		System.out.println(noeDis[i]);
	}
	Vector pdbBNew   = pp.newPdbByTranslation(pdbVecB, centerBNew);
	double [] centerANew = {-centerA[0], -centerA[1], -centerA[2]};
	Vector pdbANew   = pp.newPdbByTranslation(pdbVecA, centerANew);
	//Print the new PDBs
	Vector ppVecT = new Vector();
	ppVecT.addAll(pdbANew);
	ppVecT.addAll(pdbBNew);
	Collections.sort(ppVecT, new Pdb.PdbComparator());
	return ppVecT;
    }

    /** private utility routines **/
    /** Check magnitude of difference of scalars. **/
    private static void check(double x, double y) {
	double eps = Math.pow(2.0, -52.0);
	if (x == 0 & Math.abs(y) < 10*eps) return;
	if (y == 0 & Math.abs(x) < 10*eps) return;
	if (Math.abs(x-y) > 10*eps*Math.max(Math.abs(x),Math.abs(y))) {
	    throw new RuntimeException("The difference x-y is too large: x = " 
				       + Double.toString(x) + "  y = " + Double.toString(y));
	}
    }

    /** Check norm of difference of "vectors". **/
    private static void check(double[] x, double[] y) {
	if (x.length == y.length ) {
	    for (int i=0;i<x.length;i++) {
		check(x[i],y[i]);
	    } 
	} else {
	    throw new RuntimeException("Attempt to compare vectors of different lengths");
	}
    }

    /** Check norm of difference of arrays. **/
    private static void check(double[][] x, double[][] y) {
	Matrix A = new Matrix(x);
	Matrix B = new Matrix(y);
	check(A,B);
    }

    /** Check norm of difference of Matrices. **/
    private static void check(Matrix X, Matrix Y) {
	double eps = Math.pow(2.0,-52.0);
	if (X.norm1() == 0. & Y.norm1() < 10*eps) return;
	if (Y.norm1() == 0. & X.norm1() < 10*eps) return;
	if (X.minus(Y).norm1() > 1000*eps*Math.max(X.norm1(),Y.norm1())) {
	    throw new RuntimeException("The norm of (X-Y) is too large: " +  Double.toString(X.minus(Y).norm1()));
	}
    }

    /** Shorten spelling of print. **/
    private static void print (String s) {
	System.out.print(s);
    }

    /** Print appropriate messages for successful outcome try **/
    private static void try_success (String s, String e) {
	print(">    " + s + "success\n");
	if ( e != "" ) {
	    print(">      Message: " + e + "\n");
	}
    }
    /** Print appropriate messages for unsuccessful outcome try **/
    private static int try_failure (int count,String s,String e) {
	print(">    " + s + "*** failure ***\n>      Message: " + e + "\n");
	return ++count;
    }
}
		    
