/*****************************************************************************************
 * PhiPsi.java

 * Version:           0.1

 * Description: 

 * This class represent the \phi and \psi angled for each residue and
 * implement Various methods for exact solution and systematic search
 * including the computation of rotation to the frist peptide plane,
 * the DFS-search, and the solver of the quartic equation and the
 * computation of \phi and \psi angles from two vectors in consecutive
 * planes etc.

 * authors:
 * initials    name            organization                email
 * ---------   --------------  ------------------------    ---------------------------
 *  LW         Lincong Wang   Dartmouth College           wlincong@cs.dartmouth.edu

 * history:
 * when        who    what
 * --------    ----   ----------------------------------------------------------
 * 06/25/04    LW     initial version for publication [Wang & Donald, J Biomol NMR 29: 223-242, 2004]

/**********************************************************************************************

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.

 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 *             Contact Info:
 *                                                     Bruce Randall Donald
 *                                                     HB 6211
 *                                                     Dartmouth College
 *                                                     Hanover, NH 03755
 *                                                     brd@cs.dartmouth.edu
 *
 * If you use publish any results derived from the use of this program
 * please cite: Exact Solutions for Internuclear Vectors and Backbone
 * Dihedral Angles from NH Residual Dipolar Couplings in Two Media,
 * and Their Application in a Systematic Search Algorithm for
 * Determining Protein Backbone Structure. L. Wang and
 * B. R. Donald. Journal of Biomolecular NMR 29: 223-242 (2004)
 *
 *
 * Copyright (C) 2004  Lincong Wang and Bruce R. Donald
 *
 * <signature of Bruce Donald>, 25 June 2004 Bruce Donald, Professor of Computer Science

 *************************************************************************************************/

import java.io. *;
import java.util. *;

public class PhiPsi implements Cloneable{
    private int residueNo;  //Though Phi and Psi are between two residues 
    private String residue; //We count it as belonging to the first of the two
    private double phi;
    private double psi;

    public PhiPsi(){
	residueNo = 0;
	residue = null;
        phi = 0.0;
        psi = 0.0;
    }	
    public PhiPsi(int No){
	residueNo = No;
	residue = ""; 
	phi = 0.0;
        psi = 0.0;
    }
    public PhiPsi(int no, String re, double x, double y){
	residueNo = no;
	residue = re;
        phi = x;
        psi = y;
    }
    //getting the values	
    public int getResidueNo(){
	return residueNo;
    }	
    public String getResidue(){
	return residue;
    }	
    public double getPhi(){
	return phi;
    }
    public double getPsi(){
	return psi;
    }
    public String toString(){
  	String desc = String.valueOf(residueNo) + "  " + residue
  	    +"  " + String.valueOf(phi*180/Math.PI) + "  " 
	    + String.valueOf(psi*180.0/Math.PI);
	return desc;	
    }
    public static class PPComparator implements Comparator{
	public int compare(Object o1, Object o2){
	    PhiPsi n1 = (PhiPsi)o1;
	    PhiPsi n2 = (PhiPsi)o2;
	    int d1 = n1.getResidueNo();
	    int d2 = n2.getResidueNo();
	    if (d1 < d2)
		return -1;
	    else if (d1 > d2)
		return 1;
	    else return 0;
	}
    }
    /** calculate the length between two vectors v1 and v2 
     * The returned angle is the [0,Pi] region
     */
    public double length(double[] v1){
	double v1Len = Math.sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]);
	return v1Len;
    }
    /** calculate the angles between two vectors v1 and v2 
     * The returned angle is in the [0,Pi] range
     */
    public double interAngle(double[] v1, double[] v2){
	double v1Len = Math.sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]);
	double v2Len = Math.sqrt(v2[0]*v2[0] + v2[1]*v2[1] + v2[2]*v2[2]);
	double c = (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2]) / (v1Len * v2Len);
	return Math.acos(c);  //acos [0, PI] so should be OK
    }

    /** cal the directioal cosines of the vector vec */
    public double[] dirCos(double [] vec) {
	double len = 0.0;
	double [] dirs = new double[vec.length];
	for (int i=0; i<vec.length; i++)
	    len += vec[i] * vec[i];
	for (int j=0; j<vec.length; j++)
	    dirs[j] = vec[j] / Math.sqrt(len);
	return dirs;
    }

    /**
     * Compute the global rotation matrix for rotating a coordinate
     * frame such as the PDB frame where the two vectors NH(i)->N(i) (nToh)
     * and N(i)->CA(i) (nToca) defined into the following frame:
     * +Z-axis along the bond NH(i)->N(i), "->" means from nuclear NH(i) to nuclear N(i).
     * +Y-axis in the peptide plane i,and the angle between +Y and the N(i)->CA(i) < 90 degree.
     * +X-axis is defined based on the right-handedness.
     @param nToh  the direction cosine of N->NH vector
     @param nToca the direction cosine of N->CA vector
     @return All the rotation Matrix
    */
    public Matrix RgCal(final double[] nToh, final double[] nToca){
        double[] n2h  = dirCos(nToh); //cal directional cosines
        double[] n2ca = dirCos(nToca);
        double theta = Math.PI - interAngle(nToh, nToca);
        double xNH   = n2h[0],   yNH = n2h[1],   zNH = n2h[2];
        double xNCA  = n2ca[0], yNCA = n2ca[1], zNCA = n2ca[2];
        double[][] mat = new double[3][3];
        Matrix A = new Matrix(3,3);

        double sinTheta = 9.99;
        double alpha1 = 9.99, beta1 = 9.99, gamma1 = 9.99;
        if ( Math.abs(zNH - 1.0) < Const.eps) { //when the NH vector is on +Z direction
            if ( Math.abs(zNCA - 1.0) > Const.eps && Math.abs(zNCA + 1.0) > Const.eps ){
                sinTheta  = Math.sqrt(1.0 - zNCA * zNCA); //under such condition cosTheta = zNCA
                mat[0][0] = yNCA / sinTheta;
                mat[0][1] = xNCA / sinTheta;
                mat[0][2] = 0.0;
                mat[1][0] = -xNCA / sinTheta;
                mat[1][1] = yNCA / sinTheta;
                mat[1][2] = 0.0;
                mat[2][0] = 0.0;
                mat[2][1] = 0.0;
                mat[2][2] = 1.0;
                return (new Matrix(mat)).transpose(); //return directly the transposed matrix
            } else { //impossible for normal PDB file
                System.out.println("zNCA is 1 or -1");
                System.exit(0);
            }
        }else if ( Math.abs(zNH + 1.0) < Const.eps) { //when the NH vector is on +Z direction
            mat[0][0] = -yNCA / sinTheta;
            mat[0][1] = xNCA  / sinTheta;
            mat[0][2] = 0.0;
            mat[1][0] = xNCA / sinTheta;
            mat[1][1] = yNCA / sinTheta;
            mat[1][2] = 0.0;
            mat[2][0] = 0.0;
            mat[2][1] = 0.0;
            mat[2][2] = 1.0;
            return (new Matrix(mat)).transpose();
        }else {
            beta1 = Math.acos(zNH);
            double sinAlpha = yNH / Math.sin(beta1);
            double cosAlpha = xNH / Math.sin(beta1);
            if ( cosAlpha >= 0 )
                alpha1 = Math.asin(sinAlpha);
            else alpha1 = Math.PI-Math.asin(sinAlpha);
            double sinGamma = (zNCA+Math.cos(beta1)*Math.cos(theta))/(Math.sin(theta)*Math.sin(beta1));
            double cosGamma = (yNCA*Math.cos(alpha1)-xNCA*Math.sin(alpha1))/Math.sin(theta);
            if (cosGamma >= 0 )
                gamma1 = Math.asin(sinGamma);
            else gamma1 = Math.PI-Math.asin(sinGamma);
        }
        //make an Euler rotation matrix from three Euler angles
        Matrix mm = A.eulerMat(alpha1, beta1, gamma1);
        //invert the +z axis: point along the NH->N direction for the sake of easy math
        Matrix MM = A.rotationMat(Math.PI, "+y");
        return (MM.times(mm));
    }

     /* compute global rotation (Rg) with two noncolinear vectors, Rg
      * is the rotation matrix for rotating the PDB coordinate system
      * into the following system:
      * +z-axis along the bond NH(i)->N(i), "->" means from atom N(i) to atom NH(i). 
      * +y-axis in the peptide plane i,and the angle  between +y and the N(i)->CA(i) < 90 degree.  
      * +x-axis is defined based on the right-handedness.
      @param nToh the directional cosine of N->NH vector
      @param nToca the directional cosine of N->CA vector
      @param rightHand to compare with the rotation matrix from the SVD method. 
      @return the rotation Matrix
     */
    public Matrix RgCal(double[] nToh, double[] nToca, boolean rightHand){
	double[] n2h = dirCos(nToh); //cal directional cosines
	double[] n2ca = dirCos(nToca);
	double theta = Math.PI - interAngle(nToh, nToca);
	double xNH = n2h[0], yNH = n2h[1], zNH = n2h[2];
	double xNCA = n2ca[0], yNCA = n2ca[1], zNCA = n2ca[2];
	double[][] mat = new double[3][3];
	double sinTheta = 9.99;
	Matrix A = new Matrix(3,3);
	double alpha1=9.99,  beta1=9.99,  gamma1=9.99;
        if ( Math.abs(zNH - 1.0) < Const.eps) { //when the NH vector is on +Z direction
            if ( Math.abs(zNCA - 1.0) > Const.eps && Math.abs(zNCA + 1.0) > Const.eps ){
		sinTheta = Math.sqrt(1.0-zNCA*zNCA); //under such condition cosTheta = zNCA
		mat[0][0] = yNCA / sinTheta;
		mat[0][1] = xNCA / sinTheta;
		mat[0][2] = 0.0;
		mat[1][0] = -xNCA / sinTheta;
		mat[1][1] = yNCA / sinTheta;
		mat[1][2] = 0.0;
		mat[2][0] = 0.0;
		mat[2][1] = 0.0;
		mat[2][2] = 1.0;
		return (new Matrix(mat)).transpose();
	    } else { //impossible for normal PDB file
		System.out.println("zNCA is 1 or -1");
		System.exit(0);
	    }
	}else if ( Math.abs(zNH + 1.0) < Const.eps) { //when the NH vector is on +Z direction
	    mat[0][0] = -yNCA / sinTheta;
	    mat[0][1] = xNCA / sinTheta;
	    mat[0][2] = 0.0;
	    mat[1][0] = xNCA / sinTheta;
	    mat[1][1] = yNCA / sinTheta;
	    mat[1][2] = 0.0;
	    mat[2][0] = 0.0;
	    mat[2][1] = 0.0;
	    mat[2][2] = 1.0;
	    return (new Matrix(mat)).transpose(); //return directly need transpose
	}else {
	    beta1 = Math.acos(zNH);  //Should be minus
	    double sinAlpha = yNH / Math.sin(beta1);
	    double cosAlpha = xNH / Math.sin(beta1);
	    if (( cosAlpha > 0) || (cosAlpha == 0) )
		alpha1 = Math.asin(sinAlpha);
	    else if ( cosAlpha < 0)
		alpha1 = Math.PI - Math.asin(sinAlpha);
	    double sinGamma = (zNCA + Math.cos(beta1) * Math.cos(theta)) / (Math.sin(theta)*Math.sin(beta1));
	    double cosGamma = (yNCA * Math.cos(alpha1) - xNCA*Math.sin(alpha1)) / Math.sin(theta);
	    if (( cosGamma > 0) || (cosGamma == 0) )
		gamma1 = Math.asin(sinGamma);
	    else if ( cosGamma < 0 )
		gamma1 = Math.PI - Math.asin(sinGamma);
	}//make a euler rotation matrix from three Euler angles
	Matrix MM = A.rotationMat(Math.PI, "+y");
 	Matrix mm = A.eulerMat(alpha1, beta1, gamma1);
	if (rightHand)
	    return MM.times(mm);
	return (Const.mLeftHand.times(MM.times(mm)));
    }

    /**
     * construct the coeficiencies of the quartic equation from Saupe
     * elements, RDC data and the relative rotation Matrix between
     * the two POFs; the Saupe matrix and the relative rotation matrix are
     * computed by SVD method for average helix or other methods
     @param Syy diagnolized Saupe element for medium 1
     @param Szz diagnolized Saupe element for medium 1
     @param Syy1 diagnolized Saupe element for medium 2
     @param Szz1 diagnolized Saupe element for medium 2
     @param rPOF the relative POFs between two matrices
     @param coefs save the constructed coefficiencies
     @return true if the construction is successful, otherwise, false.
     */	
    public boolean quarticCoef(double Syy, double Szz, double Syy1, double Szz1,
		       double rdc1, double rdc2, Matrix rPOF, double[] coefs){
	double a1 = -Syy - 2.0*Szz; //Sxx=-Syy-Szz, a1 = Sxx-Szz
	double b1 = Syy - Szz;
	double r1 = rdc1 - Szz;
	//use the matrix element directly
	double r11 = rPOF.get(0,0);
	double r12 = rPOF.get(0,1);
	double r13 = rPOF.get(0,2);
	double r21 = rPOF.get(1,0);
	double r22 = rPOF.get(1,1);
	double r23 = rPOF.get(1,2);
	double r31 = rPOF.get(2,0);
	double r32 = rPOF.get(2,1);
	double r33 = rPOF.get(2,2);

	double ax = -Syy1 - 2.0*Szz1;
	double bx = Syy1 - Szz1;
	double a2 = (ax*r11*r11+bx*r21*r21) - (ax*r13*r13 + bx * r23 *r23);
	double b2 = (ax*r12*r12+bx*r22*r22) - (ax*r13*r13 + bx * r23 *r23);
	double c1 = (ax*2*r11*r12+bx*2*r21*r22);
	double c2 = (ax*2*r11*r13+bx*2*r21*r23);
	double c3 = (ax*2*r12*r13+bx*2*r22*r23);
	double r2 = rdc2 - Szz1 - (ax*r13*r13 + bx * r23 *r23);

	double d8 = a2*a2 + c2*c2;
	double d7 = 2*a2*c1 + 2*c2*c3;
	double d6 = c1*c1 + 2*a2*b2 + c2*c2 + c3*c3;
	double d5 = -2*a2*r2 - c2*c2;
	double d4 = 2*b2*c1 + 2*c2*c3;
	double d3 = -2*r2*c1 - 2*c2*c3;
	double d2 = -2*r2*b2 - c3*c3;
	double d1 = b2*b2+c3*c3;
	if (a1 == 0.0 || b1 == 0.0){
	    System.out.println(" Saupe elements: Sxx-Szz is zero");
	    return false;
	}else if ( (r1/a1 < 0.0) && (r1/b1 > 0.0) || (r1/a1 > 0.0) && (r1/b1 < 0.0) ){
	    System.out.println("Hyperbola"); //Sth. wrong if it is a Hyperbola equation??
	    return false;
	}else if ( (r1/a1 < 0.0) || (r1/b1 < 0.0) ){
// 	    System.out.println("Nagetive"); //Sth. wrong if it is a Hyperbola equation??
	    return false;
	}
	double a = Math.sqrt(r1/a1);
	double b = Math.sqrt(r1/b1);
	double e1 = d8*a*a*a*a + d1*b*b*b*b - d6*a*a*b*b;
	double e2 = d4* a*b*b*b - d7 * a*a*a*b; 
	double e3 = 2*(d1*b*b*b*b +d2*b*b - d5*a*a - d8*a*a*a*a);
	double e4 = d7*a*a*a*b + d4*a*b*b*b + 2*d3*a*b;
	double e0 = 4*r2*r2 + d8*a*a*a*a + 2*d5*a*a + 2*d2*b*b + d1*b*b*b*b + d6*a*a*b*b;
	coefs[0] = e1*e1 + e2*e2;  //The five coeffs [0,4] for the quartic equation
	coefs[1] = 2*e1*e3+2*e2*e4;
	coefs[2] = 2*e1*e0 + e3*e3 + e4*e4 - e2*e2;
	coefs[3] = 2*e3*e0 -2*e2*e4;
	coefs[4] = e0*e0 - e4*e4;
	coefs[5] = a1; //pass the other coefficients over too for convenience.
	coefs[6] = b1;	
	coefs[7] = r1;	
	coefs[8] = a2;
	coefs[9] = b2;	
	coefs[10] = r2;	
	coefs[11] = c1;
	coefs[12] = c2;	
	coefs[13] = c3;	
	return true;
    }

    /** 
     * solve the quartic equation by eigenValues method (cf. NR book),
     * which is about 2-fold slower than other methods.
     @param coefs 14 coefficiencies for the quartic equation and for checking solutions.
     @return a vector of \phi and \psi solutions.
     */
    public Vector quarticSolve(double[] coefs){
	double c1 = coefs[0];
	double c2 = coefs[1];
	double c3 = coefs[2];
	double c4 = coefs[3];
	double c0 = coefs[4];
	double a1 = coefs[5];;
	double b1 = coefs[6];	
	double r1 = coefs[7];	
	double a2 = coefs[8];
	double b2 = coefs[9];	
	double r2 = coefs[10];	
	double cc1 = coefs[11];
	double cc2 = coefs[12];	
	double cc3 = coefs[13];	
	if (a1 == 0.0 || b1 == 0.0){
	    System.out.println(" Saupe elements: Sxx-Szz is zero");
	    System.exit(1);
	}
	double a = Math.sqrt(r1/a1);
	double b = Math.sqrt(r1/b1);
	double xr = 0.0, yr = 0.0, zr = 0.0;
	double x = 0.0, y = 0.0, z = 0.0;
	double r1Cal=0.0, r2Cal = 0.0;
	final int [] indexI  = {1,1,1, 1,1,-1, 1,-1,1, 1,-1,-1, -1,1,1, -1,1,-1, -1, -1, 1, -1,-1,-1};
	final int eightFold = 24; //For convenience
	final double eps = 1.0E-8;
	Vector rootVec = new Vector(); //a root vector in PDB Class format
	double [] polar = new double[2]; //polarFrom3D(nToNH1);
	double [] nhVec = new double[3]; 
	if( c1 != 0.0){
	    double[][] mArr = { {-c2/c1, -c3/c1, -c4/c1, -c0/c1},
                                {1.0,    0.0,    0.0,    0.0},
                                {0.0,    1.0,    0.0,    0.0},
                                {0.0,    0.0,    1.0,    0.0} };
	    Matrix mm = new Matrix(mArr);
	    EigenvalueDecomposition eigs = new  EigenvalueDecomposition (mm, false);
	    double [] realValues = eigs.getRealEigenvalues();
	    for (int i=0; i<realValues.length; i++){
		xr = a * Math.sqrt(0.50 * (1.0 - realValues[i]));
		yr = b * Math.sqrt(0.50 * (1.0 + realValues[i]));
		if ( (xr*xr + yr*yr) < 1.0){
		    zr = Math.sqrt(1.0 - xr*xr - yr*yr);
		    for (int m = 0; m < eightFold; m += 3){
			x = indexI[m] * xr;
			y = indexI[m+1] * yr;
			z = indexI[m+2] * zr;
			r2Cal = a2 * x*x + b2 * y*y + cc1 * x*y + cc2 * x*z + cc3 * y*z;
			r1Cal = a1 * x*x + b1 * y*y;
			if ( Math.abs(r1Cal - r1 ) < eps && Math.abs(r2Cal - r2 ) < eps){
			    rootVec.add(new double[]{x, y, z}); 
			}
		    }
		}
	    }
	}
	return rootVec;
    }

   /**
    **********************************************************************************
    * This method cal the \Phi and \Psi pair from two consecutive NH vectors EXACTLY
    * assuming the six angles are known.
    * These are the steps (necessary rotations) taken from NH1 to NH2.

     I:   Rotate along +x to bring the +y in the direction N(i)->CA(i)
     angle(NH(i)-N(i)-CA(i)) = 2.0*pi / 3.0; % standard
     theta_1 = angle(NH(i)-N(i)-CA(i)) - pi/2.0;
     
     R_1x = [1 0 0; 0 cos(theta_1) sin(theta_1); 0 -sin(theta_1) cos(theta_1)];

     I':   Rotate along +y to bring the bond N(i)->NH(i) into the plane of 
           C(i-1)->N(i) and N(i)->CA(i).the dihedral angle theta9 defined as
           the angle between the plane defined by (NH(i)->N(i)->CA(i)) 
	   and the plane defined by (CO(i-1)->N(i)->CA(i)).

	   R_9y = [cos(theta9) 0 -sin(theta9); 0 1 0;  sin(theta9) 0 cos(theta9)] ;

     II:  Rotate along the +y-axis with dihedral angle phi  by its definition
     y in the N(i)->CA(i) direction, z in the plane defined by N(i)->CA(i)->CO(i)

     R_2y = [cos(phi) 0 -sin(phi); 0 1 0;  sin(phi) 0 cos(phi)] ;

     III: Rotate along +x to bring +z in the direction CA(i)->CO(i)
     theta_3 = -(angle(N(i)-CA(i)-CO(i)) - pi/2) = 109 * pi / 180.0;  % standard

     R_3x = [1 0 0; 0 cos(theta_3) sin(theta_3); 0 -sin(theta_3) cos(theta_3)];

     IV:  Rotate along +z with psi as required by the definition, Please note pi
     needs to be added to be consistent with the definition
     z is in the direction along CA(i)->CO(i) and 
     y in the plane defined by CA(i)->CO(i)->N(i+1).

     R_4z = [cos(psi+pi) sin(psi+pi) 0; -sin(psi+pi) cos(psi+pi) 0; 0 0 1] ;

     V:   Rotate along +x to bring +y in the direction of CO(i)->N(i+1)
     theta_5 = angle(CA(i)-CO(i)-N(i+1)) - pi/2

     R_5x = [1 0 0; 0 cos(theta_5) sin(theta_5); 0 -sin(theta_5) cos(theta_5)];

     VI:  Rotate along +y with the dihedral angle omegaNH defined as
     the angle between the plane defined by (CA(i)->CO(i)->N(i+1)) and the plane 
     defined by (CO(i)->N(i+1)->NH(i+1)). theta_7 = omegaNH
     y is along the CO(i)->N(i+1) direction.
     z is in the plane defined CO(i)->N(i+1)->NH(i+1)

     R_6y = [cos(omegaNH) 0 -sin(omegaNH); 0 1 0;  sin(omegaNH) 0 cos(omegaNH)] ;

     VII: Rotate along +x to bring +z along the bond NH(i+1)->N(i+1)
     theta_7 = angle(CO(i)-N(i+1)-NH(i+1)) - pi /2 = pi/3;      %stardand?

     R_7x = [1 0 0; 0 cos(theta_7) sin(theta_7); 0 -sin(theta_7) cos(theta_7)];

     The final frame should be exactly like the first frame.

     The total Rotations

     R = R_7x*R_6y*R_5x*R_4z*R_3x*R_2y*R_9y*R_1x;      %the total rotation
 
     let Rg be the unique global rotation from the PDB frame (XYZ) (or any other) to
     the first xyz-frame in peptide plane i then we should have

      [0 0 -1] = R(i,i+1)*Rg*[X(i+1) Y(i+1) Z(i+1)]; % For N(i+1)->NH(i+1)
      [0 0 -1] = Rg*[X(i) Y(i) Z(i)];                % For N(i)->NH(i)

     In other words,

      [X(i+1) Y(i+1) Z(i+1)] = inv(Rg) * inv(R(i,i+1)) * Rg * [X(i) Y(i) Z(i)]; % Equ(1)
     where inv means matrix inverse. All these inverses should exist

     ************************ ****************************************************
     *
     @param rg    the rotation matrix from a unique global frame to a local frame defined above
     @param nh2   the direction cosine for amide proton
     @return  two pairs of phipsi solutions.
     * The six angles are assumed to be constants
     **/    
    public double[] PhiPsiCalRDC(Matrix rg, double[] nh2){
	Matrix rgInv = rg.transpose();
   	double [] v1 = Const.dirCosNHcnt;
	double [] v2 = Const.r9y.times(Const.r1x.times(rg.times(dirCos(nh2))));
	double x1 = v1[0], y1 = v1[1], z1 = v1[2];
	double x2 = v2[0], y2 = v2[1], z2 = v2[2];
	double r = Math.sqrt(x2*x2 + z2*z2);
	double [] phiPsiPair = new double[]{9.99, 9.99, 9.99, 9.99};
	if (Math.abs(r) < 0.01*Const.eps){ //r should != 0.0, otherwise there are errors
  	    System.out.println("Error in Phipsi Cal:  r = "+r);
  	    return phiPsiPair;
  	}
	double r1 = Math.abs((z1 + y2 * Math.sin(Const.theta3)) / (r * Math.cos(Const.theta3)));
  	if ( (r1 - 1.0) > Const.eps){ //r1 should <= 1.0, otherwise there are errors
//   	    System.out.println("Error in Phipsi Cal:  r1 = "+r1);
  	    return phiPsiPair;
  	}
	double Sc = z2 / r; 
	double Cc = x2 / r ; 
	double theta = 0.0;
	if ( Cc >= 0.0 )
	    theta = Math.asin(Sc);
	else if (Cc < 0)
	    theta = Math.PI-Math.asin(Sc);

	double phi1 = -theta+Math.asin((z1+y2*Math.sin(Const.theta3))/(r*Math.cos(Const.theta3)));
	if (phi1 < (-Math.PI))
	    phi1 += 2*Math.PI;
	else if (phi1 > Math.PI)
	    phi1 -= 2*Math.PI;
	double phi2 = Math.PI-theta-Math.asin((z1+y2*Math.sin(Const.theta3))/(r*Math.cos(Const.theta3)));
	if (phi2 < (-Math.PI))
	    phi2 += 2*Math.PI;
	else if (phi2 > Math.PI)
	    phi2 -= 2*Math.PI;
	double b = x2*Math.cos(phi1) - z2*Math.sin(phi1);
	double a = -x2*Math.sin(phi1)*Math.sin(Const.theta3) - y2*Math.cos(Const.theta3)
	           -z2*Math.cos(phi1)*Math.sin(Const.theta3);
	r1 = (x1*x1 + y1*y1);
	double dd = a*a + b*b;
	double sinPsi, cosPsi;
	double psi1 = 9.99, psi2 = 9.99;
	if ( !(r1 == 0.0)){
	    sinPsi = (a*x1+b*y1) / r1; 
	    cosPsi = (a*y1-b*x1) / r1;
	    if ( cosPsi >= 0.0 )
		psi1 = Math.asin(sinPsi);
	    else if ( cosPsi < 0)
		psi1 = Math.PI-Math.asin(sinPsi);
	    if (psi1 < (-Math.PI))
		psi1 += 2*Math.PI;
	    else if (psi1 > Math.PI)
		psi1 -= 2*Math.PI;
	}else{ //Such situation should not happened
	    System.out.println("Psi can NOT be calculated");
	    System.exit(1);
	}
	//The second pair
	b = x2*Math.cos(phi2)-z2*Math.sin(phi2);
	a = -x2*Math.sin(phi2)*Math.sin(Const.theta3) - y2*Math.cos(Const.theta3)
	    -z2*Math.cos(phi2)*Math.sin(Const.theta3);
	if ( !(r1 == 0.0)){
	    sinPsi = (a*x1+b*y1) / r1; 
	    cosPsi = (a*y1-b*x1) / r1;
	    if ( cosPsi >= 0.0 )
		psi2 = Math.asin(sinPsi);
	    else if ( cosPsi < 0)
		psi2 = Math.PI-Math.asin(sinPsi);
	    if (psi2 < (-Math.PI))
		psi2 += 2*Math.PI;
	    else if (psi2 > Math.PI)
		psi2 -= 2*Math.PI;
	}else { //Such situation should not happened
	    System.out.println("Psi can NOT be calculated");
	    System.exit(1);
	}
	phiPsiPair = new double[]{phi1, psi1, phi2, psi2};
  	return phiPsiPair;
    }

    /**
     * A method similar to the previous except that only return \Phi and \Psi in the 
     * left half.
     @param two when only two Phipsi pairs (in the left of Rma plot)
     are returned
     @return an array in the order of \phi and \psi
     */
    public double[] PhiPsiCalRDC(Matrix rg, double[] nh2, boolean two){
	Matrix rgInv = rg.transpose();
   	double [] v1 = Const.dirCosNHcnt;
	double [] v2 = Const.r9y.times(Const.r1x.times(rg.times(dirCos(nh2))));
	double x1 = v1[0], y1 = v1[1], z1 = v1[2];
	double x2 = v2[0], y2 = v2[1], z2 = v2[2];
	double r = Math.sqrt(x2*x2 + z2*z2);
	double [] phiPsiPair = new double[]{9.99, 9.99};
	if (Math.abs(r) < 0.01*Const.eps){ //r should != 0.0, otherwise there are errors
   	    System.out.println("Error in Phipsi Cal:  r = "+r);
  	    return phiPsiPair;
  	}
	double r1 = Math.abs((z1 + y2*Math.sin(Const.theta3)) / (r*Math.cos(Const.theta3)));
  	if ( (r1 - 1.0) > Const.eps){ //r1 should <= 1.0, otherwise there are errors
//   	    System.out.println("Error in Phipsi Cal:  r1 = "+r1);
  	    return phiPsiPair;
  	}
	double Sc = z2 / r; 
	double Cc = x2 / r ; 
	double theta = 0.0;
	double phi = 9.99;
	if ( Cc >= 0.0 )
	    theta = Math.asin(Sc);
	else if (Cc < 0)
	    theta = Math.PI - Math.asin(Sc);

	double phi1 = -theta + Math.asin((z1+y2*Math.sin(Const.theta3)) / (r*Math.cos(Const.theta3)));
	if (phi1 < (-Math.PI))
	    phi1 += 2*Math.PI;
	else if (phi1 > Math.PI)
	    phi1 -= 2*Math.PI;
	double phi2 = Math.PI - theta - Math.asin((z1+y2*Math.sin(Const.theta3))/(r*Math.cos(Const.theta3)));
	if (phi2 < (-Math.PI))
	    phi2 += 2*Math.PI;
	else if (phi2 > Math.PI)
	    phi2 -= 2*Math.PI;

	if (phi1 > 0.0 && phi2 > 0.0) {//System.out.println(" Sign error in Phi "+ (phi /cst) +"  "+(phi1/cst)+"  "+(phi2/cst));
	    return phiPsiPair;
	}else if ( phi1 <= 0.0)
	    phi = phi1;
	else if (phi2 <= 0.0)
	    phi = phi2;

	double b = x2*Math.cos(phi) - z2*Math.sin(phi);
	double a = -x2*Math.sin(phi)*Math.sin(Const.theta3) - y2*Math.cos(Const.theta3)
	           -z2*Math.cos(phi)*Math.sin(Const.theta3);
	r1 = (x1*x1+y1*y1);
	double dd = a*a + b*b;
	double sinPsi, cosPsi;
	double psi = 9.99;
	if ( !(r1 == 0.0)){
	    sinPsi = (a*x1 + b*y1) / r1; 
	    cosPsi = (a*y1 - b*x1) / r1;
	    if ( cosPsi >= 0.0 )
		psi = Math.asin(sinPsi);
	    else if ( cosPsi < 0)
		psi = Math.PI-Math.asin(sinPsi);
	    if (psi < (-Math.PI))
		psi += 2*Math.PI;
	    else if (psi > Math.PI)
		psi -= 2*Math.PI;
	}else { //Such situation should not happened
	    System.out.println("Psi can NOT be calculated");
	    System.exit(1);
	}
	phiPsiPair = new double[]{phi, psi};
  	return phiPsiPair;
    }

    /* A recusive function to compute all the backbone Phi/Psi for n residues
       @param n2CA1     the N->CA vector of the first residue 
       @param nh1     the N->NH vector of the first residue 
       @param allRoots  a vector of all NH vector solved from the quartic equation.
       @param depth     the depth of the DFS-tree, begining with "1"
       @param ppVec     save the computed \phi and \psi angle
       @param phiS      save the computed intermediate \phi angle
       @param psiS      save the computed intermediate \psi angle
       @param pathIndex save the index of the search path
       @param startInd  index into the vector with all the solutions for two consecutive RDCs.
       @param ramaFilter Ramachandran filter in the order of [phiLow, phiHigh, psiLow, psiHigh]
       @return true if there is a path going to the bottom, otherwise, false.
       *All the \phi and \psi solutions are returned in the vector ppVec.
     */
    public boolean phiPsiChain (double[] n2CA1, double[] nh1, Root [] allRoots, int depth, Vector ppVec, 
				double[] phiS, double[] psiS, int[] pathIndex, int startInd, double[] ramaFilter){
	Vector rootVec2  = allRoots[depth].getRootVec();
	final int lastNo = allRoots.length;  //It is also the number of residues
	double [] ppp  = new double [2];
	double [] nh2  = new double [3];  //the 2nd NH vector
	double [] n2CA = new double [3];  //the 2nd NH vector
	Matrix rgInv  = new Matrix(3, 3);
	Matrix r2yInv = new Matrix(3, 3);
	Matrix r4zInv = new Matrix(3, 3);
	Matrix rg = RgCal(nh1, n2CA1);
	int pos = 0;;
	int i = 0;
	int pIndex, parent = 0;
	boolean status = false;
	Pdb pp = new Pdb();
	if (depth < (lastNo - 1 ) ){ //when have not research the leaves, do DFS search
 	    for (i = startInd; i < rootVec2.size(); i++){
		nh2 = (double[])(rootVec2.elementAt(i));
    		ppp = PhiPsiCalRDC(rg, nh2, true);
		if (ppp[0] > ramaFilter[0] && ppp[0] <ramaFilter[1] && ppp[1] > ramaFilter[2] && ppp[1] < ramaFilter[3]){
		    phiS[depth - 1] = ppp[0];
		    psiS[depth - 1] = ppp[1];
		    rgInv  = rg.transpose();
		    r2yInv = rg.rotationMat(ppp[0], "-y");  //we need their transpose
		    r4zInv = rg.rotationMat(ppp[1] + Math.PI, "-z"); 
		    n2CA = rgInv.times((Const.matL).times(r2yInv.times(Const.r3xInv.times(r4zInv.times(Const.matR)))));
		    pathIndex[depth] = i;  //the index to all NH vector solutions 
		    status = phiPsiChain(n2CA, nh2, allRoots, depth+1, ppVec, phiS, psiS, pathIndex, 0, ramaFilter);
		    if (!status){//if the paths originating from the startInd failed to rearch the leaves.
			parent = depth; //backtrack one depth up
			if ( parent > 0){
			    pIndex = pathIndex[parent];
			    pos = pIndex + 1;
			    if (pos < (allRoots[parent].getRootVec()).size()){
				status = phiPsiChain(n2CA1, nh1, allRoots, depth, ppVec, phiS, psiS, pathIndex,pos,ramaFilter);
				if ( !status)
				    return false;
			    }
			}
		    }
		}
	    }
 	    return false;
	}else{  //when have not research the leaves, return all the Phi/Psi pairs
	    boolean flag = false;
	    for (i = 0; i < rootVec2.size(); i++){
		nh2 = (double[])(rootVec2.elementAt(i));
    		ppp = PhiPsiCalRDC(rg, nh2, true);
		if (ppp[0] > ramaFilter[0] && ppp[0] <ramaFilter[1] && ppp[1] > ramaFilter[2] && ppp[1] < ramaFilter[3]){
		    pathIndex[depth] = i;
		    for (int j = 0; j < lastNo-2; j++){
			ppVec.addElement(new Double(phiS[j]));
			ppVec.addElement(new Double(psiS[j]));
		    }
		    ppVec.addElement(new Double(ppp[0]));
		    ppVec.addElement(new Double(ppp[1]));
		    flag = true;
		}
	    }
	    return flag;
	}
    }

    protected Object clone(){
        try {
	    Object s = super.clone();     
	    return s;                     
        } catch (CloneNotSupportedException e) {
	    throw new InternalError();
        }
    }
}
